/*
 * $Id: Config.java 5981 2024-10-29 16:34:16Z carsten $
 *
 * Copyright 2010/2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po;

import static de.cm.osm2po.misc.ArrayUtils.stringArrayConcat;
import static de.cm.osm2po.misc.Utils.TS;
import static de.cm.osm2po.misc.Utils.findFileOrDir;
import static de.cm.osm2po.misc.Utils.getJarDir;
import static de.cm.osm2po.misc.Utils.getJarFile;
import static de.cm.osm2po.primitives.VarTypeDesk.registerType;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import java.net.URLStreamHandlerFactory;
import java.security.ProtectionDomain;
import java.sql.Connection;
import java.sql.Driver;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.TreeMap;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import de.cm.osm2po.ShutdownHookSequence.ShutdownAware;
import de.cm.osm2po.converter.graph.DefaultGraphBuildInterceptor;
import de.cm.osm2po.converter.graph.GraphBuildInterceptor;
import de.cm.osm2po.converter.joiner.DefaultJoinFilter;
import de.cm.osm2po.converter.joiner.JoinFilter;
import de.cm.osm2po.converter.parser.OsmParser;
import de.cm.osm2po.converter.parser.OsmXmlParser;
import de.cm.osm2po.converter.parser.QuickNet;
import de.cm.osm2po.converter.tiler.DefaultNodeTagResolver;
import de.cm.osm2po.converter.tiler.DefaultRelationTagResolver;
import de.cm.osm2po.converter.tiler.DefaultWayTagResolver;
import de.cm.osm2po.converter.tiler.NodeTagResolver;
import de.cm.osm2po.converter.tiler.RelationTagResolver;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.converter.tiler.WayTagResolver;
import de.cm.osm2po.extensions.BndExtension;
import de.cm.osm2po.extensions.DbExtension;
import de.cm.osm2po.extensions.MlgExtension;
import de.cm.osm2po.extensions.PoiExtension;
import de.cm.osm2po.extensions.SndExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.logging.LogConsoleWriter;
import de.cm.osm2po.logging.LogFileWriter;
import de.cm.osm2po.logging.LogWriter;
import de.cm.osm2po.misc.OtelAgentWrapper;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.model.Membership;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.OsmPolygon;
import de.cm.osm2po.model.OsmRelation;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.model.Restriction;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.postp.PostProcessor;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarTypeDesk;
import de.cm.osm2po.routing.AbstractRouter;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayEdge;
import de.cm.osm2po.routing.OverlayNode;
import de.cm.osm2po.routing.OverlayVertexRouter;
import de.cm.osm2po.routing.RoutingResultSegment;
import de.cm.osm2po.routing.VertexRouter;

/**
 * Uebersetzt Properties frei definierter oder per {@link Main#main(String[])}
 * eingelesener Parameter in Members oder reicht die {@link Properties} als
 * vereinfachte Fassade nach aussen. Zudem implementiert diese Klasse das
 * {@link Iterable}-Interface und liefert hier die Schluessel aller Properties.<br>
 * Es existieren drei Orte, wo sie gesetzt bzw. auch wieder ueberschrieben werden
 * koennen.<br>Die Prioriaet ist:
 * <ol>
 * <li>CLI (Kommandozeilen-Parameter)
 * <li>Konfigurationsdatei (default ist <code>osm2po.config</code>)
 * <li><code>osm2po.selfreg</code> (Datei im Classpath-Root)
 * </ol>
 * Die Konfigurationsdatei wird an den folgenden Orten gesucht:
 * <ol>
 * <li>osm2po.config aus catalina.base/conf, falls Tomcat</li>
 * <li>osm2po.config aus aktuellem oder aus dem Binaer/Jar-Verzeichnis</li>
 * <li>osm2po.config aus ClassPath-Root z.B. aus Plugins-jar</li>
 * <li>Fallback: osm2po.config aus dem Core-Package-Jar (Default-Settings)</li>
 * </ol>
 * Alternativ kann mittels Parameter <code>config=&lt;Pfad&gt;</code> eine
 * andere Konfigurationsdatei gesetzt werden.<br>
 * Sowie <code>config=</code> (ohne Pfad), als auch ein falscher Pfad,
 * verhindert ein Lesen der Konfiguration. Der Parameter selbst kann per
 * CLI oder mittels der Datei <code>osm2po.selfreg</code> (im Classpath-Root)
 * gesetzt werden. Ferner ist es moeglich, eine Konfiguration pro
 * {@link Props#copyCase(String...)} vorzuhalten.<br>
 * Hierzu muss der Parameter explizit erweitert werden:<br>
 * <code>config.&lt;case&gt;=PathToConfig</code><br>
 * Danach wird versucht, die Datei mit den folgenden Methoden zu finden:
 * <ol>
 * <li>Pfad als {@link URL} ausprobieren. (somit auch http moeglich) </li>
 * <li>Pfad als lokales {@link File} suchen</li>
 * <li>Pfad relativ zur aktuellen Umgebung (z.B. Jar) suchen.</li>
 * <li>Pfad als Classpath verwenden und dort suchen</li>
 * </ol>
 * <b>Wichtig:</b><br>
 * Liegen mehrere osm2po.selfreg-Dateien im ClassPath (z.B. pro plugin.jar),
 * so ist die Verarbeitungsreihenfolge beliebig. Es wird aber immer
 * eine osm2po.config nach allen osm2po.selfregs verarbeitet.
 *
 * @author (c) 2014 - Carsten Moeller - info@osm2po.de
 * @see Props
 */
public class Config implements ShutdownAware, Iterable<String> {

    private Log log;
    private Props props;
    private String command;
    private String prefix;
    private File workDir;
    private long reservedXmx;

    private URL sourceFile;

    private Class<? extends WayTagResolver> wayTagResolverClass;
    private Class<? extends NodeTagResolver> nodeTagResolverClass;
    private Class<? extends RelationTagResolver> relationTagResolverClass;
    private Class<? extends GraphBuildInterceptor> graphBuildInterceptorClass;
    private Class<? extends JoinFilter> joinFilterClass;
    private Map<Integer, Class<? extends AbstractRouter<?>>> routerClassMap =
        new HashMap<Integer, Class<? extends AbstractRouter<?>>>();
    private Map<Integer, Class<? extends PostProcessor>> postProcessorClassMap =
        new TreeMap<Integer, Class<? extends PostProcessor>>(); //Order-Safe!
    private List<File> files = new ArrayList<File>();
    private List<File> dirs = new ArrayList<File>();
    private Map<String, Class<? extends OsmParser>> osmParserClassMap =
        new HashMap<String, Class<? extends OsmParser>>();
    private Map<String, VarBean> beanMap = new HashMap<String, VarBean>();

    /**
     * Liefert das Prefix des Postprozessors in Abhaengikeit
     * vom Index.
     * @param index int Index
     * @return Prefix des Prozessors ohne abschliessendem Punkt.
     */
    public final static String POSTP(int index) {
        return "postp." + index;
    }

    private static ClassLoader classLoader;
    private static URL[] jarUrls;
    static {
        jarUrls = findPluginJars();
        classLoader = URLClassLoader.newInstance(jarUrls, Config.class.getClassLoader());
        registerTypes();
        URL.setURLStreamHandlerFactory(new ClassPathStreamHandlerAndFactory());
    }

    public static void logInfoJarUrls(Log log) {
        for (URL plugin : jarUrls) log.info(
                "Plugin " + plugin.getFile() + " loaded");
    }

    /**
     * Laedt Klassen aus dem eigenen {@link ClassLoader}, der
     * auch externe Jars aus dem <code>osm2po-plugins</code>
     * Verzeichnis beruecksichtigt.
     * @param className String Vollstaendiger Name der Klasse
     * @return Initialisierte {@link Class}
     * @throws ClassNotFoundException Klasse existiert nicht
     */
    public static Class<?> loadClass(String className) throws ClassNotFoundException {
        // true: ruft static{}-Initialisierer der Klassen.
        //       Wichtig bei Aufruf von VarTypeDesk.registerType() aus Plugins heraus!
        return Class.forName(className, true, classLoader);
    }

    public static InputStream getResourceAsStream(String resourcePath) {
        // PITFALL: Wenn keine Datei, dann wird hier DIR ausgelesen!
        return classLoader.getResourceAsStream(resourcePath);
    }

    private static boolean typesRegistered = false;
    public static void registerTypes() {
        // Primitive Typen in VarTypeDesk laufen von -128 aufsteigend
        // Positive Typen sind fuer custom reserviert.
        // -1 bis -10 sind fuer interne Erweiterungen gedacht.

        // ACHTUNG! Kein forName(), *.class landet im RootClassLoader!
        if (typesRegistered) return;
        registerType((byte) -11, LatLon.class);
        registerType((byte) -12, OsmNode.class);
        registerType((byte) -13, OsmWay.class);
        registerType((byte) -14, OsmRelation.class);
        registerType((byte) -15, Node.class);
        registerType((byte) -16, Way.class);
        registerType((byte) -17, Vertex.class);
        registerType((byte) -18, SegmentedWay.class);
        registerType((byte) -19, LatLons.class);
        registerType((byte) -20, Edge.class);
        registerType((byte) -21, RoutingResultSegment.class);
        registerType((byte) -22, Tile.class);
        registerType((byte) -23, TileManager.class);
        registerType((byte) -24, OverlayNode.class);
        registerType((byte) -25, OverlayEdge.class);
        registerType((byte) -26, Membership.class);
        registerType((byte) -27, OsmPolygon.class);
        registerType((byte) -28, Restriction.class);
        registerType((byte) -29, MlgExtension.class);
        registerType((byte) -30, SndExtension.class);
        registerType((byte) -31, BndExtension.class);
        registerType((byte) -32, PoiExtension.class);
        registerType((byte) -33, DbExtension.class);
        registerType((byte) -34, QuickNet.class);
        typesRegistered = true;
    };

    @Override
    public final Iterator<String> iterator() {
        return this.props.iterator();
    }

    /**
     * Ruft {@link #Config(String[], String...)} mit <code>null</code>
     * und ...args. Somit implizit auch No-Args-CTOR.
     * @param args VarArgs-Array of String
     */
    public Config(String... args) {
        this(null, args);
    }

    /**
     * Konstruktor.
     * @param args1 String[] Array of String
     * @param args2 String... Argumente als optionale VarArgs.
     */
    public Config(String[] args1, String... args2) {
        try {
            final String[] args = stringArrayConcat(args1, args2);

            this.log = Log.stderr(Level.WARN).stashMode(true);
            this.props = new Props();

            this.initMemberDefaults();
            this.readSelfRegistriesIntoProperties();

            Properties cliProps = mainArgsToProps(args);
            this.props.putAll(cliProps); // Alles mit CLI-Props ueberschreiben

            this.readConfigFileIntoProperties(); // config= koennte jetzt abweichend gesetzt sein
            this.props.putAll(cliProps); // Nochmal ueberschreiben, wirkt auf .case.xxx

            // case abarbeiten - kann ueberall herkommen!
            this.props.setProperty("case", this.getProperty("case", "default"));
            String[] cases = this.getProperty("case").split("-"); // Multi-Case-Seperator
            for (int i = 0; i < cases.length; i++) cases[i] = cases[i].trim();
            this.props = this.props.expandKeys();
            this.props = this.props.copyCase(cases);

            // Die folgenden Geschichten erst nach Case-Filter!
            this.callCustomConfigurer();

            // Override cliProps, die Dritte - denn case.xxx hier aufgeloest.
            // Zudem muss auch der CustomConfigurer ueberschreibbar sein
            this.props.putAll(cliProps); // TODO .case.xx set in Clis appear again, but dont hurt

            // coreProps immer - nach all dem Custom-Zeugs - wichtig auch fuer UnitTest
            this.addCoreProps();

            this.overrideSimpleMembersFromProperties();
            this.overrideFilesAndDirsFromProperties();
            this.overrideClassesFromProperties(); // uses ClassLoader

            // Caution: Param graph.file can already be set via param/config here
            final File graphFile = this.findGraphFile(false);
            //e.g. .run.service.filter.3.keystore=[graph.dir]/apikeys.txt
            //e.g. .run.service.filter.3.backup=[graph.dir]/apikeys.backup
            if (graphFile != null) {
                this.props.setProperty("graph.file", graphFile.getAbsolutePath());
                this.props.setProperty("graph.dir", graphFile.getParentFile().getAbsolutePath());
            }

            this.overrideLoggersFromProperties(); // uses ClassLoader

            this.registerVarTypes();

            OtelAgentWrapper.init(this);

            this.beanMap = this.loadBeans("loadbean.");
            this.log.stashFlush().stashMode(false);

        } catch (Throwable t) {
            // Wenn hier angekommen, dann kann main() nix mehr loggen.
            this.log.stashFlush().stashMode(false).fatal(t);
            this.log.closeAllWriters();
            System.exit(-1);
        }
    }

    @Override
    public void shutDown() {
        this.close();
    }

    public final void close() {
        this.log.info("Closing Config at " + TS());
        this.clearBeans();
        this.log.closeAllWriters();
    }

    @Override
    public final String toString() {
        return this.props.toString();
    }

    public final String getRawProperty(String key) {return this.props.getRawProperty(key);}
    public final String getProperty(String key) {return this.props.getProperty(key);}
    public final Props getProps() {return this.props;}
    public final Log getLog() {return this.log;}
    public final String getProperty(String key, String defaultValue) {return this.props.getProperty(key, defaultValue);}
    public final List<File> getFiles() {return this.files;}
    public final List<File> getDirs() {return this.dirs;}
    public final URL getSourceFile() {return this.sourceFile;}
    public final String getPrefix() {return this.prefix;}
    public final String getCommand() {return this.command;}
    public final long getReservedXmx() {return this.reservedXmx;}
    public final Map<Integer, Class<? extends AbstractRouter<?>>> getRouterClasses() {return this.routerClassMap;}
    public final Class<? extends WayTagResolver> getWayTagResolverClass() {return this.wayTagResolverClass;}
    public final Class<? extends NodeTagResolver> getNodeTagResolverClass() {return this.nodeTagResolverClass;}
    public final Class<? extends RelationTagResolver> getRelationTagResolverClass() {return this.relationTagResolverClass;}
    public final Class<? extends GraphBuildInterceptor> getGraphBuildInterceptorClass() {return this.graphBuildInterceptorClass;}
    public final Class<? extends JoinFilter> getJoinFilterClass() {return this.joinFilterClass;}
    public final Map<Integer, Class<? extends PostProcessor>> getPostProcessorClassMap() {return this.postProcessorClassMap;}

    /**
     * Wandelt main(args) in Properties um.
     * @param args String[] main-Argumente NULLABLE.
     * @return {@link Properties} NOT NULL.
     * <br><br><b>Wichtig:</b><br>
     * Es gelten folgende Besonderheiten:
     * <ul>
     * <li>Parameter sind in der Form key=value anzugeben.</li>
     * <li>keys ohne = oder =value werden zu key=true</li>
     * <li>Wildcards *.* werden nach "dirs" und "files" als CSV geschrieben</li>
     * <li>Wurden "dirs" oder "files" bereits gesetzt, so werden Wildcards ignoriert</li>
     * </ul>
     */
    public static Properties mainArgsToProps(String[] args) {
        Properties props = new Properties();
        String files = null, dirs = null;
        if (args != null) for (int i = 0; i < args.length; i++) {
            String key = args[i].trim();
            File file = new File(key);
            if (file.exists()) {
                if (file.isFile()) {
                    if (null == files) files = ""; else files += ",";
                    files += file.getAbsolutePath();
                } else {
                    if (null == dirs) dirs = ""; else dirs += ",";
                    dirs += file.getAbsolutePath();
                }
            } else { // no local file
                String val = "true";
                if (key.startsWith("http://") || key.startsWith("https://")) {
                    val = key;
                    key = "remoteFile";
                } else {
                    int equalsSignPos = key.indexOf('=');
                    if (equalsSignPos > 0) {
                        val = key.substring(equalsSignPos + 1);
                        key = key.substring(0, equalsSignPos);
                    }
                }
                props.setProperty(key.trim(), val.trim());
            }
        }
        if (files != null) {
            props.setProperty("files", files);
        }
        if (dirs != null) {
            props.setProperty("dirs", dirs);
        }
        return props;
    }

    /** (Re-)Initialisiert die Standardwerte. */
    private void initMemberDefaults() {
        this.nodeTagResolverClass = DefaultNodeTagResolver.class;
        this.relationTagResolverClass = DefaultRelationTagResolver.class;
        this.wayTagResolverClass = DefaultWayTagResolver.class;
        this.graphBuildInterceptorClass = DefaultGraphBuildInterceptor.class;
        this.joinFilterClass = DefaultJoinFilter.class;
        this.osmParserClassMap.put("osm", OsmXmlParser.class);
        this.prefix = "osm";
        this.workDir = new File(this.prefix);
        this.reservedXmx = 0x2000000; // 32 MB
        this.command = null;
        this.routerClassMap.put(0, VertexRouter.class);
        this.routerClassMap.put(1, OverlayVertexRouter.class);
    }

    /**
     * Findet alle Jars in den Lib-Verzeichnissen.
     * Durchsucht wird osm2po-plugins, osm2po-lib und das
     * {@link ProtectionDomain}-Verzeichnis.
     * Befindet sich ein Jar im Jar, so wird dieses entpackt.
     * @return Array von {@link URL} NOT NULL
     * @throws Exception beliebiger Natur
     */
    private static URL[] findPluginJars() {
        // ClassLoader fuer JARs aus'm lib-Verzeichnis bauen.
        Vector<URL> urls = new Vector<URL>();
        File[] pluginDirs = new File[] {
            findFileOrDir("osm2po-plugins", false),
            findFileOrDir("osm2po-lib", false),
            getJarDir()
        };

        File coreJar = getJarFile();
        final String coreJarName = null == coreJar ? null  : coreJar.getName();

        for (int p = 0; p < pluginDirs.length; p++) {
            if (null == pluginDirs[p]) continue;
            File[] libDirFiles = pluginDirs[p].listFiles();
            try {
                for (int i = 0; i < libDirFiles.length; i++) {
                    File file = libDirFiles[i];
                    if (file.isFile() && file.getName().endsWith(".jar")) {
                        if (file.getName().equalsIgnoreCase(coreJarName)) {
                            continue; // dont load self if started from jar
                        }
                        List<File> jarFiles = extractJarsInJar(file);
                        jarFiles.add(0, file); // Haupt-Jar vorne einfuegen. Immer!
                        for (File jarFile : jarFiles) {
                            urls.add(jarFile.toURI().toURL());
                        }
                    }
                }
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        URL[] urlArr = new URL[urls.size()];
        for (int i = 0; i < urls.size(); i++) urlArr[i] = urls.get(i);
        return urlArr;
    }

    /**
     * Extrahiert Jars aus einem Jar in physikalische Dateien.
     * Als Zielverzeichnis wird {@link File#getParentFile()} verwendet.
     * @param jarFile {@link File}
     * @return {@link List} of {@link File} von extrahierten Jars.
     * @throws Exception beliebiger Natur
     */
    private static List<File> extractJarsInJar(File jarFile) throws Exception {
        List<File> jarList = new ArrayList<File>();
        ZipInputStream zipIs = new ZipInputStream(new FileInputStream((jarFile)));
        for (ZipEntry entry = zipIs.getNextEntry(); entry != null; entry = zipIs.getNextEntry()) {
            String name = entry.getName();
            if (name.endsWith(".jar")) {
                name = name.substring(name.lastIndexOf("/") + 1);
                File extJarFile = new File(jarFile.getParentFile(), name);
                if (!extJarFile.exists()) {
                    byte[] buf = new byte[0x1000];
                    FileOutputStream fos = new FileOutputStream(extJarFile);
                    for (int n = zipIs.read(buf); n != -1; n = zipIs.read(buf)) {
                        fos.write(buf, 0, n);
                    }
                    fos.close();
                    jarList.add(extJarFile);
                }
            }
        }
        zipIs.close();
        return jarList;
    }

    /**
     * Sucht im Classpath-Root nach allen <code>osm2po.selfreg</code>-Dateien
     * und ueberschreibt die internen Properties mit den gefundenen
     * Werten. Achtung: Die Reihenfolge ist beliebig.
     */
    private void readSelfRegistriesIntoProperties() {
        final String REG = "osm2po.selfreg";
        try {
            Enumeration<URL> resources = classLoader.getResources(REG);
            while (resources.hasMoreElements()) {
                URL url = resources.nextElement();
                InputStream is = url.openStream();
                Props regProps = new Props();
                regProps.load(is);
                is.close();
                this.log.debug("Registered Properties in\n" + url + ":\n" + regProps);
                this.props.putAll(regProps);
            }
        } catch (Throwable t) {
            this.log.error(t);
        }
    }

    /**
     * Ueberschreibt die internen Properties final mit Werten aus einer
     * Konfigurations-Datei.
     * @throws Throwable
     */
    private void readConfigFileIntoProperties() throws Throwable {
        final String CONFIG = "osm2po.config";
        URL url = null;

        String extConf = this.getProperty(
                "config." + this.getProperty("case"), this.getProperty("config"));

        // Prio1: External config set via Argument oder Property
        if (extConf != null) {
            if ("".equals(extConf.trim())) {
                log.warn("Config file disabled");
                return;
            }
            try {
                url = new URI(extConf).toURL(); // Auch http moeglich
            } catch (Exception e) {
                File file = new File(extConf);
                if (file.exists()) {
                    url = file.toURI().toURL();
                } else {
                    File localFile = findFileOrDir(extConf, true);
                    if (localFile != null) {
                        url = localFile.toURI().toURL();
                    } else {
                        url = classLoader.getResource(extConf);
                    }
                }
            }

            if (null == url) {
                this.log.error("config=" + extConf + " not found");
                return;
            }
        }

        // Prio2: Default osm2po.config in catalina.base/conf if tomcat
        if (null == url) {
            String catalina = System.getProperty("catalina.base");
            if (catalina != null) {
                File catalinaConf = new File(catalina + "/conf/" + CONFIG);
                if (catalinaConf.exists()) url = catalinaConf.toURI().toURL();
            }
        }

        // Prio3: Default osm2po.config in current or jar directory
        if (null == url) {
            File localFile = findFileOrDir(CONFIG, true);
            if (localFile != null) url = localFile.toURI().toURL();
        }

        // Prio4: Config from Root-Classpath even in Plugins
        if (null == url) {
            url = classLoader.getResource(CONFIG);
        }

        // Prio5: Fallback to internal defaults
        if (null == url) {
            url = Config.class.getResource(CONFIG);
        }

        if (null == url) {
            this.log.error("No configruation file found!");
        } else {
            try {
                this.log.info("Reading Configuration from " + url);
                InputStream is = url.openStream();
                this.props.load(is);
                is.close();
                if ("file".equals(url.getProtocol())) {
                    File file = new File(url.getPath());
                    this.props.setProperty("config.dir", file.getParent());
                    this.props.setProperty("config.file", file.getName());
                }

            } catch (Throwable t) {
                this.log.fatal(t);
            }
        }
    }

    private void callCustomConfigurer() throws Exception {
        Configurer configurer = null;
        String configClass = this.getProperty("configurer.class");
        if (configClass != null) configurer =
            (Configurer) loadClass(configClass).getDeclaredConstructor().newInstance();
        if (configurer != null) {
            log.info("Using additional config from "
                    + configurer.getClass().getSimpleName());
            configurer.configure(this.props);
        }
    }

    @SuppressWarnings("unchecked")
    private void overrideLoggersFromProperties() {
        this.log.closeAllWriters();

        Level logLevel = Level.byName(getProperty("log.level", "debug"));
        boolean logThreadId = Boolean.valueOf(getProperty("log.thread", "false"));
        this.log.setLogLevel(logLevel);
        this.log.setLogThread(logThreadId);

        boolean writerExists = false;
        for (int i = 0; i < 100; i++) {
            String logClassKey = "log." + i + ".class";
            String writerClass = this.props.getProperty(logClassKey);
            if (writerClass != null) {
                writerClass = writerClass.trim();
                try {
                    if ("".equals(writerClass)) {
                        this.log.warn(logClassKey + " disabled");
                    } else {
                        Class<? extends LogWriter> lwClazz =
                            (Class<? extends LogWriter>) loadClass(writerClass);
                        this.log.addLogWriter(lwClazz.getDeclaredConstructor().newInstance().open(this, i));
                        writerExists = true;
                    }
                } catch (Throwable t) {
                    this.log.fatal(t);
                }
            }
        }
        if (!writerExists) {
            this.log.addLogWriter(new LogConsoleWriter().open(this, 0));
            this.log.addLogWriter(new LogFileWriter().open(this, 1));
        }
    }

    private void overrideFilesAndDirsFromProperties() throws Throwable {
        String baseDir;

        if (!this.dirs.isEmpty()) {
            log.error("Config: OverrideDirs called again. Clearing internal dir-List");
            this.dirs.clear();
        }
        if (!this.files.isEmpty()) {
            log.error("Config: OverrideFiles called again. Clearing internal file-List");
            this.files.clear();
        }

        String strDirs = props.getProperty("dirs");
        if (strDirs != null) {
            baseDir = null;
            for (String strDir : strDirs.split(",")) {
                strDir = strDir.trim();
                if (strDir.startsWith("*")) {
                    baseDir = strDir.substring(1).trim();
                    continue;
                }
                File dir = new File(baseDir, strDir);
                if (dir.isDirectory()) this.dirs.add(dir);
                else log.warn(dir.getAbsolutePath() + " not found");
            }
        }

        String strFiles = props.getProperty("files");
        if (strFiles != null) {
            baseDir = null;
            for (String strFile : strFiles.split(",")) {
                strFile = strFile.trim();
                if (strFile.startsWith("*")) {
                    baseDir = strFile.substring(1);
                    continue;
                }
                File file = new File(baseDir, strFile);
                if (file.isFile()) this.files.add(file);
                else log.warn(file.getAbsolutePath() + " not found");
            }
        }

        // seit 12.10.20: Keine Unterscheidung mehr zwischen file und remoteFile
        String sourcePath = this.getProperty("file", this.getProperty("remoteFile"));
        if (sourcePath != null) try {
            this.sourceFile = new URI(sourcePath).toURL();

        } catch (Exception ignore) {
            File localFile = new File(sourcePath);
            if (localFile.isFile()) this.sourceFile = localFile.toURI().toURL();
            else this.sourceFile = classLoader.getResource(sourcePath);

        }
        if (null == sourceFile && this.files.size() > 0) {
            this.sourceFile = this.files.get(0).getAbsoluteFile().toURI().toURL();
            this.files.remove(0);
        }
    }

    private void overrideSimpleMembersFromProperties() throws Throwable {
        String strPrefix = this.props.getProperty("prefix");
        if (strPrefix != null) this.prefix = strPrefix.trim();
        else this.props.setProperty("prefix", this.prefix);

        String strDir = this.props.getProperty("workDir", this.prefix);
        this.workDir = new File(strDir.trim());
        this.props.setProperty("workDir", workDir.getAbsolutePath()); // write back if null

        String strCommand = this.props.getProperty("cmd");
        if (strCommand != null) this.command = strCommand.trim();

        String strReservedXmx = this.props.getProperty("reservedXmx");
        if (strReservedXmx != null)
            this.reservedXmx = Long.valueOf(strReservedXmx).longValue() * 0x100000; // mal 1Mb

    }

    /**
     * Liefert das Arbeitsverzeichnis.
     * Existiert dieses noch nicht, so wird es hier lazy angelegt.
     * @return File workdir
     */
    public File getWorkDir() {
        // Lazy!
        if (!this.workDir.exists()) {
            this.workDir.mkdirs();
        }
        return this.workDir;
    }

    /**
     * Liefert einen Parser fuer eine Datei (-Extension)
     * @param uri Pfad
     * @return Class des {@link OsmParser}, Fallback: {@link OsmXmlParser}-Class.
     */
    public Class<? extends OsmParser> getOsmParserClass(String uri) {
        Class<? extends OsmParser> osmParserClass = OsmXmlParser.class;
        for (Entry<String, Class<? extends OsmParser>> entry : this.osmParserClassMap.entrySet()) {
            String key = entry.getKey();
            if (uri.endsWith(key)) return entry.getValue();
        }
        return osmParserClass;
    }

    private void addCoreProps() {
        Date now = new Date();

        DateFormat fd = new SimpleDateFormat("yyyy-MM-dd");
        DateFormat ft = new SimpleDateFormat("HH:mm");

        this.props.setProperty("core.timestamp", "" + System.currentTimeMillis());
        this.props.setProperty("core.date", fd.format(now));
        this.props.setProperty("core.time", ft.format(now));
        this.props.setProperty("core.version", Version.getVersion());

        // Tomcat BUG: ProtectionDomain may be NULL
        // see: https://bz.apache.org/bugzilla/show_bug.cgi?id=58096
        File domainDir = getJarDir();
        String coreDomain = null == domainDir ? "" : domainDir.getAbsolutePath();
        this.props.setProperty("core.domain", coreDomain );
    }

    private void registerVarTypes() {
        for (int i = -128; i <= 127; i++) {
            String key = "varType." + i;
            String val = this.props.getProperty(key);
            if (val != null) {
                try {
                    registerType((byte)i, val, classLoader);
                } catch (Throwable t) {
                    log.error(t);
                }
            }
        }
    }

    /**
     * Versucht im aktuellen Kontext eine lokale Graphdatei zu finden.
     * @param mustExist boolean, Datei muss bereits existieren ja/nein.
     * @return File graphFile oder NULL, wenn nicht gefunden und mustExist gesetzt.
     */
    public File findGraphFile(boolean mustExist) {
        File graphFile = null;

        // Prio 1 per param/config gesetzt
        String configGraphFilePath = this.getProperty("graph.file");
        if (configGraphFilePath != null) {
            graphFile = new File(configGraphFilePath);

        } else {
            // File-Param bzw. ohne Param als Commandline-Argument
            if (this.sourceFile != null && "file".equals(sourceFile.getProtocol())) {
                graphFile = new File(this.sourceFile.getPath());
                if (!Graph.isGraphFile(graphFile)) graphFile = null;
           }
           // ggf. auch unsortiert in einer der *.*-Dateien. Dann heraussuchen
           if (null == graphFile) for (File file : this.getFiles()) {
               if (file.getName().endsWith(".gph") && Graph.isGraphFile(file)) {
                   graphFile = file;
                   break;
               }
           }
        }

        // Last Fallback to Default
        if (null == graphFile) {
            graphFile = new File(this.getWorkDir(), this.getPrefix() + "_2po.gph");
        }

        if (mustExist && !Graph.isGraphFile(graphFile)) {
            graphFile = null;
        }

        return graphFile;
    }


    public boolean putBean(final String name, VarBean bean) {
        if (this.beanMap.containsKey(name)) return false;
        this.beanMap.put(name, bean);
        return true;
    }

    public VarBean getBean(final String name) {
        return this.beanMap.get(name);
    }

    public void deleteBean(final String name) {
        VarBean varBean = this.beanMap.remove(name);
        if (varBean != null) varBean.clear();
    }

    public void clearBeans() {
        for (Entry<String, VarBean> entry : this.beanMap.entrySet()) {
            VarBean bean = entry.getValue();
            if (bean != null) bean.clear();
        }
    }

    /**
     * Laedt als Dateien serialisierte {@link Var}s und {@link VarBean}s
     * aus der Konfiguration in eine Map von {@link VarBean}s.
     * Standardmaessig wird diese Methode bei der Initialisierung mit
     * <code>loadbean.</code> gerufen. Eine Konfiguration koennte folglich
     * so aussehen:<br>
     * <code>loadbean.mybean=path/to/varbean.file</code> oder<br>
     * <code>loadbean.mybean.33=path/to/var.file</code><br>
     * Auf die Objekte kann danach mittels {@link #getBean(String)}
     * unter dem Namen "mybean" zugegriffen werden.<br>
     * @param prefixWithDot Prefix der property
     * @return {@link Map} of {@link VarBean} NOT NULL
     */
    public final Map<String, VarBean> loadBeans(final String prefixWithDot) {
        Map<String, VarBean> map = new HashMap<String, VarBean>();
        for (String key : this) {
            if (key.startsWith(prefixWithDot)) {
                String payload = key.substring(prefixWithDot.length()); // Prefix entfernen
                String[] parts = payload.split("\\.");
                String name = parts[0].trim();
                String type = parts.length > 1 ? parts[1].trim() : null;

                File file = new File(getProperty(key));
                if (file.isFile()) {
                    InStream inStream = new InStreamDisk(file);
                    VarBean varBean = null;
                    if (null == type) {
                        varBean = new VarBean().readFromStream(inStream);
                    } else {
                        Var<?> var = VarTypeDesk.newInstanceOf(Byte.valueOf(type));
                        var.readFromStream(inStream);
                        varBean = new VarBean(var);
                    }
                    inStream.close();
                    if (map.containsKey(name)) {
                        log.info("Bean '" + name + "' already exists");
                    } else {
                        map.put(name, varBean);
                        log.info("Bean '" + name + "' loaded from\n" + file.getPath());
                    }
                }
            }
        }
        return map;
    }

    @SuppressWarnings("unchecked")
    private void overrideClassesFromProperties() throws Throwable {

        String strWayTagResolverClass = this.props.getProperty("wayTagResolver.class");
        if (strWayTagResolverClass != null) this.wayTagResolverClass =
            (Class<? extends WayTagResolver>) loadClass(strWayTagResolverClass.trim());

        String strNodeTagResolverClass = this.props.getProperty("nodeTagResolver.class");
        if (strNodeTagResolverClass != null) this.nodeTagResolverClass =
            (Class<? extends NodeTagResolver>) loadClass(strNodeTagResolverClass.trim());

        String strRelationTagResolverClass = this.props.getProperty("relationTagResolver.class");
        if (strRelationTagResolverClass != null) this.relationTagResolverClass =
            (Class<? extends RelationTagResolver>) loadClass(strRelationTagResolverClass.trim());

        String strGraphBuildInterceptorClass =
            this.props.getProperty("graph.build.interceptor.class");
        if (strGraphBuildInterceptorClass != null) this.graphBuildInterceptorClass =
            (Class<? extends GraphBuildInterceptor>) loadClass(strGraphBuildInterceptorClass.trim());

        String strJoinFilterClass =
            this.props.getProperty(JoinFilter.PROP_PREFIX + "class");
        if (strJoinFilterClass != null) this.joinFilterClass =
            (Class<? extends JoinFilter>) loadClass(strJoinFilterClass.trim());

        for (int i = -128; i <= 127; i++) {
            String key, val;
            key = POSTP(i) + ".class";
            val = this.props.getProperty(key);
            if (val != null) {
                try {
                    // Hier hat Java ein Problem:
                    // Wenn kein PostProcessor, dann kein Fehler!
                    Class<? extends PostProcessor> clazz =
                        (Class<? extends PostProcessor>) loadClass(val.trim());
                    this.postProcessorClassMap.put(i, clazz);
                } catch (Exception e) {
                    this.log.warn(val + " not found");
                }
            }

            key = "router." + i + ".class";
            val = this.props.getProperty(key);
            if (val != null) {
                try {
                    Class<? extends AbstractRouter<?>> clazz =
                        (Class<? extends AbstractRouter<?>>) loadClass(val.trim());
                    this.routerClassMap.put(i, clazz);
                } catch (Exception e) {
                    this.log.warn(val + " not found");
                }
            }

        }

        for (Entry<Object, Object> entry : props.entrySet()) {
            String key = (String) entry.getKey();
            String val = (String) entry.getValue();
            final String prefix = "osmParser.";
            if (key.startsWith(prefix)) {
                String extension = key.substring(prefix.length()).trim().toLowerCase();
                try {
                    Class<? extends OsmParser> clazz =
                        (Class<? extends OsmParser>) loadClass(val.trim());
                    this.osmParserClassMap.put(extension, clazz);
                } catch (Exception e) {
                    this.log.warn(val + " not found");
                }
            }
        }

    }

    /**
     * Oeffnet eine Datenbank-{@link Connection}.<br>
     * Folgende Verbindungsparameter werden aus den {@link Props}
     * der {@link Config} gezogen.<br>
     * Defaults in Klammern:
     * <ul>
     * <li><code>db.driver</code> (org.postgresql.Driver)</li>
     * <li><code>db.connect</code> (jdbc:postgresql://localhost:5432/osm2po)</li>
     * <li><code>db.property.user</code> (postgres)</li>
     * <li><code>db.property.password</code> (postgres)</li>
     * </ul>
     * Weitere Parameter koennen ebenfalls mittels <code>db.property.*</code>
     * uebergeben werden.
     *
     * @param prefix String NULLABLE, Default ist 'db' ohne Punkt, siehe oben.
     * @return {@link Connection}
     * @throws Exception beliebiger Natur
     */
    public final Connection openDbConnection(String prefix) throws Exception {
        String pfx = null == prefix ? "db" : prefix;
        String drv = getProperty(pfx +
                ".driver", "org.postgresql.Driver");
        String con = getProperty(pfx +
                ".connect", "jdbc:postgresql://localhost:5432/osm2po");
        Properties dbprops = new Properties();
        dbprops.setProperty("user", "postgres");
        dbprops.setProperty("password", "postgres");

        final String propPre = pfx + ".property.";
        final int propPreLen = propPre.length();
        for (String key : this) {
            if (key.startsWith(propPre)) {
                String val = getProperty(key);
                String prop = key.substring(propPreLen);
                dbprops.setProperty(prop, val);
            }
        }

        // Driver nicht ueber DriverManager laden, da wir
        // einen eigenen ClassLoader verwenden.
        Class<?> driverClazz = loadClass(drv);
        Driver driver = (Driver) driverClazz.getDeclaredConstructor().newInstance();

        return driver.connect(con, dbprops);
    }

    /**
     * Handler und gleichzeitig Factory, um fuer {@link URL}
     * auch das <code>classpath:</code>-Protokoll zuzulassen.
     * @author Carsten
     */
    private static class ClassPathStreamHandlerAndFactory
    extends URLStreamHandler implements URLStreamHandlerFactory {
        @Override
        public URLStreamHandler createURLStreamHandler(String protocol) {
            return ("classpath".equals(protocol)) ? this : null; // this intance or new?
        }
        @Override
        protected URLConnection openConnection(URL url) throws IOException {
            return new ClassPathURLConnection(url);
        }
        private static class ClassPathURLConnection extends URLConnection {
            public ClassPathURLConnection(URL url) {super(url);}
            @Override
            public void connect() throws IOException {}
            @Override
            public InputStream getInputStream() throws IOException {
                InputStream result = classLoader.getResourceAsStream(url.getFile());
                if (null == result) throw new IOException(
                        "Could not open InputStream for URL '" + this.url + "'");
                return result;
            }
        }
    }

}