/*
 * $Id: Main.java 5981 2024-10-29 16:34:16Z carsten $
 *
 * Copyright 2012-2024 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po;

import static de.cm.osm2po.misc.ArrayUtils.stringArrayConcat;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.Utils.beep;
import static de.cm.osm2po.misc.Utils.getJavaInfo;
import static de.cm.osm2po.misc.Utils.getJavaSpecVersion;
import static de.cm.osm2po.misc.Utils.openBrowser;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.URL;
import java.util.Date;
import java.util.Map.Entry;

import de.cm.osm2po.converter.Compressor;
import de.cm.osm2po.converter.GraphBuilder;
import de.cm.osm2po.converter.Joiner;
import de.cm.osm2po.converter.Merger;
import de.cm.osm2po.converter.Processor;
import de.cm.osm2po.converter.Segmenter;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.converter.parser.OsmParser;
import de.cm.osm2po.converter.parser.OsmParserHandler;
import de.cm.osm2po.converter.parser.OsmXmlParser;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.Debug;
import de.cm.osm2po.misc.Threader;
import de.cm.osm2po.misc.Threader.Executor;
import de.cm.osm2po.postp.PostProcessor;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.RateLimitServer;

public class Main {

    private static long[] logCounters = new long[Level.values().length];
    private static void cumulateLogCounters(Log log) {
        long[] lc = log.getCounters();
        for (int i = 0; i < logCounters.length; i++) {
            logCounters[i] += lc[i];
        }
    }

    public static void main(final String[] args) { // Wir nur von Commandline gerufen
        // Explizit args2 null setzen, sonst Self-Call und DeadLoop
        main(args, (String[]) null);
    }

    public static void main(String[] args1, String... args2) {
        final String[] args = stringArrayConcat(args1, args2);

        printLicenceHeader(System.err);
        System.err.println();

        // Hilfe anzeigen, wenn Intuitiv-Parameter eingegeben wurde
        for (int i = 0; i < args.length; i++) {
            String arg = args[i].toLowerCase();
            // --help durchlassen, schleswig-holstein nicht.
            if (arg.length() <= 6
                    && (arg.contains("-h") || arg.equals("help"))) {
                printHelp(System.out);
                return;
            }
        }

        final Config config = new Config(args);
        Config.logInfoJarUrls(config.getLog());

        String strJobs = config.getProperty("jobs");
        if (null == strJobs) {
            run(config);
        } else {
            Threader threader = new Threader(strJobs, new Executor() {
                @Override public void execute(String command) {
                    Main.run(new Config(args, "case=" + command));
                }
            });

            Log log = config.getLog();
            log.info("Executing " + threader + " at " + new Date());
            threader.run();
            log.logWarnCounters(logCounters); // static global
            config.close();
        }
    }

    private static void run(Config config) {
        final Log log = config.getLog();
        final ShutdownHookSequence onShutdown = new ShutdownHookSequence(log);

        // Schliesst Dateien und Threads, z.B. Logger
        onShutdown.addLast(config);

        try {
            // Kommando analysieren
            String cmd = config.getCommand();
            if (null == cmd) cmd = "tjsgpr";

            if (cmd.indexOf('c') != -1) {
            	cmd = cmd.replace("c", "");
            	cmd += "tjsp";
            }

            if ("d".equals(cmd)) { // Debugger-Dumper - Ohne Logger!
                Debug.dump2poFile(config);
                return;
            }

            System.gc();

            log.info("Running osm2po " + Version.getVersion()
                    + " with cmd=" + cmd + " - " + MB());

            log.info(getJavaInfo());
            double javaSpecVersion = getJavaSpecVersion();
            if (javaSpecVersion < 1.7) log.warn("JAVA " + javaSpecVersion
                    + " detected but some features require 1.7 or higher");

            String caseInfo = config.getProperty("caseInfo");
            if (caseInfo != null) log.note(caseInfo);

            log.debug("Final Properties are:\n" + config.getProps());

            if (cmd.indexOf('p') != -1) {
                if (config.getPostProcessorClassMap().isEmpty()) {
                    log.warn("cmd 'p' set but no postprocessors injected");
                }
            }
            if (cmd.indexOf('m') != -1) {
                if (cmd.contains("j")) throw new RuntimeException(
                    "cmd=m (Merger) cannot be combined with cmd=j (Joiner)");
            }

            if (cmd.indexOf('p') != -1) runPostProcessors(config, true);
            if (cmd.indexOf('t') != -1) runTiler(config);
            if (cmd.indexOf('z') != -1) runProcessor(config, new Compressor());
            if (cmd.indexOf('m') != -1) runProcessor(config, new Merger());
            if (cmd.indexOf('j') != -1) runProcessor(config, new Joiner());
            if (cmd.indexOf('s') != -1) runProcessor(config, new Segmenter());
            if (cmd.indexOf('g') != -1) runProcessor(config, new GraphBuilder());
            if (cmd.indexOf('p') != -1) runPostProcessors(config, false);

            log.logWarnCounters();
            cumulateLogCounters(log);

            if (cmd.indexOf('l') != -1) {
                log.info("Starting " + RateLimitServer.class.getSimpleName()
                        + " at " + new Date().toString());
                try {
                    RateLimitServer rateLimitServer = new RateLimitServer(config);
                    onShutdown.addFirst(rateLimitServer);
                } catch (Exception e) {
                    log.fatal(RateLimitServer.class.getSimpleName() + " not started.");
                    log.info("Calling System.exit(-1)");
                    System.exit(-1);
                }
            }

            if (cmd.indexOf('r') != -1) {
                log.info("Starting " + Osm2poService.class.getSimpleName()
                        + " at " + new Date().toString());
                Osm2poService osm2poService = new Osm2poService();
                if (osm2poService.startService(config)) {
                    if (Boolean.valueOf(config.getProperty("desktop", "false"))) {
                        beep(1);
                        openBrowser(osm2poService.getHttpUrl());
                    }
                    onShutdown.addFirst(osm2poService);
                } else {
                    log.info("Calling System.exit(-1)");
                    System.exit(-1);
                }
            }

        } catch (Throwable t) {
            if (log != null) {
                log.fatal(t);
                cumulateLogCounters(log);
                log.logWarnCounters(logCounters);
            } else {
                t.printStackTrace();
            }
            System.exit(-1); // Da sind evtl. noch mehr jobs in der Schleife
        }
    }

    @SuppressWarnings("unchecked")
    private static void runTiler(Config config) throws Throwable {
        Log log = config.getLog();
        log.info("Starting Tiler at " + new Date().toString());

        Class<? extends OsmParser> parserClass = null;
        InputStream osmIn = null;

        // Parser explizit gesetzt?
        // Nicht 'osmParser' verwendet, da Prefix fuer Suffix-Erkennung!
        String parserClassName = config.getProperty("parser.class");
        if (parserClassName != null) {
            Class<?> clazz = Config.loadClass(parserClassName);
            parserClass = (Class<? extends OsmParser>) clazz;
        }

        URL sourceFile = config.getSourceFile();
        if (null == sourceFile) {
            if (null == parserClass) { // Kann Parser ohne InputStream sein
                parserClass = OsmXmlParser.class;
                log.info("Reading XML from STDIN");
                osmIn = System.in;
            }

        } else { // SourceFile given as URL
            log.info("Reading from " + sourceFile);
            if (null == parserClass) {
                parserClass = config.getOsmParserClass(sourceFile.getFile());
            }

            if ("file".equals(sourceFile.getProtocol())) {
                File localFile = new File(sourceFile.getPath());
                log.info("File last modified at " + new Date(localFile.lastModified()));

                osmIn = new BufferedInputStream(
                        new FileInputStream(localFile), 0x10000);
                if (null == parserClass)
                    parserClass = config.getOsmParserClass(localFile.getName());

            } else { // http, ftp and other
                osmIn = sourceFile.openStream();
            }
        }

        log.info("Using parser " + parserClass.getName());
        OsmParser parser = parserClass.getDeclaredConstructor().newInstance();

        String handlerClass = config.getProperty(
                "parser.handler.class", Tiler.class.getName());
        log.debug("With handler " + handlerClass);
        OsmParserHandler handler =
            (OsmParserHandler) Config.loadClass(handlerClass).getDeclaredConstructor().newInstance();

        handler.open(config);
        parser.open(osmIn, handler, config);
        parser.parse();
        parser.close();
        handler.close();

        log.info("Tiler finished at " + new Date().toString());
    }

    private static void runProcessor(Config config, Processor processor) throws Throwable {
        Log log = config.getLog();
        final String processorName = processor.getClass().getSimpleName();
        log.info("Starting " + processorName + " at " + new Date().toString());

        processor.init(config);
        processor.run();
        processor.close();

        log.info(processorName +  " finished at " + new Date().toString());
    }

    private static void runPostProcessors(Config config, boolean asPreProcessor) {
        Log log = config.getLog();
        for (Entry<Integer, Class<? extends PostProcessor>> postpEntry :
            config.getPostProcessorClassMap().entrySet()) {

            // negative IDs sind PreProcessors!
            if (asPreProcessor ^ postpEntry.getKey() < 0) {
                continue;
            }

            log.info("Starting PostProccessor[" +  postpEntry.getKey()+ "] at "
                    + new Date() + "\n" + postpEntry.getValue().getName());
            Class<? extends PostProcessor> ppClass = postpEntry.getValue();
            try {
                PostProcessor pp = ppClass.getDeclaredConstructor().newInstance();
                pp.run(config, postpEntry.getKey());
                log.info("PostProcessor finished at " + new Date().toString());

            } catch (Throwable t) {
                if (t instanceof InstantiationException) {
                    log.error("Could not instantiate " + ppClass.getName());
                } else {
                    log.error(t);
                }

                String key = Config.POSTP(postpEntry.getKey()) + ".failOnError";
                boolean failOnError = Boolean.valueOf(
                        config.getProperty(key, "true"));
                if (failOnError) {
                    throw new RuntimeException(t);
                } else {
                    log.info("Tip! Set '" + key + "=true' to exit on this error");
                }
            }
        }
    }

    private static void printHelp(PrintStream ps) {
        try {
            InputStream is = Main.class.getResourceAsStream("osm2po-help.txt");
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);
            String line;
            ps.println();
            while ((line = br.readLine()) != null) ps.println(line);
            br.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static void printLicenceHeader(PrintStream ps) {
        try {
            String prgVersion = Version.getVersion();
            String prgCreated = Version.getCreated();
            String prgYear = prgCreated.substring(0, 4);

            InputStream is = Main.class.getResourceAsStream("header.txt");
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x1000);
            for (int b = is.read(); b != -1; b = is.read()) baos.write(b);
            String header = String.format(
                    new String(baos.toByteArray()), prgVersion, prgYear);
            ps.write(header.getBytes());
            ps.write("\n".getBytes());

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

}
