/*
 * $Id: Log.java 5964 2024-10-26 13:58:54Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;
import static de.cm.osm2po.logging.Log.Level.ALERT;
import static de.cm.osm2po.logging.Log.Level.BUGFIX;
import static de.cm.osm2po.logging.Log.Level.DATA;
import static de.cm.osm2po.logging.Log.Level.DEBUG;
import static de.cm.osm2po.logging.Log.Level.ERROR;
import static de.cm.osm2po.logging.Log.Level.FATAL;
import static de.cm.osm2po.logging.Log.Level.INFO;
import static de.cm.osm2po.logging.Log.Level.MESSAGE;
import static de.cm.osm2po.logging.Log.Level.NOTE;
import static de.cm.osm2po.logging.Log.Level.PROGRESS;
import static de.cm.osm2po.logging.Log.Level.WARN;
import static de.cm.osm2po.logging.LogConsoleWriter.ERR;
import static de.cm.osm2po.logging.LogConsoleWriter.OUT;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * Ein kleiner Logger. Simpel aber gut.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public final class Log implements Iterable<LogWriter> {

    public static enum Level {
        /** Undefiniert, intern verwendet. */
        NULL(0, false, "      "),
        /** Ausgabe von Daten und Dumps. */
        DATA(1, false, "DATA  "),
        /** Korrigiete oder gerettete Daten. */
        BUGFIX(2, true, "FIX   "),
        /** Detaillierte Entwicklungs-Infos. */
        DEBUG(3, false, "DEBUG "),
        /** Verarbeitungsfortschritt anzeigen. */
        PROGRESS(4, false,  "...   "),
        /** Info Zusatz z.B. fuer Ergebnisanzeigen. */
        NOTE(5, false, "NOTE  "),
        /** Warn Info unkritischer Natur. */
        WARN(6, true, "WARN  "),
        /** Wichtige Programminformation. */
        INFO(7, false, "INFO  "),
        /** Kritische Programminfo, ohne Crash. */
        ERROR(8, true, "ERROR "),
        /** Sehr wichtige Info, z.B. fuer externe Systeme oder Mail-Trigger */
        MESSAGE(9, false, "INFO  "),
        /** Alarm Info, z.B. fuer externe Systeme oder Mail-Trigger */
        ALERT(10, true, "ALERT "),
        /** Sehr kritische Programminfo, wahrscheinlich mit Crash. */
        FATAL(11, true, "FATAL ");

        public final int severity;
        public final boolean negative;
        public final String prefix;

        Level(final int severity, boolean negative, final String prefix) {
            this.severity = severity;
            this.negative = negative;
            this.prefix = prefix;
        }

        public static String spacer() {
            return NULL.prefix;
        }

        public static Level bySeverity(int severity) {
            for (Level level : Level.values()) {
                if (severity == level.severity) return level;
            }
            return NULL;
        }

        public static Level byName(final String name) {
            try {
                return Level.valueOf(name.trim().toUpperCase());
            } catch (Exception ignore) {
                return NULL;
            }
        }
    }

    private Level logLevel;
    private boolean logThread;
    private volatile boolean stashMode;

    private List<LogWriter> logWriterList = new ArrayList<LogWriter>();
    private List<String> stashList = new ArrayList<String>();

    private long[] logCounters = new long[Level.values().length];

    /** @return long Array of number of calls per level */
    synchronized public final long[] getCounters() {
        return Arrays.copyOf(this.logCounters, this.logCounters.length);
    }

    /** resets the counter of log calls */
    synchronized public final void resetCounters() {
        Arrays.fill(this.logCounters, 0L);
    }

    /** @return Log ohne {@link LogWriter}. */
    public static Log empty() {
        return new Log();
    }

    /** @return Log mit stdout-Writer Level Warn. */
    public static Log stdout() {
        return stdout(WARN);
    }

    /** @return Log mit stderr-Writer Level Error  */
    public static Log stderr() {
        return stderr(ERROR);
    }

    public static Log stdout(Level level) {
        return new Log(level).addLogWriter(
                new LogConsoleWriter().setPrintTo(OUT).setLogLevel(level));
    }

    public static Log stderr(Level level) {
        return new Log(level).addLogWriter(
                new LogConsoleWriter().setPrintTo(ERR).setLogLevel(level));
    }

    public Log() {
        this.logLevel = INFO;
    }

    public Log(Level level) {
        this.logLevel = level;
    }

    public final Log setLogLevel(Level logLevel) {
    	this.logLevel = logLevel;
    	return this;
    }

    public final Level getLogLevel() {
    	return this.logLevel;
    }

    public final Log setLogThread(boolean logThread) {
        this.logThread = logThread;
        return this;
    }

    public final boolean isLogThread() {
        return this.logThread;
    }

    /**
     * Fuegt einen neuen Logger hinzu.
     * <br><b>Wichtig:</b><br>
     * Steht dieser initial auf {@link Level#NULL}, so
     * erbt er das Level seines Parents.
     * @param logWriter {@link LogWriter}
     * @return this
     */
    public final Log addLogWriter(LogWriter logWriter) {
        this.logWriterList.add(logWriter);
        return this;
    }

    /**
     * Schliesst alle {@link LogWriter} und entfernt sie aus der
     * internen Liste.
     */
    public void closeAllWriters() {
    	if (this.logWriterList != null)
    	    for (LogWriter logWriter : this.logWriterList) {
    	        try {logWriter.close();} catch (Throwable ignore) {};
    	    }
    	this.logWriterList = new ArrayList<LogWriter>();
    }

    @Override
    public final Iterator<LogWriter> iterator() {
        return this.logWriterList.iterator();
    }

    public final void data(String msg) {this.log(msg, DATA, null);}
    public final void bugfix(String msg) {this.log(msg, BUGFIX, null);}
    public final void debug(String msg) {this.log(msg, DEBUG, null);}
    public final void note(String msg) {this.log(msg, NOTE, null);}
    public final void warn(String msg) {this.log(msg, WARN, null);}
    public final void info(String msg) {this.log(msg, INFO, null);}
    public final void error(String msg) {this.log(msg, ERROR, null);}
    public final void fatal(String msg) {this.log(msg, FATAL, null);}
    public final void message(String msg) {this.log(msg, MESSAGE, null);}
    public final void alert(String msg) {this.log(msg, ALERT, null);}
    public final void progress(String msg) {this.log(msg, PROGRESS, null);}

    public final void error(Throwable t) {this.error(translateError(t, 10));}
    public final void fatal(Throwable t) {this.fatal(translateError(t, 10));}

    public final boolean isEnabled(Level logLevel) {
        return logLevel.severity >= this.logLevel.severity;
    }

    public final synchronized Log stash(Level level, final String msg) {
        this.stashList.add(level + "|" + msg);
        return this;
    }

    public final synchronized Log stashMode(boolean on) {
        this.stashMode = on;
        return this;
    }

    public final synchronized Log stashFlush() {
        boolean currentStashMode = this.stashMode;
        this.stashMode = false;
        for (String entry : this.stashList) {
            int idx = entry.indexOf('|');
            final Level level = Level.valueOf(entry.substring(0, idx));
            final String message = entry.substring(idx +1);
            this.log(message, level, null);
        }
        this.stashList.clear();
        this.stashMode = currentStashMode;
        return this;
    }

    public final synchronized void log(String message,
            Level logLevel, LogWriter sender) {

        if (null == this.logWriterList) return;

        if (this.stashMode) { // Wichtig: Immer ohne LevelFilter
            this.stash(logLevel, message); // Merken fuer verzoegerte Ausgabe

        } else {
            this.logCounters[logLevel.ordinal()]++;
            if (logLevel.severity < this.logLevel.severity) return;

            if (this.logThread) {
                Thread ct = Thread.currentThread();
                if (PROGRESS == logLevel) {
                    message = "T" + ct.getId() + " " + message;
                } else {
                    message = "T" + ct.getId() + "-"
                        + ct.getName().toUpperCase() + " " + message;
                }
            }

            List<LogWriter> brokenWriterList = null;
            for (LogWriter lw : this.logWriterList) {
                try {
                    if (lw != sender) lw.log(message, logLevel);

                } catch (Throwable t) {
                    if (null == brokenWriterList) brokenWriterList = new ArrayList<LogWriter>();
                    try {lw.close();} catch (Throwable ignore) {}

                    this.stash(FATAL, translateError(t, 10));
                    this.stash(WARN, lw.getClass().getSimpleName() + " closed and removed");
                    brokenWriterList.add(lw);
                }
            }
            if (brokenWriterList != null) {
                this.logWriterList.removeAll(brokenWriterList);
            }
        }
    }

    public static String translateError(Throwable t, int stackTraceLen) {
        String msg = null;
        String clazz = null;
        String errString = "Unknown Error";

        if (t instanceof InvocationTargetException) { // reflected error
            t = ((InvocationTargetException) t).getTargetException();
        }

        if (t != null) {

            if (t.getClass() != null) {
                clazz = t.getClass().getName();
            }
            msg = t.getMessage();
            if (null == msg && t.getCause() != null) {
                msg = t.getCause().getMessage();
            }

            errString = "Exception at " + new Date().toString()
                + "\nMessage: " + (null == msg ? "Unknown" : msg)
                + "\nClass: " + (null == clazz ? "Unknown" : clazz)
                + "\nStacktrace:";

            if (null == t.getStackTrace()) {
                errString += "null";
            } else {
                int lines = t.getStackTrace().length;
                if (0 == lines) {
                    errString += " is empty!";
                } else {
                    int more = lines;
                    for (int i = 0; i < stackTraceLen && i < lines; i++, more--) {
                        errString += "\n" + t.getStackTrace()[i].toString();
                    }
                    if (more > 0) errString += "\n" + more + " more...";
                }
            }
        }

        return errString;
    }

    public final void logWarnCounters() {
        this.logWarnCounters(this.logCounters);
    }

    public final void logWarnCounters(long[] counters) {
        String lwc = null;;
        for (Level level : Level.values()) {
            if (level.negative) {
                final long counter = counters[level.ordinal()];
                if (counter > 0) {
                    if (null == lwc) lwc = "Log statistics:";
                    lwc += ' ' + level.name() + "s:" + counter;
                }
            }
        }
        if (lwc != null) {
            // Hier nur Warn senden, da LogMailer und Co auf errors reagieren
            this.log(lwc, WARN, null);
            this.logCounters[WARN.ordinal()]--; // sich selbst nicht mitzaehlen
        }
    }

    /**
     * Erzeugt ein Log, das sich nicht selbst aufruft.<br>
     * Sinnvoll, wenn ein Logger, z.B. ein {@link LogMailWriter}
     * Fehler-Logs generiert, die allerdings nicht in den eigenen
     * Log-Kanal geschickt werden sollen.
     * @param sender {@link LogWriter} Verursacher
     * @return neues {@link Log}
     */
    public final Log createNonSelfCallingWriterLog(final LogWriter sender) {
        return new Log(this.logLevel).addLogWriter(new LogWriterAdapter() {
            @Override public void log(String message, Level logLevel) {
                Log.this.log(message, logLevel, sender);
            }
        });
    }

}
