/*
 * $Id: LogConsoleWriter.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;

import static de.cm.osm2po.logging.Log.Level.FATAL;
import static de.cm.osm2po.logging.Log.Level.INFO;
import static de.cm.osm2po.logging.Log.Level.PROGRESS;
import static de.cm.osm2po.misc.Utils.stringCropMid;

import java.io.PrintStream;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log.Level;

public class LogConsoleWriter implements LogWriter {

    public final static int NULL = 0;
    public final static int OUT = 1;
    public final static int ERR = 2;

    private Level logLevel = INFO;
    private PrintStream printStream;
    private boolean isLastLogProgress;
    private int lastProgressMsgLen;

    public LogConsoleWriter() {
    	this.setPrintTo(OUT);
    }

    @Override
    public LogWriter open(Config config, int index) {
        this.logLevel = PROGRESS;
        if (config != null) {
            String strLogTo = config.getProperty("log." + index + ".to");
            if (strLogTo != null) {
                strLogTo = strLogTo.trim().toLowerCase();
                this.setPrintTo(LogConsoleWriter.NULL);
                if ("out".equals(strLogTo))
                    this.setPrintTo(LogConsoleWriter.OUT);
                if ("err".equals(strLogTo))
                    this.setPrintTo(LogConsoleWriter.ERR);
            }
            Level logLevel = Level.byName(config.getProperty(
                    "log." + index + ".level", "progress"));
            this.setLogLevel(logLevel);
        }
        return this;
    }

    @Override
	public LogWriter setLogLevel(Level logLevel) {
		this.logLevel = logLevel;
		return this;
	}

    @Override
    public Level getLogLevel() {
        return this.logLevel;
    }

    public LogConsoleWriter setPrintTo(int printTo) {
        switch (printTo) {
        case OUT:
            this.printStream = System.out;
            break;
        case ERR:
            this.printStream = System.err;
            break;
        default:
            this.printStream = null;
            break;
        }
        return this;
    }

    @Override
    public void log(String message, Level logLevel) {
        if (logLevel.severity < this.logLevel.severity) return;

        String msg = (logLevel.prefix + message).replace("\n", "\n" + Level.spacer());

        if (this.printStream != null) {
            if (PROGRESS == logLevel) {
                msg = stringCropMid(msg, 80);

                this.isLastLogProgress = true;
                this.printStream.print("\r" + msg);

                // Laengenausgleich
                int msgLen = msg.length();
                for (int i = msgLen; i < this.lastProgressMsgLen; i++)
                    this.printStream.print(" ");
                this.lastProgressMsgLen = msgLen;

            } else {
                if (isLastLogProgress) {
                    this.isLastLogProgress = false;
                    this.printStream.print("\r" + msg);

                    // Laengenausgleich
                    int msgLen = msg.length();
                    for (int i = msgLen; i < this.lastProgressMsgLen; i++)
                        this.printStream.print(" ");
                    this.lastProgressMsgLen = 0;

                    this.printStream.println();

                } else {
                    this.printStream.println(msg);
                }
            }
        }
    }

    @Override
    public void close() {this.logLevel = FATAL;}

}
