/*
 * $Id: LogFileWriter.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log.Level;

/**
 * Ein einfacher Datei-Logger.
 * <br>Config-Beispiel:<br>
 * <pre>
 * log.2.class=de.cm.osm2po.logging.LogFileWriter
 * log.2.level=debug (default)
 * log.2.append=true (default true=anfuegen, false=neu)
 * log.2.file=[workDir]/[prefix]_2po.log (default Dateipfad)
 * log.2.maxSizeMb=0 (default), &gt;0 RollingBackupFileSizeTrigger
 * </pre>
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 */
public class LogFileWriter implements LogWriter {
    private static final int FILESIZECHECKINTERVAL = 1000;

    private PrintStream printStream;
    private File file;
    private int logCounter = FILESIZECHECKINTERVAL; // Bei Neustart 1x checken
    private boolean append;
    private long maxFileLen;
    private Level logLevel = Level.DEBUG; // Default
    private Log parentLog;

    @Override public LogWriter open(Config config, int index) {
        this.logLevel = Level.DEBUG;
        this.parentLog = config.getLog();

        File file = null;
        boolean append = true;

        String extFile = config.getProperty("log." + index + ".file");
        if (extFile != null) file = new File(extFile);
        else file = new File(config.getWorkDir(), config.getPrefix() + "_2po.log");

        append = Boolean.valueOf(config.getProperty(
                "log." + index + ".append", "true"));
        this.logLevel = Level.byName(config.getProperty(
                "log." + index + ".level", "debug"));
        this.maxFileLen = 0x100000L/*MB*/ * Integer.valueOf(config.getProperty(
                "log." + index + ".maxSizeMb", "0"));

        this.setFile(file, append);
        return this;
    }


    @Override
	public LogWriter setLogLevel(Level logLevel) {
		this.logLevel = logLevel;
		return this;
	}

    /**
     * Setzt das LogFile.<br>
     * Diese Methode kann sowohl initial als auch zur Laufzeit
     * aufgerufen werden, um die Datei zu wechseln.
     * Ruft dafuer in jedem Fall vorher {@link #close()}.
     * @param file {@link File}
     * @param append boolean Anfuegen ja/nein
     * @return this
     */
    public LogFileWriter setFile(File file, boolean append) {
        this.close(); // safe!
        file.getParentFile().mkdirs(); // ensure dir exists
        try {
            this.file = file;
            this.append = append;
            this.printStream = new PrintStream(
                    new FileOutputStream(file, append));
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        return this;
    }

    @Override
    public Level getLogLevel() {
        return this.logLevel;
    }

    @Override
    public void log(String message, Level logLevel) {
        if (null == this.printStream) return;
        if (logLevel.severity < this.logLevel.severity) return;
        if (Level.PROGRESS == logLevel) return;
        String msg = (logLevel.prefix + message).replace("\n", " | ");

        this.printStream.println(msg);

        // RollingFile-Backup
        if (this.maxFileLen > 0 && FILESIZECHECKINTERVAL < (++this.logCounter)) {
            this.logCounter = 0;
            // file.length() liefert 0, wenn nicht vorhanden!
            if (this.file.length() >= this.maxFileLen) {
                this.close();
                String ts = new SimpleDateFormat("yyMMddHHmmssSSS").format(new Date());
                File backupFile = new File(this.file.getParentFile(),
                        "BACKUP-" + ts + "-" + this.file.getName());
                boolean ok = this.file.renameTo(backupFile);
                this.setFile(this.file, this.append);
                if (!ok) { // Rename gescheitert!
                    this.maxFileLen = 0; // Ausschalten!
                    this.parentLog.log("Could not rename '"
                            + this.file.getAbsolutePath() + "' to '"
                            + backupFile.getAbsolutePath() + "'",
                            Level.ERROR, this);
                }
            }
        }
    }

    @Override
    public void close() {
        // File selbst nicht nullen, wird noch verwendet!
        if (this.printStream != null) {
            this.printStream.close();
            this.printStream = null;
        }
    }

}
