/*
 * $Id: LogMailWriter.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;

import static de.cm.osm2po.misc.Utils.TS;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.service.Smtp;

/**
 * Ein einfacher Logger, der aufgelaufene Meldungen per Mail verschickt.
 * Ausgeloest wird der Versandt entweder durch die Ueberschreitung einer
 * maximalen Anzahl Zeichen oder nach Erhalt einer {@link Level#MESSAGE}
 * Meldung oder hoeher oder aufgrund des Zeit-Intervall-Triggers.
 * <br>Das eigentliche Default-LogLevel ist {@link Level#WARN}.
 * <br><br><strong>Wichtig:</strong><br>
 * STARTTLS wird derzeit NICHT unterstuetzt. Siehe {@link Smtp}.
 * <br>Config-Beispiel:<br>
 * <pre>
 * log.2.class=de.cm.osm2po.logging.LogMailWriter
 * log.2.level=warn (default)
 * log.2.to=your@mail.com
 * log.2.from=my@mail.com
 * log.2.host=smtp.myprovider.com
 * log.2.user=myproviderUsername
 * log.2.pass=myproviderPassword
 * log.2.port=465 (default)
 * log.2.ssltls=true (default)
 * log.2.interval=60 (default, AutoSendInterval in Minuten)
 * log.2.maxchars=1024 (default Anzahl max Zeichen in Kilo)
 * log.2.prefix=OSM2PO (default Prefix fuer Betreff)
 * </pre>
 * <b>Empfehlung:</b><br>
 * Optimalerweise solllte dieser Logger an Index 0, also noch vor einem
 * etwaigen {@link LogConsoleWriter} oder {@link LogFileWriter}
 * konfiguriert werden, damit bei {@link #close()} die Aufraeum-Sequenz
 * zumindest noch in den oben genannten Logs erscheint.
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 * @see LogSlackWriter
 */
public class LogMailWriter implements LogWriter {
    private volatile boolean closed;
    private volatile Level topLoggedLevel;

    private Log log;
    private Level logLevel;
    private Smtp smtp;
    private String to;
    private StringBuilder logHistory;
    private int maxHistoryLength;
    private long counter;
    private ExecutorService senderThreads;
    private Thread intervalSenderThread;
    private long interval;
    private String subjPrefix;

    public LogMailWriter() {}
    public LogMailWriter(Smtp smtp) {this.smtp = smtp;}

    @Override
    public LogWriter setLogLevel(Level logLevel) {
        this.logLevel = logLevel;
        return this;
    }

    @Override
    public Level getLogLevel() {
        return this.logLevel;
    }

    @Override
    public LogWriter open(Config config, int index) {
        this.log = config.getLog().createNonSelfCallingWriterLog(this);
        this.closed = false;

        this.topLoggedLevel = Level.NULL;

        final String prefix = "log." + index + ".";

        this.setLogLevel(Level.byName(
                config.getProperty(prefix + "level", "warn")));

        int port = Integer.parseInt(
                config.getProperty(prefix + "port", "465"));
        boolean ssltls = Boolean.valueOf(
                config.getProperty(prefix + "ssltls", "true"));

        this.to = config.getProperty(prefix + "to");

        String host = config.getProperty(prefix + "host");
        String from = config.getProperty(prefix + "from");
        String user = config.getProperty(prefix + "user");
        String pass = config.getProperty(prefix + "pass");

        this.logHistory = new StringBuilder();
        this.maxHistoryLength = 1024 * Integer.parseInt(
                config.getProperty(prefix + "maxchars", "1024"));
        this.interval = 60000 * Integer.parseInt(
                config.getProperty(prefix + "interval", "60"));

        this.subjPrefix = config.getProperty(prefix + "prefix", "OSM2PO");

        if (null == this.smtp) {
            this.smtp = new Smtp( // Wg. UnitTest, see Constructor
                    log, from, host, port, ssltls, user, pass);
        }
        this.senderThreads = Executors.newFixedThreadPool(5);

        this.intervalSenderThread = new Thread(new IntervalSender());
        this.intervalSenderThread.setName(
                LogMailWriter.class.getSimpleName() + "IntervalThread");
        this.intervalSenderThread.start();

        return this;
    }

    @Override
    public void log(String msg, Level logLevel) {
        if (this.closed) return;
        if (logLevel.severity < this.logLevel.severity) return;
        if (logLevel == Level.PROGRESS) return;

        synchronized (this) {
            if (logLevel.severity > this.topLoggedLevel.severity)
                this.topLoggedLevel = logLevel;
            this.logHistory.append('[' + logLevel.name() + "] " + msg).append('\n');

            if (logLevel.severity >= Level.MESSAGE.severity
                    || this.logHistory.length() >= this.maxHistoryLength)
                this.sendHistory();
        }
    }

    private synchronized void sendHistory() {
        String subjLevelStr = this.topLoggedLevel.name();
        String subject = this.subjPrefix + "-LOG[" + (++this.counter) + "]-" + subjLevelStr;
        String message = this.logHistory.toString(); // tatsaechlich eine Kopie
        this.logHistory.setLength(0);
        this.topLoggedLevel = Level.NULL;
        this.senderThreads.execute(new Sender(
                this.log, this.smtp, this.to, subject, message));
    }

    @Override
    public void close() {
        if (this.closed) return;
        try {
            this.log.info("SMTP starting close sequence");

            this.closed = true;
            this.intervalSenderThread.interrupt();
            this.sendHistory();
            this.senderThreads.shutdown();

            if (this.senderThreads.awaitTermination(5, TimeUnit.SECONDS)) {
                this.log.info("SMTP closed");
            } else {
                this.log.warn("SMTP closed - not all mails delivered");
            }

        } catch (Throwable t) {
            this.log.error(t);
        }
    }

    private class IntervalSender implements Runnable {
        @Override
        public void run() {
            final LogMailWriter lm = LogMailWriter.this;
            try {
                while (!lm.closed) {
                    Thread.sleep(lm.interval);
                    if (lm.closed) return; // Doppelt haelt besser

                    if (lm.logHistory.length() > 0) {
                        lm.log.info("SMTP " + lm.getClass().getSimpleName()
                                + "-Interval reached. Flushing - " + TS());
                        lm.sendHistory();
                    } else {
                        lm.log.info(
                                "SMTP LogHistory is empty. Nothing sent - " + TS());
                    }
                }

            } catch (InterruptedException ie) {
                lm.log.info("SMTP " + lm.getClass().getSimpleName()
                        + this.getClass().getSimpleName() + " interrupted");
            } catch (Throwable t) {
                lm.log.error(t);
            }
        }
    }

    private static class Sender implements Runnable {
        private static final int maxErrs = 3;
        private static final long sleepAfterSendErr = 60000; // 1 Min.

        private final Log log;
        private final Smtp smtp;
        private final String to;
        private final String subject;
        private final String message;

        public Sender(Log log, Smtp smtp, String to, String subject, String message) {
            this.log = log;
            this.smtp = smtp;
            this.to = to;
            this.subject = subject;
            this.message = message;
        }

        @Override
        public void run() {
            for (int errs = 1; errs <= maxErrs; errs++) { // Sei hartnaeckig
                try {
                    this.log.progress("SMTP Log to " + this.to);
                    if (this.smtp.send(this.subject, this.to, this.message)) {
                        this.log.info("SMTP '" + this.subject + "' sent to " + this.to);
                        return;

                    } else {
                        this.log.warn("SMTP failed [" + errs + "/" + maxErrs + "]");
                        Thread.sleep(sleepAfterSendErr);
                    }

                } catch (InterruptedException ie) {
                    this.log.info(LogMailWriter.class.getSimpleName() + ".Sender interrupted");
                    break;

                } catch (Exception e) {
                    this.log.error(e);
                    break;
                }
            }
            this.log.error("SMTP '" + this.subject + "' discarded");
        }

    }

}
