/*
 * $Id: LogUtilConsoleWriter.java 5991 2024-11-01 10:20:54Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.logging.ConsoleHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log.Level;

public class LogUtilConsoleWriter implements LogWriter {
// TODO: Nach System.exit() ist dieser Logger still wie die Nacht.

    private Level logLevel;
    private Logger logger;

    private static DateTimeFormatter DF = DateTimeFormatter.ofPattern("yy-MM-dd HH:mm:ss.[SSS]");
    private static ZoneId zoneId = ZoneId.systemDefault();

    public LogUtilConsoleWriter() {

        Formatter formatter = new Formatter() {
            @Override
            public String format(LogRecord record) {
                String message = record.getMessage();
                if (null == message) message = "<NULL>";

                StringBuilder sb = new StringBuilder(0x100);
                // record.getInstant() not in Java8
                sb.append(DF.format(LocalDateTime.ofInstant(
                        Instant.ofEpochMilli(record.getMillis()), zoneId)));
                sb.append(" [");
                sb.append(pad(record.getLevel().toString(), 7));
                sb.append("] ");
                sb.append(message.replace("\n", "\n| "));
                sb.append("\n");
                return sb.toString();
            }
        };

        Handler handler = new ConsoleHandler();
        handler.setLevel(java.util.logging.Level.ALL);
        handler.setFormatter(formatter);

        this.logLevel = Level.NULL; // delegate all

        this.logger = Logger.getLogger("osm2po");
        this.logger.setLevel(java.util.logging.Level.ALL);
        this.logger.setUseParentHandlers(false); // log4j calls this additivity
        this.logger.addHandler(handler);
        this.logger.setUseParentHandlers(false); // else each log appears at least twice
    }

    private static String pad(String s, int length) {
        char[] sarr = null == s ? new char[0] : s.toCharArray();
        char[] carr = new char[length];
        for (int i = 0; i < carr.length; i++) {
            if (i < sarr.length) carr[i] = sarr[i];
            else carr[i] = ' ';
        }
        return new String(carr);
    }

    /**
     * @return neue Log-Instanz mit einem {@link LogUtilConsoleWriter}
     */
    public static Log logInstance() {
        return new Log(Level.NULL).addLogWriter(new LogUtilConsoleWriter());
    }

    @Override
    public LogWriter open(Config config, int index) {
        if (config != null) {
            Log log = config.getLog();
            if (log != null) {
                log.debug("LogWriter[" + index + "] uses "
                        + this.logger.getClass().getName());
            }
        }
        return this;
    }

    @Override
    public void log(String message, Level logLevel) {
        if (logLevel.severity < this.logLevel.severity) return;
        if (logLevel.severity < Level.FATAL.severity) // wegen Stacktrace
//        	message = message.replace("\n", " ");
        try {
            // LEVEL_PROGRESS (4) ignored
	        switch (logLevel) {
	        case DATA: this.logger.finest(message); break;
	        case BUGFIX: this.logger.finer(message); break;
	        case DEBUG: this.logger.fine(message); break;
	        case NOTE: this.logger.fine(message); break;
	        case WARN: this.logger.warning(message); break;
	        case INFO: this.logger.info(message); break;
	        case ERROR: this.logger.severe(message); break;
	        case MESSAGE: this.logger.info(message); break;
            case ALERT: this.logger.severe(message); break;
	        case FATAL: this.logger.severe(message); break;
            default:
	        }

        } catch (Exception e) {
        	throw new RuntimeException(e);
        }
    }

	@Override
	public LogWriter setLogLevel(Level logLevel) {
		this.logLevel = logLevel;
		return this;
	}

	@Override public Level getLogLevel() {return this.logLevel;}
    @Override public void close() {}

}
