/*
 * $Id: OtelAgentWrapper.java 6003 2024-12-01 16:04:45Z carsten $
 *
 * Copyright 2024 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Executor;
import java.util.concurrent.ThreadPoolExecutor;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.GlobalOpenTelemetry;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.OpenTelemetry;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.common.AttributeKey;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.common.Attributes;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.common.AttributesBuilder;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.metrics.LongCounter;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.metrics.LongHistogram;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.metrics.Meter;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.metrics.ObservableLongGauge;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.trace.Span;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.trace.SpanBuilder;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.trace.SpanKind;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.trace.StatusCode;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.api.trace.Tracer;
import io.opentelemetry.javaagent.shaded.io.opentelemetry.context.Context;

/**
 * Diese Klasse kapselt Methoden des Opentelemetry-JavaAgent.
 * Das ermoeglicht manuellen Code fuer OTEL zu schreiben, also Spans
 * und Metriken, ohne dass zur Laufzeit- oder Kompilierzeit
 * Probleme auftreten, falls der JavaAgent nicht verwendet wird.
 * @author (c) 2024 - Carsten Moeller, info@osm2po.de
 */
public class OtelAgentWrapper {
    private OtelAgentWrapper() {/*static*/}

    private static OpenTelemetry otel;
    private static Tracer otelTracer;
    private static Meter otelMeter;
    private static Log log;

    public static final SpanWrapper NOOP_SPAN = new NoOpSpan();

    public static void init(final Config config) {
        final String serviceName = config.getProperty("otel.service.name", "osm2po");
        log = config.getLog();
        init(serviceName, log);
    }

    public static void init(final String serviceName, Log log) {
        try {
            otel = GlobalOpenTelemetry.get();
            otelTracer = otel.getTracer(serviceName);
            otelMeter = otel.getMeter(serviceName);

        } catch (Throwable t) {
            log.warn("OTEL disabled. Set -javaagent:<pathTo>/opentelemetry-javaagent.jar");
        }
    }

    /**
     * Wrappt einen {@link Executor} fuer OTEL.
     * Ohne diesen Mechanismus wuerden {@link ThreadPoolExecutor}-{@link Thread}s in den
     * falschen Traces landen bzw. die Spans wuerden falschen Parents zugewiesen
     * werden, da die {@link Thread#getId()} im Pool wiederverwendet wird.
     * Mit immer neuen {@link Thread}s, also Thead-IDs waere dies kein Problem.
     * @param executor {@link Executor}
     * @return {@link Executor}, mit OTEL als gewrappte Variante, sonst gleich Parameter.
     */
    public static Executor wrapExecutor(final Executor executor) {
        if (null == otel) return executor;
        else return Context.taskWrapping(executor);
    }

    private static Attributes toOtelAttribs(String[] kvPairs) {
        if (null == kvPairs || kvPairs.length < 2) return Attributes.empty();
        AttributesBuilder builder = Attributes.builder();
        for (int i = 1; i < kvPairs.length; i += 2) builder.put(kvPairs[i-1], kvPairs[i]);
        return builder.build();
    }

    public enum SpanKindWrapper {
        CLIENT, SERVER, INTERNAL;
    }

    public static interface LongMeter {
        void set(long value, String... attributesKeyValPairs);
        default void close(){}; // e.g. for async threads
    }

    public static SpanWrapper startSpan(final String spanName, final SpanKindWrapper spanKind) {
        if (null == otelTracer) {
            return NOOP_SPAN;
        } else {
            SpanBuilder spanBuilder = otelTracer.spanBuilder(spanName);
            spanBuilder.setSpanKind(SpanKind.valueOf(spanKind.name()));
            Span span = spanBuilder.startSpan();
            span.makeCurrent();
            return new OtelSpan(span);
        }
    }

    public static SpanWrapper currentSpan() {
        if (null == otel) return NOOP_SPAN;
        return new OtelSpan(Span.current());
    }

    public static interface SpanWrapper {
        SpanWrapper setAttribute(final String key, final Object value );
        SpanWrapper setAttribute(final String key, final long value );
        SpanWrapper setOk(boolean ok);
        SpanWrapper addEvent(final Object value);
        SpanWrapper addEvent(final String key, final Object value);
        SpanWrapper recordException(final Throwable t);
        void end();
    }

    private final static class NoOpSpan implements SpanWrapper {
        @Override public SpanWrapper setAttribute(String key, Object value) {return this;}
        @Override public SpanWrapper setAttribute(String key, long value) {return this;}
        @Override public SpanWrapper setOk(boolean ok) {return this;}
        @Override public SpanWrapper addEvent(Object value) {return this;}
        @Override public SpanWrapper addEvent(String key, Object value) {return this;}
        @Override public SpanWrapper recordException(Throwable t) {return this;}
        @Override public void end() {}
    }

    private final static class OtelSpan implements SpanWrapper {
        private Span otelSpan;

        public OtelSpan(final Span otelSpan) {
            Objects.requireNonNull(otelSpan);
            this.otelSpan = otelSpan;
        }

        @Override
        public final SpanWrapper setAttribute(final String key, final Object value ) {
            this.otelSpan.setAttribute(key, value.toString());
            return this;
        }

        @Override
        public final SpanWrapper setAttribute(final String key, final long value ) {
            this.otelSpan.setAttribute(key, value);
            return this;
        }

        @Override
        public final SpanWrapper setOk(boolean ok) {
            this.otelSpan.setStatus(ok ? StatusCode.OK : StatusCode.ERROR);
            return this;
        }

        @Override
        public final SpanWrapper addEvent(final Object value) {
            this.otelSpan.addEvent(value.toString());
            return this;
        }

        @Override
        public final SpanWrapper addEvent(final String key, final Object value) {
            this.otelSpan.addEvent(key, Attributes.of(AttributeKey.stringKey(key), value.toString()));
            return this;
        }

        @Override
        public final SpanWrapper recordException(final Throwable t) {
            this.otelSpan.recordException(t);
            return this;
        }

        @Override
        public final void end() {
            this.otelSpan.end();
        }
    }


    // ---------- METRICS --------------------------------------------

    /**
     * Erstellt einen Counter.
     * @param name String Name of Meter
     * @return {@link LongMeter}
     */
    public static LongMeter createCounter(final String name) {
        return new Counter(name);
    }

    /**
     * Erstellt einen asynchronen Gauge, der den hoechsten Long-Wert sendet.
     * Nach jedem Senden wird der Max-Wert reinitialisiert.
     * @param name String Name of Meter
     * @return {@link LongMeter}
     */
    public static LongMeter createMaxGauge(final String name) {
        return new MaxGauge(name);
    }

    /**
     * Erstellt ein Histogramm von Long-Werten.
     * @param name String Name of Meter
     * @param buckets VarArg-Array von Long
     * @return {@link LongMeter}
     */
    public static LongMeter createHistogram(final String name, long... buckets) {
        Long[] bucks = new Long[buckets.length];
        for (int i = 0; i < buckets.length; i++) bucks[i] = Long.valueOf(buckets[i]);
        return new Histogram(name, bucks);
    }


    public final static class Counter implements LongMeter {
        private final String name;
        private LongCounter meter;

        private Counter(final String name) {
            this.name = name;
        }

        private boolean initOk() {
            if (null == this.meter) {
                if (null == otelMeter) return false;
                this.meter = otelMeter.counterBuilder(this.name).build();
            }
            return true;
        }

        @Override
        public void set(long value, String... kvPairs) {
            if (this.initOk()) this.meter.add(value, toOtelAttribs(kvPairs));
        }
    }

    public final static class Histogram implements LongMeter {
        private final String name;
        private final List<Long> bucketBoundaries;
        private LongHistogram meter;

        private Histogram(final String name, Long[] buckets) {
            this.name = name;
            this.bucketBoundaries = Arrays.asList(buckets);
        }

        private boolean initOk() {
            if (null == this.meter) {
                if (null == otelMeter) return false;
                this.meter = otelMeter.histogramBuilder(this.name).ofLongs()
                    .setExplicitBucketBoundariesAdvice(this.bucketBoundaries)
                    .build();
            }
            return true;
        }

        @Override
        public void set(long value, String... kvPairs) {
            if (this.initOk()) this.meter.record(value, toOtelAttribs(kvPairs));
        }
    }

    public final static class MaxGauge implements LongMeter {
        private final String name;
        private ObservableLongGauge meter;
        private long max = Long.MIN_VALUE;
        private Attributes attributes;

        private MaxGauge(final String name) {
            this.name = name;
        }

        private boolean initOk() {
            if (null == this.meter) {
                if (null == otelMeter) return false;
                this.meter = otelMeter.gaugeBuilder(name).ofLongs().buildWithCallback(
                    measurement->{
                        final long value = Long.MIN_VALUE == this.max ? 0 : this.max;
                        if (null == this.attributes) {
                            measurement.record(value); // null attribs is not allowed
                        } else {
                            measurement.record(value, this.attributes);
                        }
                        this.max = Long.MIN_VALUE;
                        this.attributes = null;
                    }
                );
            }
            return true;
        }

        @Override
        public void set(long value, String... kvPairs) {
            if (value > this.max) this.max = value;
            if (this.initOk()) this.attributes = toOtelAttribs(kvPairs);
        }

        @Override
        public void close() {
            if (this.meter != null) {
                this.meter.close();
                this.meter = null;
            }
        }

    }

}
