package de.cm.osm2po.misc;

import java.util.ArrayList;
import java.util.List;

/**
 * Ein String-Parser, der aus einer Regel Kommando-Namen
 * erstellt, die ueber einen angeschlossenen {@link Executor}
 * ausgefuehrt werden.<br>
 * Die Jobs laufen in {@link Thread}s und werden je nach
 * Operator sequentiell oder parallel gestartet.<br>
 * Die Grammatik:<br>
 * <ul>
 * <li><code>a,b</code> Sequentielle Ausfuehrung</li>
 * <li><code>a|b</code> Parallele Ausfuehrung</li>
 * <li><code>a|b,c</code> a parallel zu b, danach erst c</li>
 * <li><code>(a,b)|c</code> b erst nach a, parallel zu beiden c</li>
 * </ul>
 * <b>Wichtig:</b><br>
 * Als Token werden nur alphanumerische Bezeichner a-z, A-Z, 0-9
 * und "-", sowie "_" akzeptiert. Falsche Zeichen werden mit einer
 * {@link IllegalArgumentException} quittiert.<br>
 * Leerzeichen werden grundsaetzlich ignoriert. Das gilt auch
 * fuer Tokens der Form "a b c", was zu "abc" wird.<br>
 * Wird vor einer oeffnenden Klammer kein "|" oder "," gesetzt,
 * so wird "|" angenommen und der Klammerausdruck parallel verarbeitet.
 * @author (c) 2021 - Carsten Moeller, info@osm2po.de
 */
public class Threader implements Runnable {
    public interface Executor {void execute(String command) throws Exception;}

    private String name = null;
    private Threader next = null; // next sequential job
    private List<Threader> parallelJobs = new ArrayList<>(); // parallel jobs
    private Executor executor;

    public Threader(String rule, Executor executor) {
        this(executor, rule, new int[] {0});
    }

    @Override public String toString() {
        return name != null ? name : parallelJobs + (next != null ? ">" + next : "");
    }

    @Override public final void run() {
        try {
            if (this.executor != null) this.executor.execute(this.name);
            List<Thread> parallelThreadList = new ArrayList<Thread>();
            for (Threader job : this.parallelJobs) {
                Thread thread = new Thread(job);
                if (job.name != null) thread.setName(job.name);
                parallelThreadList.add(thread);
            }
            for (Thread thread : parallelThreadList) thread.start();
            for (Thread thread : parallelThreadList) thread.join();
            if (this.next != null) this.next.run();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }


    private Threader(Executor executor, String name) {
        this.executor = executor;
        this.name = name;
    }

    private Threader(Executor executor, String rule, int[] idx) {
        String token = "";
        while (idx[0] <= rule.length()) {
            char c = idx[0] == rule.length() ? '$' : rule.charAt(idx[0]++);
            if (c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z' || c >= '0' && c <= '9'
                    || c == '-' || c == '_') {
                token += c;

            } else if (c != ' ') { // Skip WS - Falle: Token "a b" wird zu "ab"
                if (token.length() > 0) { // always
                    // interpretiert ) und ( als |, wenn | selbst nicht gesetzt
                    this.parallelJobs.add(new Threader(executor, token));
                    token = ""; // reset
                }

                switch (c) {
                case '|' : continue; // just better to read
                case '$' : return;
                case ',' :
                    this.next = new Threader(executor, rule, idx);
                    break;
                case '(' :
                    this.parallelJobs.add(new Threader(executor, rule, idx));
                    idx[0]++; // skip my corresponding ")"
                    break;
                case ')' :
                    idx[0]--; // skip deeper recurse level caused by this.next
                    return;
                default: throw new IllegalArgumentException(
                        "Unsupported Character: '" + c + "'");
                }
            }
        }
    }

}
