/*
 * $Id: Utils.java 5930 2024-10-11 15:31:47Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static java.lang.StrictMath.floor;

import java.awt.Desktop;
import java.awt.Toolkit;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.math.BigInteger;
import java.net.URI;
import java.net.URL;
import java.nio.charset.Charset;
import java.security.ProtectionDomain;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.Properties;
import java.util.Random;

public class Utils {
    public final static Charset UTF8 = Charset.forName("UTF-8");

    private Utils() {/*static*/}

    private static Random random = new Random();

    private static byte[] LOCODE_ENC = new byte[256];
    static {
        Arrays.fill(LOCODE_ENC, (byte)-1);
        for (byte i = 0; i < 10; i++) {LOCODE_ENC['0'+i] = i;}
        for (byte i = 0; i < 26; i++) {LOCODE_ENC['A'+i] = (byte) (i+10);}
    }

    //############## ByteArray / HexString Konvertierung #####

    private static final char[] hexChars = "0123456789ABCDEF".toCharArray();

    public static String bytesToHex(byte[] byteArray) {
        int len = byteArray.length;
        char[] hex = new char[len * 2];
        for (int idx = 0; idx < len; idx++) {
            hex[idx * 2] = hexChars[(byteArray[idx] >>> 4) & 0xF];
            hex[(idx * 2) + 1] = hexChars[byteArray[idx] & 0xF];
        }
        return new String(hex);
    }

    public static byte[] hexToBytes(String s) {
        char[] chars = s.toCharArray();
        byte[] bytes = new byte[chars.length / 2];
        for (int i = 0; i < bytes.length; i++) {
            int v1 = chars[i * 2];
            int v0 = chars[i * 2 + 1];
            if (v1 > 57) v1 -= 7;
            if (v0 > 57) v0 -= 7;
            bytes[i] = (byte) ((v1 & 0xF) << 4 | (v0 & 0xF));
        }
        return bytes;
    }

    //##################### JAVA-Version #########################

    public static final double getJavaSpecVersion() {
        double javaSpec = 1.0;
        try {javaSpec = Double.parseDouble(
                System.getProperty("java.specification.version"));
        } catch (Exception e) {}
        return javaSpec;
    }

    public static final String getJavaInfo() {
        String jvendor = System.getProperty("java.vendor");
        String jversion = System.getProperty("java.version");
        return "Java " + jversion + " (" + jvendor + ")";
    }

    //########################### FORMAT ###########################

    public final static NumberFormat DECIMAL_FORMAT_0_DIGITS =
        new DecimalFormat("#,##0");
    public final static NumberFormat DECIMAL_FORMAT_2_DIGITS =
        new DecimalFormat("#,##0.00");
    public final static NumberFormat DECIMAL_FORMAT_4_DIGITS =
            new DecimalFormat("#,##0.0000");
    public final static DateFormat DATE_TIME_FORMAT_TS =
        new SimpleDateFormat("yyMMdd-HH:mm.ssSSS");

    /**
     * Formatiert einen long als Dezimal-String mit TausenderTrenner.
     * @param l long.
     * @return Dezimal-String.
     */
    public final static String DF(long l) {
        return DECIMAL_FORMAT_0_DIGITS.format(l);}

    /**
     * Formatiert einen double als Dezimal-String mit TausenderTrenner
     * und vier Nachkommastellen.
     * @param d double.
     * @return Dezimal-String.
     */
    public final static String DF4(double d) {
        return DECIMAL_FORMAT_4_DIGITS.format(d);}

    /**
     * Formatiert einen double als Dezimal-String mit TausenderTrenner
     * und zwei Nachkommastellen.
     * @param d double.
     * @return Dezimal-String.
     */
    public final static String DF2(double d) {
        return DECIMAL_FORMAT_2_DIGITS.format(d);}

    /**
     * Formatiert ein Date zu {@link #DATE_TIME_FORMAT_TS}.
     * @param d Date.
     * @return Date-String.
     */
    public final static String DF(Date d) {
        return DATE_TIME_FORMAT_TS.format(d);
    }

    /**
     * Formatiert das aktuelle Date zu {@link #DATE_TIME_FORMAT_TS}.
     * @return Date-String.
     */
    public final static String TS() {
        return DATE_TIME_FORMAT_TS.format(new Date());
    }

    /**
     * Rundet einen double auf 7 Nachkommastellen
     * @param dval double
     * @return double gerundet auf 7 Nachkommastellen
     */
    public final static double roundE7(double dval) {
        return ((long)floor(dval * 1e7 + 0.5d)) / 1e7;
    }

    /**
     * Rundet einen double auf 5 Nachkommastellen
     * @param dval double
     * @return double gerundet auf 5 Nachkommastellen
     */
    public final static double roundE5(double dval) {
        return ((long)floor(dval * 1e5 + 0.5d)) / 1e5;
    }

    /**
     * Zerlegt einen long-integer in einzelne Bytes
     * getrennt durch Komma. Das niedrigste Byte
     * steht danach links. Die Laenge variiert je
     * nach Groesse. So laesst sich diese Funktion auch
     * auf ints oder shorts anwenden. Bei negativen Werten
     * hier gegebenenfalls vorher ausmaskieren.
     * @param val long
     * @return String formatiert als Byte-Kette.
     */
    public final static String toOctetString(long val) {
        String s = "";
        NumberFormat nf = new DecimalFormat("000");
        do {
            long octet = val & 0xFFL;
            String o = nf.format(octet);
            s = (0 == s.length()) ? o : o + '.' + s;
            val >>>= 8;
        } while (val != 0);
        return s;
    }

    //######################## LONG INT/INT ########################

    public static long hilo(int hi, int lo) {
        return ((long)hi << 32) | (lo & 0xFFFFFFFFL);
    }
    public static int loOf(long hilo) {
        return (int) (hilo & 0xFFFFFFFFL);
    }
    public static int hiOf(long hilo) {
        return (int) (hilo >> 32);
    }

    //########################### MEMORY ###########################

    /**
     * Liefert die ungefaehre Menge des noch freien RAMs als String
     * Information.
     * @return Verbleibender RAM-Speicher als String Info.
     */
    public static String MB() {
        long ram = getMemoryAvailable();
        ram >>>= 20; // MegaByte
        return DF(ram) + "M";
    }

    /**
     * Liefert die ungefaehre Menge des noch freien RAMs als long.
     * @return Verbleibender RAM-Speicher als long.
     */
    public static long getMemoryAvailable() {
        long free = Runtime.getRuntime().freeMemory();
        long total = Runtime.getRuntime().totalMemory();
        long max = Runtime.getRuntime().maxMemory();
        long available = max - total + free;
        return available;
    }

    public static String substringRight(final String s, int n) {
        if (null == s) return "";
        if (s.length() < 2) return s;
        return s.substring(s.length() - n, s.length());
    }

    public static String msToDhms(long ms) {
        long seconds = ms / 1000;
        long minutes = seconds / 60;
        seconds -= minutes * 60;
        long hours = minutes / 60;
        minutes -= hours * 60;
        long days = hours / 24;
        hours -= days * 24;
        return substringRight("0" + days, 2)
            + ":" + substringRight("0" + hours, 2)
            + ":" + substringRight("0" + minutes, 2)
            + ":" + substringRight("0" + seconds, 2);
    }

    //########################### FILE IO ###########################

    public static void copyFile(File from, File to) {
        try {
            InputStream in = new FileInputStream(from);
            OutputStream out = new FileOutputStream(to);
            byte[] buf = new byte[4096];
            int len;
            while ((len = in.read(buf)) > 0) out.write(buf, 0, len);
            in.close();
            out.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static synchronized File createTmpFile(File dir) {
        long rndMask = 0x7FFFFFFFFFFFFFFFL;
        File file = new File(dir,
                String.valueOf(random.nextLong() & rndMask) + ".tmp");
        file.deleteOnExit();
        return file;
    }

    /**
     * Liefert das Laufzeit-Hauptverzeichnis unabhaengig davon,
     * ob die Binaerklassen entpackt oder als Jar vorliegen.
     * @return {@link File} Verzeichnis oder null.
     */
    public static File getJarDir() {
        File dir = null;
        try {
            ProtectionDomain pd = Utils.class.getProtectionDomain();
            if (pd != null) {
                URL url = pd.getCodeSource().getLocation();
                dir = new File(url.toURI().getPath()).getParentFile();
            }
        } catch (Exception e) {}
        return dir;
    }


    /**
     * Liefert das osm2po-core.jar als {@link File}, wenn von
     * hier gestartet und nicht im Entwicklungsmodus.
     * @return {@link File} JarFile, wenn binaer mittels java -jar gestartet.
     */
    public static File getJarFile() {
        try {
            ProtectionDomain pd = Utils.class.getProtectionDomain();
            if (pd != null) {
                URL url = pd.getCodeSource().getLocation();
                File file = new File(url.toURI().getPath());
                if (file.isFile() && file.getName().endsWith(".jar")) {
                    return file;
                }
            }
        } catch (Exception e) {}
        return null;
    }

    /**
     * Versucht innerhalb der Laufzeitumgebung ein Verzeichnis
     * oder eine Datei zu finden, indem zuerst im aktuellen
     * Verzeichnis und danach im Laufzeit-Hauptverzeichis
     * mittels {@link #getJarDir()} gesucht wird.
     * @param name String Name des Verzeichnisses oder Datei
     * @param findFile boolean true: Finde Datei, sonst Verzeichnis.
     * @return Gefundener Datei-Pfad oder null.
     */
    public static File findFileOrDir(String name, boolean findFile) {
        File file = new File(name);
        if (!file.exists()) {
            file = new File(getJarDir(), name); // dir laut Doku null-sicher!
            if (!file.exists()) return null;
        }
        if (file.isDirectory() && findFile) {
            return null;
        }
        return file;
    }

    /**
     * Liest eine Datei komplett in ein ByteArray ein.
     * @param dataFile {@link File}
     * @return byte[] ByteArray
     * @throws Exception beliebiger Natur
     */
    public static byte[] loadFileIntoByteArray(File dataFile) throws Exception {
        long fileSize = dataFile.length();
        if (fileSize > Integer.MAX_VALUE -1) { // +1 = EOF
            throw new Exception("File too large!");
        }

        // Datei ins Memory lesen
        byte[] memory = new byte[(int) fileSize];
        FileInputStream fis = new FileInputStream(dataFile);
        byte[] buf = new byte[4096];
        for (int pos = 0, len = fis.read(buf); len > -1; pos += len, len = fis.read(buf)) {
            System.arraycopy(buf, 0, memory, pos, len);
        }
        fis.close();

        return memory;
    }

    public static byte[] loadStreamIntoByteArray(InputStream is) throws Exception {
        if (is != null) {
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x10000);
            byte[] buf = new byte[0x100];
            for (int nRead = is.read(buf); nRead != -1; nRead = is.read(buf))
                baos.write(buf, 0, nRead);
            return baos.toByteArray();
        } else return null;
    }

    //########################### AWT ###########################

    public static void openBrowser(final String url) {
        try {
            if (Desktop.isDesktopSupported()
                    && Desktop.getDesktop().isSupported(Desktop.Action.BROWSE)) {
                Desktop.getDesktop().browse(new URI(url));
            }
        } catch (Throwable t) {
            t.printStackTrace(System.err);
        }
    }

    public static void beep(int count) {
        try {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            for (int i = 0; i < count; i++) {
                toolkit.beep();
                Thread.sleep(500);
            }
        } catch (Throwable t) {
            t.printStackTrace(System.err);
        }
    }

    //########################### ARGS ###############################


    /**
     * Liest ein Argumenten-Array und liefert dieses
     * als {@link Properties} zurueck.
     * @param args Array of String NULLABLE
     * @return {@link Properties} NOT NULL.
     * <br><br><b>Wichtig:</b><br>
     * Diese Methode akzeptiert nur Argumente der Form <code>key=value</code>
     */
    public static Properties readArgsIntoProps(String[] args) {
        Properties props = new Properties();
        if (args != null) for (int i = 0; i < args.length; i++) {
            String s = args[i];
            int equalsSignPos = s.indexOf('=');
            if (equalsSignPos > 0) {
                String key = s.substring(0, equalsSignPos);
                String val = s.substring(equalsSignPos + 1);
                props.setProperty(key.trim(), val.trim());
            }
        }
        return props;
    }

    //########################### PRIME ############################

    /**
     * Findet die naechste Primzahl.
     * @param longNumber long zahl
     * @return Primzahl &gt; longNumber
     */
    public static long findNextPrime(long longNumber){
        final BigInteger bigTwo = BigInteger.valueOf(2);
        BigInteger bigi = BigInteger.valueOf((longNumber+1)|1);
        while(!bigi.isProbablePrime(128)) bigi = bigi.add(bigTwo);
        return bigi.longValue();
    }

    //########################## STRING ############################

    public static String stringCropMid(final String s, int maxLen) {
        if (null == s) return null; // ok
        int len = s.length();
        if (len <= maxLen) return s; // ok
        if (maxLen < 4) return "..."; // Unsinn
        int maxLenHalf = (maxLen -3) >>> 1; // -3 fuer ...
        int evenOne = 1-(maxLen & 1); // 1 wenn gerade, 0 wenn ungerade
        String s1 = s.substring(0, maxLenHalf + evenOne);
        String s2 = s.substring(len - maxLenHalf);
        return s1 + "..." + s2;
    }

    public static String stringToXmlUTF8(String text) {
        StringBuilder sb = new StringBuilder();
        for(int i = 0; i < text.length(); i++){
            char c = text.charAt(i);
            switch(c){
            case '<': sb.append("&lt;"); break;
            case '>': sb.append("&gt;"); break;
            case '\"': sb.append("&quot;"); break;
            case '&': sb.append("&amp;"); break;
            case '\'': sb.append("&apos;"); break;
            default: sb.append(c);
            }
        }
        return new String(sb.toString().getBytes(UTF8));
    }

    /**
     * Konvertiert einen un/locode in einen int.<br>
     * Ein Locode bezeichnet einen wichtigen Verkehrsknotenpunkt.
     * Er besteht aus den Buchstaben A-Z, darf aber auch die Ziffern
     * 0-9 enthalten. Ein Locode ist immer 5 Zeichen lang.
     * @param locode String locode
     * @return int encoded locode als int, -1 bei Fehler
     */
    public static int locodeToInt(final String locode) {
        if (null == locode) return -1;
        if (locode.length() != 5) return -1;
        int iloc = 0;
        int pow36 = 1;
        for (int i = 4; i >= 0; i--) { // reverse to mainain the sort order
            byte c = (byte) locode.charAt(i);
            if (c >= 65) c &=~ 0x20; // kill lower-case bit
            byte enc = LOCODE_ENC[c];
            if (-1 == enc) return -1;
            iloc +=  enc * pow36;
            pow36 *= 36;
        }
        return iloc;
    }

    @SuppressWarnings("unchecked")
    public static<T> T getAnnotationDefault(Class<? extends Annotation> annotationClass,String element) {
        try {
            Method method = annotationClass.getMethod(element,(Class[])null);
            return((T)method.getDefaultValue());
        } catch (Exception e) {
            return null;
        }
    }

    //########################### MISC ############################


    /**
     * Vorgriff auf Java14 Records.
     * @param clazz {@link Class}
     * @return record-Klasse ja/nein. Aktuell immer nein.
     */
    public static boolean isRecord(Class<?> clazz) {
        return false;
        // return targetClass.isRecord(); Java14
    }


} // EOC
