/*
   Copyright (c) 2012 Carsten Moeller, Pinneberg, Germany. <info@osm2po.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   $Id: OsmO5mParser.java 5952 2024-10-21 11:41:05Z carsten $
*/

package de.cm.osm2po.plugins.parser;

import static de.cm.osm2po.misc.Utils.UTF8;
import static de.cm.osm2po.primitives.VarNum.readSigned;
import static de.cm.osm2po.primitives.VarNum.readUnsigned;
import static java.lang.Integer.toHexString;

import java.io.InputStream;
import java.util.Arrays;
import java.util.Date;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.parser.OsmParser;
import de.cm.osm2po.converter.parser.OsmParserHandler;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamAdapter;
import de.cm.osm2po.primitives.VarString;

public class OsmO5mParser implements OsmParser {

    private final static String FORMAT = "o5m2";

    private final static byte START      = (byte) 0xff;
    private final static byte RESET      = (byte) 0xff;
    private final static byte END        = (byte) 0xfe;
    private final static byte NODE       = (byte) 0x10;
    private final static byte WAY        = (byte) 0x11;
    private final static byte RELATION   = (byte) 0x12;
    private final static byte BBOX       = (byte) 0xdb;
    private final static byte TIMESTAMP  = (byte) 0xdc;
    private final static byte HEADER     = (byte) 0xe0;
    private final static byte SYNC       = (byte) 0xee;
    private final static byte JUMP       = (byte) 0xef;

    // Relation Role-Def for:
    private final static byte REL_NODE = 48; // '0'
    private final static byte REL_WAY  = 49; // '1'
    private final static byte REL_REL  = 50; // '2'

    private InStream inStream;
    private OsmParserHandler handler;
    private Log log;

    private static byte[] buffer = new byte[0x800];

    private byte[] pairBytesKey; // Pointer to StringPair key
    private byte[] pairBytesVal; // Pointer to StringPair val

    private byte[][] pairCacheKeys = new byte[15000][];
    private byte[][] pairCacheVals = new byte[15000][];
    private int pairIdx;

    private long deltaNodeId;
    private long deltaNodeLat;
    private long deltaNodeLon;
    private long deltaTimeStamp;
    private long deltaChangeSet;
    private long deltaWayId;
    private long deltaWayNodeRef;
    private long deltaRelationId;
    private long deltaRelationNodeRef;
    private long deltaRelationWayRef;
    private long deltaRelationRelationRef;

    private byte PREVTYPE;

    private boolean arbitrary;

    @Override
    public void open(InputStream inputStream,
            OsmParserHandler osmParserHandler, Config config) {
        this.inStream = new InStreamAdapter(inputStream);
        this.handler = osmParserHandler;
        this.log = config.getLog();

        this.arbitrary = Boolean.valueOf(
                config.getProperty("parser.arbitrary"));
    }

    @Override
    public void close() {
        this.inStream.close();
    }

    @Override
    public void parse() {
        if (inStream.readByte() != START)
            throw new RuntimeException("Start Byte " + START + " expected");

        while (!inStream.isEof()) {
            byte type = inStream.readByte();
            switch (type) {
            case NODE: this.parseNode(); break;
            case WAY: this.parseWay(); break;
            case RELATION: this.parseRelation(); break;
            case HEADER: this.parseHeader(); break;
            case TIMESTAMP: this.parseTimeStamp(); break;
            case BBOX: this.parseBoundingBox(); break;
            case RESET: this.reset(); break;
            case JUMP: this.jump(); break;
            case SYNC: this.sync(); break;
            case END: this.end(); break;
            default:
                log.warn("Unknown Type 0x" + toHexString(type & 0xff));
                this.skipBytes();
            }
            PREVTYPE = type;
        }
    }

    private void reset() {
        log.debug("Reset: PrevType was 0x"
                + Integer.toHexString(PREVTYPE & 0xFF));

        this.pairIdx = 0;

        this.deltaNodeId = 0L;
        this.deltaNodeLat = 0L;
        this.deltaNodeLon = 0L;
        this.deltaTimeStamp = 0L;
        this.deltaChangeSet = 0L;
        this.deltaWayId = 0L;
        this.deltaWayNodeRef = 0L;
        this.deltaRelationId = 0L;
        this.deltaRelationNodeRef = 0L;
        this.deltaRelationWayRef = 0L;
        this.deltaRelationRelationRef = 0L;

        if (NODE == PREVTYPE) {
            if (!this.arbitrary)
                this.handler.closeNodeProcessing();
        } else if (WAY == PREVTYPE) {
            if (!this.arbitrary)
                this.handler.closeWayProcessing();
        }

    }

    private void end() {
        if (PREVTYPE == RELATION) {
            if (!this.arbitrary)
                this.handler.closeRelationProcessing();
        }
    }

    private void parseStringPairFragment(boolean forRelRef) {
        long ref = readUnsigned(this.inStream);
        if (0L == ref) {
            if (forRelRef) {
                byte type = this.inStream.readByte();
                this.pairBytesKey = new byte[] {type};
            } else {
                this.pairBytesKey = readZeroDelimeted(inStream);
            }
            this.pairBytesVal = readZeroDelimeted(inStream);
            if (this.pairBytesKey.length + this.pairBytesVal.length > 250) {
                if (log.isEnabled(Level.DATA)) log.data("Skipping StringPair\n"
                        + new VarString(this.pairBytesKey) + "=" + new VarString(this.pairBytesVal));
            } else {
                if (this.pairIdx == 15000) this.pairIdx = 0;
                this.pairCacheKeys[this.pairIdx] = this.pairBytesKey;
                this.pairCacheVals[this.pairIdx] = this.pairBytesVal;
                this.pairIdx++;
            }
        } else {
            int idx = (int) (this.pairIdx - ref);
            if (idx < 0) idx += 15000;
            this.pairBytesKey = this.pairCacheKeys[idx];
            this.pairBytesVal = this.pairCacheVals[idx];
        }
    }

    private void parseTimeStamp() {
        long len = readUnsigned(this.inStream); // Length, but why?
        long pos = this.inStream.getReadIndex();
        if (len > 0) {
            long ts = readSigned(this.inStream); // in seconds since 1970
            log.debug("File TimeStamp: " + new Date(ts * 1000));
            if (this.inStream.getReadIndex() - pos != len) {
                throw new RuntimeException("TimeStamp size does not match");
            }
        }
    }

    /**
     * Parsed Version-, TimeStamp-, ChangeSet-, Author-Fragment
     * fuer jede Entity (Node, Way, Relation).
     */
    private void parseVersionFragment() {
        long version = readUnsigned(this.inStream);
        if (version != 0L) {
            this.deltaTimeStamp += readSigned(this.inStream);
            if (this.deltaTimeStamp != 0L) {
                this.deltaChangeSet += readSigned(this.inStream);
                this.parseStringPairFragment(false); // UID (StringBytes as Number), User
            }
        }
    }

    private void parseHeader() {
        VarString header = new VarString().readFromStream(this.inStream);
        log.debug("Header: " + header);
        if (!FORMAT.equals(header.toString()))
            throw new RuntimeException(FORMAT + " expected in Header");
    }

    private void parseBoundingBox() {
        long len = readUnsigned(this.inStream);
        if (len > 0) {
            double x1 = readSigned(this.inStream) / 1e7;
            double y1 = readSigned(this.inStream) / 1e7;
            double x2 = readSigned(this.inStream) / 1e7;
            double y2 = readSigned(this.inStream) / 1e7;
            log.debug("BoundingBox: " + new LatLon(y1, x1) + "-" + new LatLon(y2, x2));
        }
    }

    private void skipBytes() {
        long len = readUnsigned(this.inStream);
        log.warn("Skipping " + len + " bytes");
        for (int i = 0; i < len; i++) {
            this.inStream.readByte();
        }
    }

    private void sync() {
        log.warn("Sync not supported");
        this.skipBytes();
    }

    private void jump() {
        log.warn("Jump not supported");
        this.skipBytes();
    }

    private void parseNode() {
        long length = readUnsigned(this.inStream);
        long lastIndex = this.inStream.getReadIndex() + length;

        this.deltaNodeId += readSigned(this.inStream);

        this.parseVersionFragment();

        this.deltaNodeLon += readSigned(this.inStream);
        this.deltaNodeLat += readSigned(this.inStream);

        double lon = this.deltaNodeLon / 1e7;
        double lat = this.deltaNodeLat / 1e7;

        this.handler.onNode(this.deltaNodeId, lat, lon);

        // Node-Tags
        while (this.inStream.getReadIndex() < lastIndex) {
            this.parseStringPairFragment(false);
            String key = new String(this.pairBytesKey, UTF8);
            String val = new String(this.pairBytesVal, UTF8);
            this.handler.onNodeTag(key, val);
        }

        this.handler.onNodeComplete();
    }

    private void parseWay() {
        long length = readUnsigned(this.inStream);
        long lastIndex = this.inStream.getReadIndex() + length;

        this.deltaWayId += readSigned(this.inStream);
        this.handler.onWay(this.deltaWayId);

        this.parseVersionFragment();

        long refsLength = readUnsigned(this.inStream);
        long lastRefIndex = this.inStream.getReadIndex() + refsLength;

        // Way-Refs
        while(this.inStream.getReadIndex() < lastRefIndex) {
            this.deltaWayNodeRef += readSigned(this.inStream);
            this.handler.onWayNdRef(this.deltaWayNodeRef);
        }

        // Way-Tags
        while (this.inStream.getReadIndex() < lastIndex) {
            this.parseStringPairFragment(false);
            String key = new String(this.pairBytesKey, UTF8);
            String val = new String(this.pairBytesVal, UTF8);
            this.handler.onWayTag(key, val);
        }

        this.handler.onWayComplete();
    }


    private void parseRelation() {
        long length = readUnsigned(this.inStream);
        long lastIndex = this.inStream.getReadIndex() + length;

        this.deltaRelationId += readSigned(this.inStream);
        this.handler.onRelation(this.deltaRelationId);

        this.parseVersionFragment();

        long refsLength = readUnsigned(this.inStream);
        long lastRefIndex = this.inStream.getReadIndex() + refsLength;

        // Relation-Refs
        while(this.inStream.getReadIndex() < lastRefIndex) {
            long deltaRef = readSigned(this.inStream);
            this.parseStringPairFragment(true);
            String role = new String(this.pairBytesVal, UTF8);
            if (REL_NODE == this.pairBytesKey[0]) {
                this.deltaRelationNodeRef += deltaRef;
                this.handler.onRelationMember("node", this.deltaRelationNodeRef, role);
            } else if (REL_WAY == this.pairBytesKey[0]) {
                this.deltaRelationWayRef += deltaRef;
                this.handler.onRelationMember("way", this.deltaRelationWayRef, role);
            } else if (REL_REL == this.pairBytesKey[0]) {
                this.deltaRelationRelationRef += deltaRef;
                this.handler.onRelationMember("relation", this.deltaRelationRelationRef, role);
            } else {
                throw new IllegalStateException(
                        "Unknown Relation RefTyp " + this.pairBytesKey);
            }
        }

        // Relation-Tags
        while (this.inStream.getReadIndex() < lastIndex) {
            this.parseStringPairFragment(false);
            this.handler.onRelationTag(
                    this.pairBytesKey.toString(), this.pairBytesVal.toString());
        }

        this.handler.onRelationComplete();
    }


    /**
     * Liest einen C-Style-ByteArrayString aus einem Stream.
     * @param inStream {@link InStream}
     * @return Array of Byte, ohne abschliessendes Byte 0.
     */
    private static synchronized byte[] readZeroDelimeted(InStream inStream) {
        int n = 0;
        for (byte b = inStream.readByte(); b != 0; b = inStream.readByte()) {
            if (buffer.length == n)
                buffer = Arrays.copyOf(buffer, buffer.length * 2);
            buffer[n++] = b;
        }
        return Arrays.copyOf(buffer, n);
    }


}
