/*
 * $Id: MlgExtensionBuilder.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2013 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;
import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.routing.Graph.SUPPORT_EDGEFLAGS;
import static de.cm.osm2po.routing.Graph.SUPPORT_LATLON;

import java.io.File;
import java.util.Arrays;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.GraphBuilder;
import de.cm.osm2po.extensions.MlgExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.primitives.IntIntBag;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.GraphHeader;

/**
 * Liest die Graph-Datei und erzeugt Multilevels.
 * Die Levels werden als {@link MlgExtension} an die Datei angehaengt.
 *
 * @author (c) 2013 - Carsten Moeller - info@osm2po.de
 */
public class MlgExtensionBuilder implements PostProcessor {
    /*
     * Eine interessante Beobachtung ist, dass mehr als 3 Levels keine
     * bedeutende Verbesserung bringen. Im Gegenteil:
     * Mittlere Levels verschwinden fast gaenzlich oder mutieren zu Fragmenten.
     * Von dieser Beobachtung ist auch der Locality-Filter betroffen,
     * der im Routing eher linear anstatt quadratisch zu funktionieren
     * braucht. Voraussetzung ist lediglich, dass das kleinste Level
     * gross genug ist, mindestens so um die 0.2 Grad.
     */

    // TODO findExits(...) arbeitet vertex-basiert. Besser: Geometriepunkte betrachten.

    public final static byte ID_MASK_BASE = 0;
    public final static byte ID_MASK_FIND_SHORTEST = 1;
    public final static byte ID_MASK_IGNORE_ONEWAYS = 2;
    public final static byte ID_MASK_IGNORE_RESTRICTIONS = 4;

    private Log log;
    private Graph graph;
    private IntIntBag cellVertexBag;
    private int[] vertexCells;
    private byte[] edgeLevels;
    private short nLevels;
    private float maxCellSize;

    @Override
    public void run(Config config, int index) throws Exception {
        this.log = config.getLog();
        File graphFile = config.findGraphFile(true);
        if (null == graphFile) return;

        String maxLevelKey = Config.POSTP(index) + ".maxLevel";
        String[] rasts = config.getProperty(maxLevelKey, "3,1.0").split(",");
        this.nLevels = Short.valueOf(rasts[0].trim());
        this.maxCellSize = Float.valueOf(rasts[1].trim());

        String idKey = Config.POSTP(index) + ".id";
        String strId = config.getProperty(idKey, "0");
        byte id = Byte.parseByte(strId);

        String mli = MlgExtension.class.getSimpleName() + "#" + id;
        log.info("Building MultiLevels " + Arrays.toString(rasts)
                + " " + mli + " for " + graphFile.getName());

        if (new GraphHeader(graphFile).isExtensionExist(MlgExtension.class, id)) {
            log.error("Extension " + mli + " already exists");
            return;
        }

        boolean findShortest = (id & ID_MASK_FIND_SHORTEST) != 0;
        boolean ignoreOneWays = (id & ID_MASK_IGNORE_ONEWAYS) != 0;
        boolean ignoreRestrictions = (id & ID_MASK_IGNORE_RESTRICTIONS) != 0;

        this.graph = new Graph(graphFile, log,
                SUPPORT_LATLON | SUPPORT_EDGEFLAGS, false, config.getProps());

        this.cellVertexBag = new IntIntBag();
        this.vertexCells = new int[graph.getEdgeEntries().length];
        this.edgeLevels = new byte[graph.getEdgeIds().length];

        for (byte level = 0; level < this.nLevels; level++)
            this.runLevel(level, findShortest, ignoreOneWays, ignoreRestrictions);

        this.graph.close();

        log.info("Writing MultiLevel-Extension to " + graphFile.getName());

        MlgExtension extension = new MlgExtension(
                (byte)this.nLevels, this.maxCellSize, this.edgeLevels);
        GraphBuilder.addExtension(graphFile, extension, id);

        if (log.isEnabled(Level.DEBUG)) logStatistics();

        this.edgeLevels = null;
        this.vertexCells = null;
        this.cellVertexBag.clear();
    }

    private void logStatistics() {
        int[] ls = new int[this.nLevels +1];
        for (int i = 0; i < this.edgeLevels.length; i++) {
            ls[this.edgeLevels[i]]++;
        }
        String s = "Distribution of " + this.nLevels + " levels:";
        for (int i = 0; i < ls.length; i++) {
            s += "\n[" + i + "] : " + DF(ls[i]) + " edges";
        }
        log.debug(s);
    }

    private void runLevel(byte level,
            boolean findShortest, boolean ignoreOneWays, boolean ignoreRestrictions) {

        MlgExtensionBuilderRouter router = new MlgExtensionBuilderRouter();

        this.createRaster(level);

        int nAllExits = 0;

        int[] cellIds = intArrayDistinct(this.cellVertexBag.getKeys().toPrimitive());
        int nCells = cellIds.length;
        log.info(DF(this.cellVertexBag.getSize())
                + " Vertices rastered into " + nCells + " cells");

        for (int c = 0; c < nCells; c++) {
            int cellId = cellIds[c];
            int[] exitVertexIds = findExits(cellId);
            int nExits = exitVertexIds.length;
            nAllExits += nExits;
            log.progress("Level " + (level + 1) + ": Cell " + DF(c + 1)
                    + " of " + DF(nCells) + " #" + cellId + " :"
                    + DF(nExits) + " Exits");

            for (int s = 0; s < nExits; s++) {
                int sourceId = exitVertexIds[s];
                router.reset();
                router.traverse(this.graph, cellId, this.vertexCells, level,
                        this.edgeLevels, sourceId, findShortest, ignoreOneWays, ignoreRestrictions);

                for (int t = 0; t < nExits; t++) {
                    if (t != s) {
                        int targetId = exitVertexIds[t];
                        if (router.isVisited(targetId)) {
                            int[] path = router.makePath(targetId);
                            for (int p = 0; p < path.length; p++) {
                                int edgeIdx = path[p];
                                this.edgeLevels[edgeIdx] = (byte) (level +1);
                            }
                        }
                    }
                }
            }
        }

        log.info("Level " + (level + 1) + ": " + DF(nAllExits) + " exits found for "
                + DF(graph.getEdgeIds().length) + " Edges");
    }

    /**
     * Findet angrenzende Vertices eines Gebiets, die nicht
     * in diesem liegen, jedoch mit einem Hop (einer Edge)
     * in dieses hinein- oder aus diesem herausfuehren.
     *
     * @param cellId int ID des Gebietes.
     * @return int[]-Array von vertices.
     */
    private int[] findExits(int cellId) {
        int n = 0;
        // Sammle alle Vertices dieser Cell
        int[] vertexIds = this.cellVertexBag.getAll(cellId);
        int[] exitVertexIds = new int[1000];
        if (vertexIds != null) { // In dieser Cell sind keine Vertices

            int[] edgeTargetIds = this.graph.getEdgeTargetIds();
            int[] edgeSourceIds = this.graph.getEdgeSourceIds();

            for (int vx = 0; vx < vertexIds.length; vx++) {
                int vertexId = vertexIds[vx];
                int[] outEdgeIdxs = graph.findOutgoingEdges(vertexId);
                for (int ex = 0; ex < outEdgeIdxs.length; ex++) {
                    int edgeIdx = outEdgeIdxs[ex];
                    int targetId = edgeTargetIds[edgeIdx];
                    // Fuehrt ausgehende Edge in ein anderes Gebiet?
                    if (this.vertexCells[targetId] != cellId) {
                        if (exitVertexIds.length == n)
                            exitVertexIds = Arrays.copyOf(exitVertexIds, exitVertexIds.length * 2);
                        exitVertexIds[n++] = targetId; // dann TargetId merken!
                    }
                }

                int[] inEdgeIdxs = graph.findIncomingEdges(vertexId);
                for (int ex = 0; ex < inEdgeIdxs.length; ex++) {
                    int edgeIdx = inEdgeIdxs[ex];
                    int sourceId = edgeSourceIds[edgeIdx];
                    // Stammt eingehende Edge aus anderem Gebiet?
                    if (this.vertexCells[sourceId] != cellId) {
                        if (exitVertexIds.length == n)
                            exitVertexIds = Arrays.copyOf(exitVertexIds, exitVertexIds.length * 2);
                        exitVertexIds[n++] = sourceId; // dann SourceId merken!
                    }
                }
            }
        }
        return intArrayDistinct(exitVertexIds);
    }

    private int calcCellId(double lat, double lon, double cellSize) {
        double y = StrictMath.floor((lat + 90) / cellSize);
        double x = StrictMath.floor((lon + 180) / cellSize);
        int ilat = (int) y;
        int ilon = (int) x;
        return ilat << 16 | ilon;
    }

    private void createRaster(byte level) {
        this.cellVertexBag.clear();

        float[] lats = this.graph.getLats();
        float[] lons = this.graph.getLons();

        float cellSize = this.maxCellSize;
        cellSize /= Math.pow(2, this.nLevels - level -1);
        log.info("Rastering Level " + (level + 1) + " (" + cellSize + ")");

        int nVertices = this.vertexCells.length;
        for (int vertexId = 1; vertexId < nVertices; vertexId++) {
            int cellId = calcCellId(lats[vertexId], lons[vertexId], cellSize);

            this.cellVertexBag.add(cellId, vertexId);
            this.vertexCells[vertexId] = cellId;

            if ((vertexId % 1000000) == 0) log.progress(
                    DF(vertexId) + " of " + DF(nVertices - 1) + " vertices rastered.");
        }

        log.progress("Sorting");
        this.cellVertexBag.sort();
    }

    public byte[] getEdgeLevels() {
        return this.edgeLevels;
    }

}
