/*
 * $Id: SndExtensionBuilder.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.converter.GraphBuilder.addExtension;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;
import static de.cm.osm2po.routing.Graph.SUPPORT_INMEMORY;
import static de.cm.osm2po.routing.Graph.isReverse;

import java.io.File;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.SndExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.primitives.Dictionary;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.GraphHeader;

/**
 * Startet die {@link Graph}-Datei und generiert aus den Strassennamen
 * ein {@link Dictionary} fuer die Rueckwaertssuche als Extension.
 * @author (c) 2016 - Carsten Moeller, info@osm2po.de
 */
public class SndExtensionBuilder implements PostProcessor {
    public final static int COMPAT_VERSION = 1;

    private Log log;

    @Override
    public void run(Config config, int index)
    throws Exception {
        this.log = config.getLog();
        File workDir = config.getWorkDir();
        File graphFile = null;

        final String prep = Config.POSTP(index) + ".";

        final boolean isolate = Boolean.valueOf( // Nicht in Graph schreiben
                config.getProperty(prep + "isolate"));
        final boolean keep = Boolean.valueOf( // DicDatei aufbewahren
                config.getProperty(prep + "keep"));
        final String extFile = config.getProperty(prep + "file");
        final byte id = Byte.valueOf(config.getProperty(prep + "extensionId", "0"));

        if (isolate) {
            log.info("Building isolated " + this.getClass().getSimpleName());

        } else {
            graphFile = config.findGraphFile(true);
            if (null == graphFile) return;

            if (new GraphHeader(graphFile).isExtensionExist(SndExtension.class, (byte)0)) {
                log.error("Extension " + SndExtension.class.getSimpleName() + " already exists");
                return;
            }
        }

        if (extFile != null) {
            File dicFile = new File(extFile);
            this.addToGraphFile(dicFile, graphFile, id);
            return;
        }

        if (graphFile != null)
            log.info("Loading " + graphFile.getAbsoluteFile());

        Level oldLogLevel = log.getLogLevel();
        log.setLogLevel(Level.ERROR);
        Graph graph = new Graph(graphFile, log, SUPPORT_INMEMORY , false, null);
        log.setLogLevel(oldLogLevel);

        // Da wir spaeter mehrere Dateipositionen im Vorraus kennen muessen,
        // wird hier nochmals eine finale Temp-Datei gebaut.
        File dicFile = new File(workDir, config.getPrefix() + "_2po_snd.ext");
        OutStream dicOs = new OutStreamDisk(dicFile);

        // Headerdaten hier noch nicht ermittelt, daher mit 0 fuellen
        dicOs.writeInt(0); // version
        dicOs.writeLong(0); // DictionaryPos
        dicOs.writeLong(0); // eofPos (size in bytes)

        long dicPos = dicOs.getWriteIndex();

        log.info("Building Dictionary for street names");
        Dictionary edgeNamesDic = new Dictionary();
        edgeNamesDic.create(workDir, log);
        int nEdges = graph.getEdgeIds().length;
        int[] edgeIds = graph.getEdgeIds();
        for (int edgeIdx = 0; edgeIdx < nEdges; edgeIdx++) {
            if (isReverse(edgeIds[edgeIdx])) continue; // forward edges only
            VarString edgeName = graph.lookupEdgeName(edgeIdx);
            if (!VarString.EMPTY.equals(edgeName))
                edgeNamesDic.add(edgeIdx, edgeName.toString().toLowerCase());
        }
        edgeNamesDic.build(dicOs);

        long eofPos = dicOs.getWriteIndex(); // = size

        dicOs.close();
        graph.close();

        // Headerdaten sind jetzt bekannt. Ueberschreiben.
        RandomAccessDisk bndRad = new RandomAccessDisk(dicFile, false);
        bndRad.setWriteIndex(0);
        bndRad.writeInt(COMPAT_VERSION);
        bndRad.writeLong(dicPos);
        bndRad.writeLong(eofPos);
        bndRad.close();

        // Jetzt kann Extension geschrieben werden
        if (!isolate) this.addToGraphFile(dicFile, graphFile, id);
        if (!(isolate || keep)) dicFile.delete();
    }

    private void addToGraphFile(File extensionFile, File graphFile, byte id) throws Exception {
        long size = extensionFile.length();
        byte varType = typeIdOf(SndExtension.class);

        log.info("Writing " + SndExtension.class.getSimpleName() + "#" + id
                + " from " + extensionFile.getName()
                + " to " + graphFile.getName()
                + " [" + (DF2((double)size / 0x100000) + "M]"));

        InStream inStream = new InStreamDisk(extensionFile);
        addExtension(graphFile, varType, id, size, inStream);
        inStream.close();
    }

}
