/*
 * $Id: JsonDeserializer.java 5930 2024-10-11 15:31:47Z carsten $
 *
 * Copyright 2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.reflection;
import static de.cm.osm2po.misc.Utils.isRecord;
import static de.cm.osm2po.reflection.Reflector.getGetter;
import static de.cm.osm2po.reflection.Reflector.getSetter;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Parameter;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.function.Function;

import de.cm.osm2po.reflection.Reflector.Discriminator;
import de.cm.osm2po.reflection.Reflector.Discriminators;
import de.cm.osm2po.reflection.Reflector.FieldGetter;
import de.cm.osm2po.reflection.Reflector.Getter;
import de.cm.osm2po.reflection.Reflector.Reflectable;
import de.cm.osm2po.reflection.Reflector.Setter;

/**
 * Ein simpler oder gar nicht mal so simpler Json-Deserializer.<br>
 * Implementiert mit vergleichbar wenigen Zeilen Code und dadurch schnell.<br>
 * Sehr gut geeignet fuer POJOS, jedoch weniger geeignet, wenn Klassen
 * unendliche Kaskaden von komplexen Generics aufweisen.<br>
 * Grundsaetzlich werden Generics aber unterstuetzt.<br><br>
 * Die Grundlage bildet der {@link Reflector}, der standardmaessig
 * alle <code>public</code> Members einer Klasse akzeptiert.<br>
 * <strong> Statics, Privates, Methods</strong>, usw.
 * muessen explizit mit {@link Reflectable} annotiert sein.<br>
 * Polymorphy laesst sich mittels {@link Discriminators} abbilden.<br>
 * Interfaces werden aktuell nicht unterstuetzt.<br>
 * Die Super-Klasse sollte folglich konkret oder abstrakt sein.<br>
 * Beispiel einer Base-Class-Annotation:
 * <pre>
 * &#64;Discriminators(property = "dtype", value =  {
 *   &#64;Discriminator(type = "c1", clazz = ChildClass1.class),
 *   &#64;Discriminator(type = "c2", clazz = ChildClass2.class)
 * })
 * </pre>
 * und im JSON dann folglich:<br>
 * <code>{"dtype":"c1", "key1":"value1", ... usw.}</code>
 * <br><br>
 * Konzeptionell neu hier ist, dass auch ein Filter gesetzt werden kann.
 * Dieser bewirkt, dass nur in benannten Kontexten gewisse
 * Attribute beruecksichtigt werden. Dieser Filter arbeitet binaer.
 * Ist der Wert 0 (default), so wird er ignoriert.
 *
 * @see #deserialize(String)
 * @see #deserialize(String, Class)
 * @author (c) 2021/2023 - Carsten Moeller, info@osm2po.de
 */
public final class JsonDeserializer {

    private JsonDeserializer() {/*STATIC*/}

    public static Object deserialize(final String json) {
        return deserialize(json, null);
    }

    public static <T> T deserialize(final String json, Class<T> clazz) {
        return deserialize(json, clazz, 0L);
    }

    public static Object deserialize(final String json, Type type) {
        return deserialize(json, type, 0L);
    }

    @SuppressWarnings("unchecked")
    public static <T> T deserialize(final String json, Class<T> clazz, long filter) {
        Type type = clazz; // dont call myself
        return (T) deserialize(json, type, filter);
    }

    public static Object deserialize(final String json, Type type, long filter) {
        if (null == json) return null;
        try {return parseValue(json.toCharArray(), new int[]{0}, type, filter);}
        catch (Exception e) {throw new IllegalArgumentException(e);}

    }

    /**
     * Simple (Marker-)String-Klasse, die nicht verarbeitet wird und so selbst einen
     * JSON beinhalten kann.
     */
    public static final class JsonString implements CharSequence {
        private final String string;
        public JsonString(String string) {this.string = string;}
        @Override public int length() {return this.string.length();}
        @Override public char charAt(int index) {return this.string.charAt(index);}
        @Override public CharSequence subSequence(int start, int end) {return this.string.substring(start, end);}
        @Override public String toString() {return this.string;}
    }

    private static Map<Class<?>, Function<String, Object>> converterMap = new HashMap<>();
    static {
        converterMap.put(Integer.class, Integer::decode);
        converterMap.put(Long.class, Long::decode);
        converterMap.put(Short.class, Short::decode);
        converterMap.put(Byte.class, Byte::decode);
        converterMap.put(Character.class, s->Character.valueOf(s.charAt(0)));
        converterMap.put(Double.class, Double::valueOf);
        converterMap.put(Float.class, Float::valueOf);
        converterMap.put(Boolean.class, Boolean::valueOf);

        converterMap.put(Integer.TYPE, Integer::decode);
        converterMap.put(Long.TYPE, Long::decode);
        converterMap.put(Short.TYPE, Short::decode);
        converterMap.put(Byte.TYPE, Byte::decode);
        converterMap.put(Character.TYPE, s->Character.valueOf(s.charAt(0)));
        converterMap.put(Double.TYPE, Double::valueOf);
        converterMap.put(Float.TYPE, Float::valueOf);
        converterMap.put(Boolean.TYPE, Boolean::valueOf);

        converterMap.put(BigInteger.class, BigInteger::new);
        converterMap.put(BigDecimal.class, BigDecimal::new);

        converterMap.put(OffsetDateTime.class, s->OffsetDateTime.parse(s));
        converterMap.put(ZonedDateTime.class, ZonedDateTime::parse);
        converterMap.put(LocalDateTime.class, s->ZonedDateTime.parse(s).toLocalDateTime());
        converterMap.put(LocalDate.class, s->ZonedDateTime.parse(s).toLocalDate());
        converterMap.put(Date.class, s->Date.from(ZonedDateTime.parse(s).toInstant()));
        converterMap.put(GregorianCalendar.class, s->GregorianCalendar.from(ZonedDateTime.parse(s)));
        converterMap.put(Calendar.class, s->{
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(Date.from(ZonedDateTime.parse(s).toInstant()));
            return calendar;
        });
    }

    @SuppressWarnings("unchecked")
    private static <T> T valueOf(final String value, Class<T> targetClass) {
        Function<String, Object> function = converterMap.get(targetClass);
        if (null == function) return null;
        return (T) function.apply(value);
    }

    // Skips all whitespaces and leaves position
    // on next valid and parsable character
    private static void skipWhiteSpace(char[] carr, int[] x) {
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case ' ': case '\n': case '\r' : case '\t': break; // whitespace
            default: x[0]--; return;
            }
        }
    }

    // Skip next Value. May be {}, [], "" or null or Number
    // x-Postion must be before or directly on this opening character
    private static int skipValue(char[] carr, int[] x) {
        skipWhiteSpace(carr, x);
        final int startPos = x[0];
        char openChar = carr[x[0]++];
        char closeChar = 0;
        int level = 1;
        if ('\"' == openChar) closeChar = '\"';
        else if ('{' == openChar) closeChar = '}';
        else if ('[' == openChar) closeChar = ']';
        else {
            openChar = 0;
            level = 0; // e.g. Number or null
        }

        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            if ('\\' == c) { // escape char
                c = carr[x[0]++];
                continue; // ignore this and next char
            }
            if (closeChar == c) level--;
            else if (openChar == c) level++;

            if (0 == level) {
                if (c == ',' || c == '\"' || c == ']' || c == '}') {
                    if (0 == closeChar) {
                        x[0]--; // eg. ends with foo:null}
                    }
                    break;
                }
            }
        }

        int nSkipped = x[0] - startPos;
        return nSkipped;
    }

    // x-position must be before or directly on leading "
    // leaves position on next valid non-whitespace character after :
    // in case } is read, null will be returned
    private static String parseKey(final char[] carr, final int[] x) {
        skipWhiteSpace(carr, x);
        char c = carr[x[0]];
        if ('}' == c)  return null;
        if (c != '\"') throw new RuntimeException(
                "'\"' expected at position " + x[0]);
        x[0]++;
        final String key = parseString(carr, x);
        skipWhiteSpace(carr, x);
        c = carr[x[0]++];
        if (c != ':') throw new RuntimeException(
                "':' expected at position " + x[0]);
        skipWhiteSpace(carr, x);
        return key;
    }

    // Deserialisiert den naechste Wert und delegiert dann aufgrund
    // von {[" oder keinem der drei Start-Zeichen
    // an die entsprechenden Funktionen
    private static Object parseValue(
            final char[] carr, final int[] x, final Type T, long filter) throws Exception {

        Class<?> targetClass = null;
        Type[] generics = null;
        if (T instanceof ParameterizedType) {
            ParameterizedType PT = (ParameterizedType) T;
            targetClass = (Class<?>) PT.getRawType(); // kann auch Interface sein
            generics = PT.getActualTypeArguments();

        } else {
            targetClass = (Class<?>) T;
        }

        if (JsonString.class == targetClass) {
            final int nSkipped = skipValue(carr, x);
            final String json = String.copyValueOf(carr, x[0]- nSkipped, nSkipped);
            final Object value = new JsonString(json);
            return value;
        }

        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '{': // Map, Properties oder Custom-Object

                // Wenn unbekannt, dann Map (Interface!) annehmen
                // Impl wird im Nachgang gesetzt
                if (null == targetClass || Object.class == targetClass) {
                    targetClass = Map.class;
                }

                // Wenn Map aber auch Properties (ebenfalls Map)
                if (Map.class.isAssignableFrom(targetClass)) {

                    // Sonderlocke: Properties sind explizite String-String-Maps
                    if (Properties.class == targetClass) {
                        generics = new Type[] {String.class, String.class};

                    } else if (Map.class == targetClass) { // und Map-Interfaces benoetigen Impl.
                        targetClass = HashMap.class;
                    }
                    if (null == generics) {
                        generics = new Type[] {String.class, Object.class};
                    }

                    final Class<?> keyClass = (Class<?>) generics[0];
                    final Class<?> valueClass = (Class<?>) generics[1];

                    if (keyClass != String.class) throw new RuntimeException(
                            "JSON-Keys must be Strings. Other types are currently not supported");

                    return parseMap(carr, x, targetClass, valueClass, filter);

                } else if (isRecord(targetClass)) {
                    return parseRecord(carr, x, targetClass);

                } else {
                    // Sonst alles andere wie Custom- Classes, Interfaces, AbstractClasses, usw.

                    // Wurde ein Diskrimator annotiert, dann Impl-Klasse ableiten.
                    Discriminators discriminators = targetClass.getAnnotation(Discriminators.class);
                    if (discriminators != null) {
                        // Henne-Ei-Problem, da Instanziierung vor Deserialisierung kommt.
                        // Worst-Case: Ist der Wert im json hinten, dann muss von hier ab der
                        // komplette String bis } oder Ende durchsucht werden.
                        String jsonType = findDiscriminatorValue(discriminators.property(), carr, x);
                        if (jsonType != null) for (Discriminator discriminator : discriminators.value()) {
                            if (discriminator.type().equals(jsonType)) {
                                if (targetClass.isAssignableFrom(discriminator.clazz())) {
                                    targetClass = discriminator.clazz(); // nur ueberschreiben, wenn gefunden
                                    break;
                                }
                            }
                        }
                    }

                    return parseCustomObject(carr, x, targetClass, filter);
                }

            case '[': // Sieht nach Array aus, dann schauen, ob primitiv oder Collection

                // Wenn unbekannt, dann Collection (Interface!) annehmen
                // Impl wird im Nachgang gesetzt
                if (null == targetClass || Object.class == targetClass) {
                    targetClass = Collection.class;
                }

                if (Collection.class.isAssignableFrom(targetClass)) {

                    if (targetClass.isInterface()) {
                        // Keine Impl, vielleicht Set, List oder Collection selbst
                        // TODO Abstrakte (Zwischen-)Klasse ebenfalls beruecksichtigen
                        // ggf. einfach auf existierenden Standard-Konstruktor pruefen
                        targetClass = ArrayList.class;
                    }
                    if (null == generics) {
                        generics = new Type[] {Object.class};
                    }

                    final Class<?> valueClass = (Class<?>) generics[0];
                    return parseCollection(carr, x, targetClass, valueClass, filter);

                } else if (targetClass.isArray()) {
                    return parsePrimitiveArray(carr, x, targetClass, filter);
                };

            case '\"': // Dann wohl String
                final String string = parseString(carr, x);
                if (String.class == targetClass) return string; // liefert ggf. auch "null"
                // liefert echte null, aber auch Date und andere Typen, wenn explizit gesetzt
                else return castToken(string, targetClass); // Trotzdem cast, da auch Datum moeglich

            case ' ': case '\n': case '\r' : case '\t': break; // whitespace != token - wichtig!

            default: // Ansonsten als TokenString lesen und dann z.B. in Zahl oder echt null wandeln
                x[0]--; // wieder auf 1. Zeichen des Tokens setzen
                final String token = tokenizeValue(carr, x);
                return castToken(token, targetClass);
            }
        }
        return null;
    }

    private static Object castToken(final String s, Class<?> targetClass) throws Exception {
        if ("null".equals(s)) return null; // !!! Vorsicht, wenn String null != "null"

        Object obj = valueOf(s, targetClass);
        if (obj != null) return obj;

        // Fallbacks
        if (String.class == targetClass || Object.class == targetClass) {
            return s.toString();
        }

        if (targetClass.isEnum()) {
            for (Object e : targetClass.getEnumConstants()) {
                if (e.toString().equals(s)) return e;
            }
        }

        throw new ClassCastException("Cannot cast '" + s + "' to Class "
                + (null == targetClass ? "null" : targetClass.getSimpleName()));
    }

    // x-Position must be one behind {
    private static String findDiscriminatorValue(
            final String discriminator, char[] carr, int[] x) throws Exception {
        int pos = x[0]; // Fuer Reset merken
        String dtype = null;
        // Read ahead
        while (x[0] < carr.length && null == dtype) {
            final String key = parseKey(carr, x); // Sollte hoffentlich vorne sein
            if (null == key) break;
            if (!discriminator.equals(key)) skipValue(carr, x);
            else dtype = (String) parseValue(carr, x, String.class, 0L);
            skipWhiteSpace(carr, x);
            if (',' != carr[x[0]++]) break;
        }

        x[0] = pos; // rescue
        return dtype;
    }


    // x-Position must be one behind "
    private static String parseString(char[] carr, int[] x) {
        StringBuilder sb = new StringBuilder();
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            if (c == '\"') break;
            if (c == '\\') { // reverse solidus = escape
                c = carr[x[0]++];
                switch (c) {
                case 'b': c = '\b'; break;
                case 'f': c = '\f'; break;
                case 'n': c = '\n'; break;
                case 'r': c = '\r'; break;
                case 't': c = '\t'; break;
                case 'u':
                    char[] hex = new char[4];
                    for (int i = 0; i < 4; i++) {
                        hex[i] = carr[x[0]++];
                    }
                    c = (char) Integer.parseInt(new String(hex), 16);
                }
            }
            sb.append(c);
        }
        return sb.toString();
    }

    // x-Position must be one behind {
    private static Object parseMap(char[] carr, int[] x,
            Class<?> targetClass, Class<?> valueClass, long filter) throws Exception {

        @SuppressWarnings("unchecked")
        Map<Object, Object> targetMap = (Map<Object, Object>) targetClass.getConstructor().newInstance();

        x[0]--; // reset Position to {
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '{': case ',':
                String key = parseKey(carr, x);
                if (key != null) { // check special case empty {}
                    Object value = parseValue(carr, x, valueClass, filter);
                    if (value != null) targetMap.put(key, value);
                }
                break;
            case '}': return targetMap;
//            case ' ': case '\n': case '\r' : case '\t': break; // whitespace
            }
        }
        return null;
    }

    // x-Position must be one behind {
    private static Object parseCustomObject(char[] carr, int[] x,
            Class<?> targetClass, long filter) throws Exception {

        Object targetObject = targetClass.getConstructor().newInstance();

        x[0]--; // reset Position to {
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '{': case ',':
                String key = parseKey(carr, x);
                if (key != null) { // check special case empty {}
                    Setter setter = getSetter(key, targetClass, filter);
                    if (null == setter) {
                        skipValue(carr, x);
                        break;
                    } else {
                        final Object value = parseValue(carr, x, setter.implType, filter);
                        setter.setValue(targetObject, value);
                    }
                }
                key = null; // reset
                break; // break switch/case but continue char-loop
            case '}': return targetObject;
            // interpret everything else as whitespace
            }
        }
        return null;
    }

    // x-Position must be one behind [
    private static Object parsePrimitiveArray (char[] carr, int[] x,
            Class<?> targetClass, long filter) throws Exception {

        Class<?> elmClass = targetClass.getComponentType();
        Object primitiveArray = Array.newInstance(elmClass, 0x400 /*1k*/);
        Object resizedArray = null;

        x[0]--; // reset Position to [
        int n = 0;
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '[': case ',':
                skipWhiteSpace(carr, x);
                if (']' ==  carr[x[0]] || ',' == carr[x[0]]) break; // Empty Value

                Object object = parseValue(carr, x, elmClass, filter);

                if (Array.getLength(primitiveArray) == n) { // ArrayOverflowHandling via Reflection
                    resizedArray = Array.newInstance(elmClass, n * 2);
                    System.arraycopy(primitiveArray, 0, resizedArray, 0, n);
                    primitiveArray = resizedArray;
                }
                Array.set(primitiveArray, n++, object); // auto-unwrappes primitive types
                break;

            case ']':
                resizedArray = Array.newInstance(elmClass, n);
                System.arraycopy(primitiveArray, 0, resizedArray, 0, n);
                primitiveArray = resizedArray;
                return primitiveArray;

                // interpret everything else as whitespace
            }
        }
        return null;
    }

    // x-Position must be one behind [
    private static Object parseCollection (char[] carr, int[] x,
        Class<?> targetClass, Class<?> elmClass, long filter) throws Exception {

        @SuppressWarnings("unchecked")
        Collection<Object> collection = (Collection<Object>) targetClass.getConstructor().newInstance();

        x[0]--; // reset Position to [
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '[': case ',':
                skipWhiteSpace(carr, x); // TODO Empty-Check schlauer loesen - auch fuer Array
                if (']' ==  carr[x[0]] || ',' == carr[x[0]]) break; // Empty Value

                Object object = parseValue(carr, x, elmClass, filter);
                collection.add(object);
                break;
            case ']': return collection;
            // interpret everything else as whitespace
            }
        }
        return null;
    }

    // x-Position must be one behind {
    private static Object parseRecord(char[] carr, int[] x,
            Class<?> targetClass) throws Exception {

        Map<String, Object> recMap = new HashMap<>();

        x[0]--; // reset Position to {
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '{': case ',':
                String key = parseKey(carr, x);
                if (key != null) { // check special case empty {}
                    Getter getter = getGetter(key, targetClass, 0);
                    if (getter instanceof FieldGetter) {
                        Field field = ((FieldGetter) getter).field;
                        Type type = field.getGenericType();
                        final Object value = parseValue(carr, x, type, 0);
                        recMap.put(key, value);

                    } else {
                        skipValue(carr, x);
                        break;
                    }
                }
                key = null; // reset
                break; // break switch/case but continue char-loop
            case '}':
                Constructor<?> ctor = targetClass.getDeclaredConstructors()[0];
                Object[] paramValues = new Object[ctor.getParameterCount()];
                int i = 0;
                for (Parameter p : ctor.getParameters()) {
                    paramValues[i++] = recMap.get(p.getName());
                }
                ctor.setAccessible(true); // for private record
                return ctor.newInstance(paramValues);
            }
            // interpret everything else as whitespace
        }
        return null;
    }


    // x-Position must be on first character
    private static String tokenizeValue(char[] carr, int[] x) { // Number or null
        StringBuilder sb = new StringBuilder();
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case ' ': case '\n': case '\r' : case '\t': break;
            case '}': case ']': case ',':
                x[0]--; // Important for Recursive call from parseArray/parseObject
                return sb.toString();
            default: sb.append(c);
            }
        }
        return sb.toString();
    }

}
