/*
 * $Id: JsonParser.java 5916 2024-07-26 19:22:17Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.reflection;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * Ein simpler Json-Parser, der Objekte lediglich
 * nach {@link HashMap} und {@link ArrayList} uebersetzt.<br>
 * Fuer komplexere Deserialisierung gegen eine Modell-Klasse ist der
 * {@link JsonDeserializer} zu bevorzugen. Dennoch ist die Einfachheit
 * dieser Variante in vielen Faellen ausreichend, wenn nicht gar
 * von Vorteil. Insbesondere dann, wenn aus sehr dynamischen oder
 * komplexen Strukturen lediglich kleine Teilmengen von
 * Interesse sind. Als Beispiel sei hier das JWT-Parsing genannt.
 * @author (c) 2019/2021 - Carsten Moeller, info@osm2po.de
 * @see #getValue(Object, String)
 */
public final class JsonParser {

    private JsonParser() {/*STATIC*/}

    public static Object parse(final String json) {
        return parse(json.toCharArray(), new int[]{0});
    }

    private static Object parse(char[] carr, int[] x) {
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '\"':
                String s = parseString(carr, x);
                if (s.length() > 18 && s.charAt(10) == 'T') try {
                    return OffsetDateTime.parse(s);} catch (Exception e) {}
                return s;
            case '{': return parseObject(carr, x);
            case '[': return parseArray(carr, x);
            case ' ': case '\n': case '\r' : case '\t': continue; // whitespace
            default: x[0]--; return parseValue(carr, x);
            }
        }
        return null;
    }

    private static String parseString(char[] carr, int[] x) {
        StringBuilder sb = new StringBuilder();
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            if (c == '\"') break;
            if (c == '\\') { // reverse solidus = escape
                c = carr[x[0]++];
                switch (c) {
                case 'b': c = '\b'; break;
                case 'f': c = '\f'; break;
                case 'n': c = '\n'; break;
                case 'r': c = '\r'; break;
                case 't': c = '\t'; break;
                case 'u':
                    char[] hex = new char[4];
                    for (int i = 0; i < 4; i++) {
                        hex[i] = carr[x[0]++]; // TODO check length
                    }
                    c = (char) Integer.parseInt(new String(hex), 16);
                }
            }
            sb.append(c);
        }

        return sb.toString();
    }

    private static HashMap<String, Object> parseObject(char[] carr, int[] x) {
        HashMap<String, Object> map = new HashMap<String, Object>();
        String key = null;
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case '\"': key = parseString(carr, x); break;
            case ':':  map.put(key, parse(carr, x)); key = null; break;
            case ',': break; // Skip! Used for old syntax where quotes for key were optional
            case '}': return map;
            case ' ': case '\n': case '\r' : case '\t': continue; // whitespace
            }
        }
        return null;
    }

    private static ArrayList<Object> parseArray(char[] carr, int[] x) {
        ArrayList<Object> arr = new ArrayList<Object>();
        while (x[0] < carr.length) {
            char c = carr[x[0]++];
            switch (c) {
            case ']': return arr;
            case ',': continue;
            case ' ': case '\n': case '\r' : case '\t': continue; // whitespace
            default: x[0]--; arr.add(parse(carr, x));
            }
        }
        return null;
    }

    private static Object parseValue(char[] carr, int[] x) {
        StringBuilder sb = new StringBuilder();
        boolean hasDot = false;
        while (x[0] <= carr.length) {
            char c = carr.length == x[0] ? 0 : carr[x[0]++];
            switch (c) {
            case '.': hasDot = true; break;
            case ' ': case '\n': case '\r' : case '\t': continue;
            case ',': case '}': case ']': case 0:
                x[0]--;
                String token = sb.toString();
                if ("null".equals(token)) return null;
                if ("".equals(token)) return null; // Wrong or rescue ok
                if ("true".equalsIgnoreCase(token)) return Boolean.TRUE;
                if ("false".equalsIgnoreCase(token)) return Boolean.FALSE;

                if (hasDot) return Double.parseDouble(token);
                else return Long.parseLong(token);
            }
            sb.append(c);
        }
        return null;
    }

    /**
     * Liefert aus einem mittels {@link #parse(String)} ermittelten
     * {@link Object} per Expression-Language einen Wert.
     * @param obj Result-Object von {@link #parse(String)}
     * @param el String Expression. Beispiel: <code>parent.children.3.name</code>
     * @return Wert des Elements im Baum als Object oder <code>null</code>
     */
    public static Object getValue(Object obj, String el) {
        if (null == obj) return null;
        if (null == el) return obj;

        Object object = obj;
        String[] tokens = el.split("\\.");
        for (int i = 0; i < tokens.length; i++) {
            // "" behandelt . am Ende oder Leerstring-Param statt null fuer el
            if ("".equals(tokens[i])) continue;

            if (object instanceof ArrayList) {
                ArrayList<?> arr = (ArrayList<?>) object;
                int x = Integer.valueOf(tokens[i]);
                object = (x >= 0 && x < arr.size()) ? arr.get(x) : null;
            } else if (object instanceof Map) {
                object = ((Map<?,?>)object).get(tokens[i]);
            } else {
                if (i < tokens.length) object = null;
                break;
            }
        }
        return object;
    }

}
