/*
 * $Id: Reflector.java 5930 2024-10-11 15:31:47Z carsten $
 *
 * Copyright 2023 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.reflection;

import static de.cm.osm2po.misc.Utils.isRecord;
import static de.cm.osm2po.reflection.Reflector.MemberAccess.Ignore;
import static de.cm.osm2po.reflection.Reflector.MemberAccess.ReadOnly;
import static de.cm.osm2po.reflection.Reflector.MemberAccess.ReadWrite;
import static de.cm.osm2po.reflection.Reflector.MemberAccess.WriteOnly;
import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.METHOD;
import static java.lang.annotation.ElementType.TYPE;
import static java.lang.annotation.RetentionPolicy.RUNTIME;
import static java.lang.reflect.Modifier.ABSTRACT;
import static java.lang.reflect.Modifier.INTERFACE;
import static java.lang.reflect.Modifier.isFinal;
import static java.lang.reflect.Modifier.isPublic;
import static java.lang.reflect.Modifier.isStatic;

import java.lang.annotation.Documented;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

/**
 * Erstellt {@link Getter}s und {@link Setter}s fuer eine Klasse
 * und erlaubt danach Zugriffe per Reflection auf deren Instanzen.<br>
 * Die Klasse muss dafuer konkret, also weder abstrakt noch generisch
 * sein. Urspruenglich wurde diese Klasse fuer den Einsatz eines
 * {@link JsonDeserializer}'s und {@link JsonBuilder}'s entwickelt,
 * sie kann jedoch auch fuer andere Zwecke eingesetzt werden.<br>
 * Der Zugriff erfolgt lazy, d.h. der erste Aufruf mittels z.B.
 * {@link #getGetter(String, Class, long)} triggert die Analyse
 * der Klasse und liegt dann im Cache.<br>
 * Default werden nur <code>public</code> {@link Field}s aufgenommen.
 * Eine Annotation mittels {@link Reflectable} kann jedoch auch
 * andere Varianten erzwingen. Dies gilt insbesondere fuer <code>statics</code>
 * und <code>methods</code>. Letztere werden zuerst analysiert.
 * D.h. ein <code>getX()</code> (wenn annotiert) wird immer
 * vor einem <code>public x</code> ausgewertet bzw. vermerkt.<br>
 * Des Weiteren unterstuetzt diese Klasse auch Polymorphie.
 * Ist eine Klasse also <code>abstract</code> und mittels
 * {@link Discriminators} annotiert, so koennen abgeleitete Instanzen
 * bei der Deserialisierung im konkreten Fall in den richtige Typ
 * uebersetzt werden. Es spricht aber auch nichts dagegen, dass
 * eine Super-Klasse konkret ist.<br>
 * Der Diskriminator selbst muss dabei nicht zwangslaeufig ein Member,
 * Getter oder static Member der Klasse selbst sein.
 * Ist er allerdings vorhanden, so wird er bevorzugt herangezogen.<br>
 * Aktuell werden keine Interfaces beruecksichtigt. Annotationen hier
 * sind also wirkunkgslos. Ebenso muss jedes
 * <code>private</code>-Member-{@link Field} explizit annotiert werden.
 * Das ist gewollt, eben um Site-Effects, also nicht gewollte Automatismen,
 * die haeufig im Jackson oder seinen Verwandten auftreten, zu unterbinden.
 * @see Getter
 * @see Setter
 * @author (c) 2021/2023 - Carsten Moeller, info@osm2po.de
 */
public final class Reflector {

    private Reflector() {/*STATIC*/}

    /**
     * Steuert den Zugriff per Reflection entweder fuer Lesend,
     * Schreibend, beides oder keinen.
     * @see Getter
     * @see Setter
     */
    public enum MemberAccess {
        ReadWrite, ReadOnly, WriteOnly, Ignore
    }

    /**
     * Steuert abweichendes Verhalten eines {@link Reflector}s.
     * @see Reflectable#value()
     * @see Reflectable#name()
     * @see Reflectable#filter()
     * @see Reflectable#implclass()
     * @see Reflectable#serializeNull()
     * @see Reflectable#order()
     * @see Reflector
     */
    @Retention(RUNTIME)
    @Target({FIELD, METHOD})
    @Documented
    public @interface Reflectable {
        /** @return {@link MemberAccess} erlauben, default {@link MemberAccess#ReadWrite} */
        MemberAccess value() default ReadWrite;
        /** @return Abweichender Name fuer Serialisierung, default "", d.h direkter Feld- oder Methoden-Name */
        String name() default "";
        /** @return BitMask as additional Member-Filter, default long 0 (No Bit set) */
        long filter() default 0L;
        /** @return Abweichende Implementations-Klasse z.B. bei Interface Members */
        Class<?> implclass() default Void.class;
        /** @return Null-Werte serialisieren ja/nein, default true */
        boolean serializeNull() default true;
        /** @return Reihenfolge bei {@link Reflector#getGetters(Class, long)} beeinflussen. Default 0. */
        int order() default 0;
    }

    /**
     * Mappt einen benannten Typ (String) auf eine konkrete {@link Class}.
     * @see Discriminators
     */
    public @interface Discriminator {
        String type();
        Class<?> clazz();
    }
    /**
     * Benennt einen Member (oder Methode) als Typ-Property (String)
     * und assoziert hierzu konkrete {@link Class}es als
     * {@link Discriminator}s. Diese werden im {@link Reflector} und
     * im {@link JsonDeserializer} herangezogen, um z.B. polymorphe
     * Probleme aufzuloesen.
     * @see Discriminator
     */
    @Retention(RUNTIME)
    @Target(TYPE)
    public @interface Discriminators {
        String property() default "type";
        Discriminator[] value();
    }

    /**
     * Abstrakte Basis-Klasse fuer alle per Reflection eingesammelten
     * {@link Getter}s und {@link Setter}s einer Klasse inklusive deren
     * <code>super</code>-Klassen.
     */
    private abstract static class Accessor {
        public final String name;
        public final long filter;

        private Accessor(final String name, long filter) {
            this.name = name;
            this.filter = filter;
        }

        protected final boolean matchesFilter(long bits) {
            if (0L == bits) return true;
            else return (this.filter & bits) != 0L;
        }

        @Override
        public String toString() {
            return this.name;
        }
    }

    /**
     * Abstrakte Basis fuer den lesenden Zugriff per Reflection
     * auf Members einer Klasse.
     * @see Reflector
     * @see FieldGetter
     * @see MethodGetter
     * @see ConstantGetter
     */
    public abstract static class Getter extends Accessor {
        public final boolean nullable;
        public final int order;
        /**
         * Liefert einen Wert per Reflection.
         * @param instance {@link Object} Instanz
         * @return {@link Object} Wert.
         * @see Reflector
         * @throws Exception Diverse
         */
        public abstract Object getValue(Object instance) throws Exception;
        public Getter(final String name, long filter, boolean nullable, int order) {
            super(name, filter);
            this.nullable = nullable;
            this.order = order;
        }
    }

    /**
     * Abstrakte Basis fuer den schreibenden Zugriff per Reflection
     * auf Members einer Klasse.
     * @see Reflector
     * @see FieldSetter
     * @see MethodSetter
     */
    public abstract static class Setter extends Accessor {
        public final Type implType;
        /**
         * Setzt einen Wert per Reflection.
         * @param instance {@link Object} Instanz
         * @param value {@link Object} der Wert selbst
         * @see Reflector
         * @throws Exception Diverse
         */
        public abstract void setValue(Object instance, Object value) throws Exception;
        public Setter(final String name, long filter, boolean nullable, final Type implType) {
            super(name, filter);
            this.implType = implType;
        }
    }

    /**
     * Konkrete Klasse eines {@link Getter}s die einen statisch/konstanten
     * Wert zurueckliefert. Hauptsaechlich verwendet fuer
     * {@link Discriminator}-Werte aus einer {@link Discriminators}-Liste.
     * @see Reflector
     * @see MethodGetter
     * @see Reflector#getGetter(String, Class, long)
     */
    public final static class ConstantGetter extends Getter {
        private final Object value;
        public ConstantGetter(final String name, final Object value, int order) {
            super(name, 0L, false, order);
            this.value = value;
        }
        @Override
        public Object getValue(Object instance) throws Exception {
            return this.value;
        }
    }

    /**
     * Konkrete Klasse fuer den lesenden Zugriff auf ein Feld einer Klasse.
     * @see Reflector
     * @see MethodGetter
     * @see Reflector#getGetter(String, Class, long)
     */
    public final static class FieldGetter extends Getter {
        public final Field field;
        public FieldGetter(final String name, long filter, boolean nullable, int order, final Field field) {
            super(name, filter, nullable, order);
            this.field = field;
        }
        @Override
        public Object getValue(Object instance) throws Exception {
            return this.field.get(instance);
        }
    }

    /**
     * Konkrete Klasse fuer den lesenden Zugriff auf eine Property via Funktion einer Klasse.
     * @see Reflector
     * @see FieldGetter
     * @see Reflector#getGetter(String, Class, long)
     */
    public final static class MethodGetter extends Getter {
        private final Method method;
        public MethodGetter(final String name, long filter, boolean nullable, int order, final Method method) {
            super(name, filter, nullable, order);
            this.method = method;
        }
        @Override
        public Object getValue(Object instance) throws Exception {
            return this.method.invoke(instance);
        }
    }

    /**
     * Konkrete Klasse fuer den schreibenden Zugriff auf ein Feld einer Klasse.
     * @see Reflector
     * @see MethodSetter
     * @see Reflector#getSetter(String, Class, long)
     */
    public final static class FieldSetter extends Setter {
        private final Field field;
        public FieldSetter(final String name, long filter, boolean nullable,
                final Field field, final Type implType) {
            super(name, filter, nullable, implType);
            this.field = field;
        }
        @Override
        public void setValue(Object instance, Object value) throws Exception {
            this.field.set(instance, value);
        }
    }

    /**
     * Konkrete Klasse fuer den schreibenden Zugriff auf eine Property via Funktion einer Klasse.
     * @see Reflector
     * @see FieldSetter
     * @see Reflector#getSetter(String, Class, long)
     */
    public final static class MethodSetter extends Setter {
        private final Method method;
        public MethodSetter(final String name, long filter, boolean nullable,
                final Method method, final Type type) {
            super(name, filter, nullable, type);
            this.method = method;
        }
        @Override
        public void setValue(Object instance, Object value) throws Exception {
            this.method.invoke(instance, value);
        }
    }

    private static class Accessors {
        public final Map<String, Getter> getterMap = new LinkedHashMap<String, Getter>();
        public final Map<String, Setter> setterMap = new HashMap<String, Setter>();
    }

    private static Map<Class<?>, Accessors> classMap = new HashMap<Class<?>, Accessors>();

    public static void clearCache() {
        classMap.clear();
    }

    public static Set<Class<?>> getCachedClasses() {
        return new HashSet<Class<?>>(classMap.keySet());
    }

    /**
     * Liefert einen {@link Setter} fuer den schreibenden Zugriff.<br>
     * Dies kann ein {@link FieldSetter} oder ein {@link MethodSetter} sein.
     * @param key String Schluessel/Alias (normalerweise der Name) eines Members
     * @param rootClass {@link Class} dieser Klasse
     * @param filter long (0=kein Filter), sonst Binaervergleich und Attribut-Filter.
     * @return {@link Setter} oder <code>null</code>, wenn nicht vorhanden.
     * @see Setter#setValue(Object, Object)
     * @see Reflector
     */
    public static Setter getSetter(final String key, final Class<?> rootClass, long filter) {
        if (null == rootClass) return null;
        Setter setter = getAccessors(rootClass).setterMap.get(key);
        if (null == setter) return null;
        if (!setter.matchesFilter(filter)) return null;
        return setter;
    }

    /**
     * Liefert einen {@link Getter} fuer den lesenden Zugriff.<br>
     * Dies kann ein {@link FieldGetter}, ein {@link ConstantGetter} oder ein
     * {@link MethodGetter} sein.
     * @param key String Schluessel/Alias (normalerweise der Name) eines Members
     * @param rootClass {@link Class} dieser Klasse
     * @param filter long (0=kein Filter), sonst Binaervergleich und Attribut-Filter.
     * @return {@link Getter}
     * @see Getter#getValue(Object)
     * @see Reflector
     */
    public static Getter getGetter(final String key, final Class<?> rootClass, long filter) {
        if (null == rootClass) return null;
        Getter getter = getAccessors(rootClass).getterMap.get(key);
        if (null == getter) return null;
        if (!getter.matchesFilter(filter)) return null;
        return getter;
    }


    public static Iterator<Getter> getGetters(final Class<?> rootClass, final long filter) {
        return getAccessors(rootClass).getterMap.values()
                .stream().filter(g->g.matchesFilter(filter)).iterator();
    }

    public static Iterator<Setter> getSetters(final Class<?> rootClass, final long filter) {
        return getAccessors(rootClass).setterMap.values()
                .stream().filter(g->g.matchesFilter(filter)).iterator();
    }

    // ---------- LAZY Accessors (calculated once per class) --------------

    private static synchronized Accessors getAccessors(final Class<?> rootClass) {
        Accessors accessors = classMap.get(rootClass);
        if (accessors != null) {
            return accessors; // if adready reflected then return it
        } else {
            accessors = new Accessors();
        }

        Class<?> currentClass = rootClass;
        Type[] genericClassTypeArguments = null;
        while (currentClass != Object.class && !currentClass.isInterface()) {

            for (Method method : currentClass.getDeclaredMethods()) {
                if (isStatic(method.getModifiers())) continue;

                Reflectable methodAnnotation = method.getAnnotation(Reflectable.class);
                if (methodAnnotation != null) { // Methods muessen Annotation haben!
                    final String methodName = method.getName();
                    final String accessorKey = (0 == methodAnnotation.name().length())
                            ? derivePropertyFromSetterOrGetterName(methodName) : methodAnnotation.name();
                    final MemberAccess memberAccess = methodAnnotation.value();
                    final boolean canRead = ReadWrite == memberAccess || ReadOnly == memberAccess;
                    final boolean canWrite = ReadWrite == memberAccess || WriteOnly == memberAccess;

                    if (canRead && method.getParameterTypes().length == 0) {
                        method.setAccessible(true);
                        accessors.getterMap.putIfAbsent(accessorKey, new MethodGetter(
                                accessorKey, methodAnnotation.filter(),
                                methodAnnotation.serializeNull(), methodAnnotation.order(), method));

                    } else if (canWrite && method.getParameterTypes().length == 1) {
                        method.setAccessible(true);
                        Class<?> annotatedImplClass = methodAnnotation.implclass();
                        if (Void.class == annotatedImplClass) annotatedImplClass = null;

                        Type methodParam0Type = resolveType(
                                method.getGenericParameterTypes()[0], annotatedImplClass, genericClassTypeArguments);

                        accessors.setterMap.putIfAbsent(accessorKey, new MethodSetter(
                                accessorKey, methodAnnotation.filter(), methodAnnotation.serializeNull(),
                                method, methodParam0Type));
                    }
                }
            }

            for (Field field : currentClass.getDeclaredFields()) {
                String accessorKey = field.getName();

                final int modifiers = field.getModifiers();
                MemberAccess memberAccess = Ignore;

                if (isPublic(modifiers) && !isStatic(modifiers)) {
                    // Statics muessen explizit annotiert sein
                    memberAccess = isFinal(modifiers) ? ReadOnly : ReadWrite;
                }

                if (isRecord(currentClass)) {
                    memberAccess = ReadOnly; // Vorgriff auf Java14
                }

                // --- Annotationen ermitteln ----->

                long filter = 0L;
                boolean serializeNull = true;
                int order = 999; // Annotierte Elemente werden so nach vorne kommen, da order=0
                Class<?> implClass = null;

                // Achtung: statics koennen als ueberschreibbar markiert werden. Wird unten behandelt.
                Reflectable fieldAnnotation = field.getAnnotation(Reflectable.class);
                if (fieldAnnotation != null) {
                    serializeNull = fieldAnnotation.serializeNull();
                    order = fieldAnnotation.order();
                    memberAccess = fieldAnnotation.value();
                    if (fieldAnnotation.name().length() > 0)
                        accessorKey = fieldAnnotation.name();
                    if (fieldAnnotation.implclass() != Void.class) {
                        implClass = fieldAnnotation.implclass();
                    }
                    filter = fieldAnnotation.filter();
                }

                // implClass hier bekannt, ggf. auch abweichend per Annotation oben
                final Type fieldType = resolveType(
                        field.getGenericType(), implClass, genericClassTypeArguments);

                // final members sind ueberschreibbar, final statics aber nicht
                if (isStatic(modifiers)) {
                    if (isFinal(modifiers)) {
                        if (WriteOnly == memberAccess) memberAccess = Ignore;
                        else if (ReadWrite == memberAccess) memberAccess = ReadOnly;
                    }
                }

                if (memberAccess != Ignore) {
                    field.setAccessible(true);

                    if (ReadWrite == memberAccess || WriteOnly == memberAccess) {
                        Setter setter = new FieldSetter(
                                accessorKey, filter, serializeNull, field, fieldType);
                        accessors.setterMap.putIfAbsent(accessorKey, setter);
                    }
                    if (ReadWrite == memberAccess || ReadOnly == memberAccess) {
                        Getter getter = new FieldGetter(
                                accessorKey, filter, serializeNull, order, field);
                        accessors.getterMap.putIfAbsent(accessorKey, getter);
                    }
                }
            }

            // In der naechsten Schleife eine Vererbungs-Hierarchie hoeher springen und
            // ggf. hier gesetzte Generic-Arguments z.B. vom Typ Class<Integer, String>
            // dort (in der Super-Klasse) fuer <T,U> ueberschreiben, so dass dann T zu Integer
            // und U zu String aufgeloest werden kann.
            Type genericSuperClass = currentClass.getGenericSuperclass();
            if (genericSuperClass instanceof ParameterizedType) {
                ParameterizedType pt = (ParameterizedType) genericSuperClass;
                Type[] superArgs = pt.getActualTypeArguments();
                for (int i = 0; i < superArgs.length; i++) {
                    // Wichtig, wenn Teilweiterleitung von Generics. Beispiel:
                    // class Baz<A,B>
                    // class Bar<B,C,D> extends Baz<Byte, B>
                    // class Foo extends Bar<Integer, String, BigInteger>
                    // <A,B> aus Baz muss hier zu <Byte, Integer> werden
                    superArgs[i] = resolveType(superArgs[i], null, genericClassTypeArguments);
                }
                genericClassTypeArguments = superArgs;
            }
            currentClass = currentClass.getSuperclass();
        }

        // Ab hier sind alle Fields und Methods aufgeloest.
        // Jetzt final pruefen, ob SuperClass Diskriminatoren annotiert hat.
        // Falls ja und nicht bereits als Member physisch vorhanden, dann
        // virtuellen Getter mit Order=-1 (=vorne) hierfuer ergaenzen.
        Class<?> superClass = rootClass.getSuperclass();
        Discriminators discriminators = superClass.getAnnotation(Discriminators.class);
        if (discriminators != null) {
            final String dtypeName = discriminators.property();
            for (Discriminator discriminator : discriminators.value()) {
                if (discriminator.clazz() == rootClass) {
                    accessors.getterMap.putIfAbsent(dtypeName,
                            new ConstantGetter(dtypeName, discriminator.type(), -1)); // -1=nach vorne
                }
            }
        }


        // Neu durchsortieren nach order first und danach alphabetisch
        ArrayList<Getter> list = new ArrayList<>(accessors.getterMap.values());
        Collections.sort(list, (a,b)->{
            int comp = a.order - b.order;
            if (0 == comp) comp = a.name.compareTo(b.name);
            return comp;
        });
        // Elemte loeschen und sortiert wieder einfuegen
        accessors.getterMap.clear();
        for (Getter getter : list) accessors.getterMap.put(getter.name, getter);

        classMap.put(rootClass, accessors); // cache for next time

        return accessors;
    }

    // Type kann hier alles sein, von konkreter Klasse bis zu "T" aus <T>
    // genericClassTypeArgs sind die <T,U>'s der besitzenden, uebergeordneten Klasse des
    // Members. Bei z.B. einem private T myVar wird T erst in abgeleiteten Klasse spezifiziert,
    // also z.B. in einer class Foo extends Bar<String> einer Bar<T>.
    // Ein T myVar der Klasse Bar wird also erst in Foo konkret zu String.
    public static Type resolveType(Type type, Class<?> implClass, Type[] genericClassTypeArgs) {
        if (type instanceof TypeVariable) { // eg. T of class<T>
            final TypeVariable<?> tv = (TypeVariable<?>) type;
            final String genericName = tv.getName(); // eg. "T" from <T,U,..>
            // Hole jetzt T und U aus der Klasse <T,U>
            TypeVariable<?>[] typeParameters = tv.getGenericDeclaration().getTypeParameters();
            for (int i = 0; i < typeParameters.length; i++) {
                if (typeParameters[i].getName().equals(genericName)) {
                    // if e.g. U matches one of Class<T,U> then position[1] found
                    if (genericClassTypeArgs != null && i < genericClassTypeArgs.length) {
                        return genericClassTypeArgs[i];
                    }
                }
            }
            throw new RuntimeException("Could not resolve type for <"
                    + genericName + "> of " + tv.getTypeName());

        } else if (type instanceof ParameterizedType) {
            ParameterizedType ptype = (ParameterizedType) type;
            if (null == implClass) implClass = (Class<?>)ptype.getRawType();
            final Type[] types = ((ParameterizedType) type).getActualTypeArguments(); // is a clone
            for (int i = 0; i < types.length; i++) {
                // seems ref-safe: SunImpl returns actualTypeArguments.clone();
                types[i] = resolveType(types[i], null, genericClassTypeArgs);
            }
            return new ParamType(implClass, types);

        } else if (type instanceof Class) {
            Class<?> clazz = (Class<?>) type;
            boolean isIorA = (clazz.getModifiers() & (ABSTRACT | INTERFACE)) != 0;
            if (isIorA && implClass != null) {
                return implClass;
            } else {
                return clazz;
            }

        } else if (type instanceof WildcardType) { // <?>
            return type;

        } else {
            return type; // fallback, hopefully never called. but safe.
        }
    }

    /**
     * Statische Fabrik fuer benutzerdefinierte, hierarchische Typen.
     * Beispiel:<br>
     * <code>T(ArrayList.class, BigInteger[].class)</code>
     * @param clazz {@link Class}, Zielklasse, darf auch abstrakt sein, z.B. Map.
     * @param generics VarArray of {@link Type} or {@link Class}
     * @return {@link ParamType}
     */
    public static ParamType T(Class<?> clazz, Type... generics) {
        return new ParamType(clazz, generics);
    }

    public final static class ParamType implements ParameterizedType {
        private Class<?> clazz;
        private Type[] generics;

        private ParamType(Class<?> clazz, Type... generics ) {
            this.clazz = clazz;
            this.generics = generics;
        }

        @Override public Type[] getActualTypeArguments() {return generics;}
        @Override public Type getRawType() {return this.clazz;}
        @Override public Type getOwnerType() {return null;}

        @Override public String toString() {
            String s = this.clazz.getName();
            if (this.generics.length > 0) {
                s += "<";
                for (int i = 0; i < this.generics.length; i++) {
                    if (i > 0) s += ",";
                    if (generics[i] instanceof ParamType) {
                        s += this.generics[i].toString();
                    } else {
                        s += ((Class<?>) this.generics[i]).getSimpleName();
                    }
                }
                s += ">";
            }
            return s;
        }
    }

    private static String derivePropertyFromSetterOrGetterName(final String name) {
        // macht aus z.B. getFoo() oder setFoo() ein kleines foo
        if (name.length() > 3) {
            if (name.startsWith("set") || name.startsWith("get")) {
                char[] carr = name.substring(3).toCharArray();
                if (carr[0] >= 'A' & carr[0] <= 'Z') {
                    carr[0] |= 0x20; // to lowercase
                    return new String(carr);
                }
            }
        }
        return name;
    }


}
