/*
 * $Id: AbstractRouter.java 5956 2024-10-22 16:04:45Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;

/**
 * Abstrakte Basisklasse fuer saemtliche Router.<br>
 * Instanzen sollten mittels {@link #newRouter(Class, AbstractGraph, Log)}
 * erzeugt werden. Wird der <code>new</code>-Operator bevorzugt, so ist
 * dies ebenfalls moeglich. Abgeleitete Klassen muessen dann zwingend den
 * {@link #AbstractRouter(AbstractGraph, Log)}-Konstruktor mittels
 * <code>super</code> rufen, um den {@link Graph} und das {@link Log}
 * zu initialisieren. Anonyme Klassen koennen mittels {@link #init()}
 * nachtraeglich konfiguriert werden. Hierfuer aber unbedingt das
 * JavaDoc lesen!
 * @since 5.4.0
 * @author (c) 2021 - Carsten Moeller - info@osm2po
 * @param <G> {@link Graph} oder {@link OverlayGraph}
 * @see #newRouter(Class, AbstractGraph, Log)
 * @see #graphClassOf(Class)
 * @see #findPath(int, int, double, RouterParameters)
 * @see #init()
 * @see VertexRouter
 * @see OverlayVertexRouter
 */
public abstract class AbstractRouter<G extends AbstractGraph> {

    final protected G graph;
    final protected Log log;

    public AbstractRouter(G graph, Log log) {
        this.graph = graph;
        this.log = log;
        this.init();
    }

    /**
     * Wird als letzte Anweisung im Konstruktor
     * {@link #AbstractRouter(AbstractGraph, Log)} aufgerufen.<br>
     * Bietet so beispielsweise anonymen Klassenkonstruktionen die
     * Moeglichkeit, komplexe Initialisierungen durchzufuehren.<br>
     * Aber <strong>Achtung</strong> bei einem unsportlichen Java-Verhalten:
     * <pre>
     * router = new VertexRouter(graph, log) {
     *
     *   // 1)
     *   // Hier gewinnt <code>foo = 99</code>, da die Initialisierung
     *   // erst nach dem Konstruktor und somit auch nach <code>init()</code>
     *   // ausgeloest wird.
     *
     *   private int foo = 99;
     *   <strong>@Override</strong> protected init() {this.foo = 33;}
     *
     *   // 2)
     *   // Bleibt die Member-Variable <code>foo</code> uninitialisiert,
     *   // so gewinnt wie erwartet die 33 aus <code>init()</code>.
     *
     *   private int foo;
     *   <strong>@Override</strong> protected init() {this.foo = 33;}
     * }
     * </pre>
     * Interessanterweise ist dies nicht bei <cite>normalen</cite>
     * (nicht anonymen) Konstrukten der Fall.<br>Hier wuerde
     * tatsaechlich immer <code>init()</code> gewinnen. Bug in Java?.
     */
    protected void init() {}

    /**
     * Erzeugt eine neue Router-Instanz aus einer Router-Klasse.
     * @param routerClass Class of Router
     * @param graph &lt;G&gt; {@link AbstractGraph}
     * @param log {@link Log} NOT NULL
     * @param <G> Graphenklasse
     * @param <R> Routerklasse
     * @return Konkrete typsichere Instanz von {@link AbstractRouter}
     * @see #newAbstractRouter(Class, AbstractGraph, Log)
     */
    public static <G extends AbstractGraph, R extends AbstractRouter<G>> R newRouter(
            Class<R> routerClass, G graph, Log log) {
        try {
            // return newAbstractRouter(routerClass, graph, log); possible but extra check not needed
            return routerClass.getDeclaredConstructor(
                    graph.getClass(), log.getClass()).newInstance(graph, log);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Analog {@link #newRouter(Class, AbstractGraph, Log)}, allerdings wird hier
     * die generische Type-Safety zwischen {@link AbstractRouter} und
     * {@link AbstractGraph} erst zur Laufzeit ueberprueft.
     * @param routerClass {@link AbstractRouter}-Klasse
     * @param graph {@link AbstractGraph}
     * @param log {@link Log}
     * @param <G> Graphenklasse
     * @param <R> Routerklasse
     * @return {@link AbstractGraph}-Instanz
     */
    public static <G extends AbstractGraph, R extends AbstractRouter<?>> R newAbstractRouter(
            Class<R> routerClass, G graph, Log log) {
        try {
            Class<?> expectedGraphClassT = graphClassOf(routerClass);
            if (expectedGraphClassT != graph.getClass()) {
                throw new IllegalStateException(graph.getClass().getSimpleName()
                        + " in " + routerClass.getSimpleName()
                        + " does not match generic type T of "
                        + expectedGraphClassT.getSimpleName());
            }
            return routerClass.getDeclaredConstructor(
                    graph.getClass(), log.getClass()).newInstance(graph, log);
        }
        catch (IllegalStateException e) {throw e;}
        catch (Throwable t) {throw new RuntimeException(t);}
    }

    /**
     * Holt die verwendete Graphen-Klasse aus dem <code>extends &lt;T&gt;</code>
     * der abgeleiteten Router-Klasse.
     * @param routerClass {@link AbstractRouter} - Implementation
     * @param <G> Class of {@link AbstractGraph}
     * @return (Graph-)Class of &lt;G&gt;
     * @throws Exception Crash
     * @see AbstractRouter
     */
    @SuppressWarnings("unchecked")
    public static <G extends AbstractGraph> Class<G> graphClassOf(
            Class<? extends AbstractRouter<?>> routerClass) throws Exception {
        while (!(routerClass.getGenericSuperclass() instanceof ParameterizedType))
            routerClass = (Class<? extends AbstractRouter<?>>) routerClass.getSuperclass();
        ParameterizedType type = (ParameterizedType) routerClass.getGenericSuperclass();
        Type typeT = type.getActualTypeArguments()[0];
        String graphClassName = typeT.toString();
        if (graphClassName.startsWith("class"))
            graphClassName = graphClassName.substring(6).trim(); // Works with Java 7
        //else graphClassName = typeT.getTypeName(); // <T> Java8+ only
        Class<G> graphClass = (Class<G>) Config.loadClass(graphClassName);
        return graphClass;
    }

    /**
     * Traversiert den Graphen bis zur Abbruchbedingung oder komplett.
     * Diese ist entweder das Erreichen der targetId oder die
     * Ueberschreitung der maxCost. Siehe Parameter.
     * @param sourceId Vertex Start Id (1-wertig)
     * @param constraints {@link RouterConstraints} Abbruchbedingung (NULLABLE).
     * @param params {@link RouterParameters} weitere Einstellungen (NULLABLE).
     * @return boolean true, sobald alle {@link RouterConstraints#targetIds} erreicht.
     * @see #makePath(int)
     * @see #findPath(int, int, double, RouterParameters)
     * @see AbstractRouter
     */
    public abstract boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params);

    /**
     * Erstellt den Pfad nach erfolgreichem
     * {@link #traverse(int, RouterConstraints, RouterParameters)}.<br>
     * Dabei werden vom Ziel die Kanten rueckwaerts durchfahren,
     * die Reihenfolge umgedreht und vorwaers sortiert zurueckgeliefert.
     * @param targetId Id des ZielVertex. Dies muss nicht zwangslaeufig
     * der Ziel-Knoten sein. Es kann auch ein beliebiger Vertex aus
     * {@link #isVisited(int)} uebergeben werden, um so weitere Pfade zu finden.
     * Dies muss jedoch von der jeweiligen Implementation des Routers unterstuetzt werden.
     * @return int[]
     *      Indizes (nicht die IDs!) der Kanten vom Start zum Ziel
     *      oder <code>null</code>, wenn nicht gefunden.
     * @see AbstractRouter
     */
    public abstract int[] makePath(int targetId);

    /**
     * Liefert true, wenn sich ein Vertex nach der Traversierung
     * in der BlackList (ClosedList) befindet, also besucht
     * oder erreicht wurde.
     * @param vertexId int Id des Vertex &gt;0.
     * @return wurde Vertex wurde besucht, ja/nein.
     * @see #makePath(int)
     * @see AbstractRouter
     */
    public abstract boolean isVisited(int vertexId);

    /**
     * Traversiert den kompletten Graphen vom uebergebenen Vertex aus.
     * @param sourceId int Start-Vertex
     * @param params optionale Parameter (NULLABLE)
     */
    public final void traverseFully(int sourceId, RouterParameters params) {
        this.traverse(sourceId, new RouterConstraints(), params);
    }

    /**
     * Traversiert den Graphen, bricht ab, sobald
     * der Ziel-Vertex besucht wurde und liefert den Pfad.<br>
     * Intern wird {@link #traverse(int, RouterConstraints, RouterParameters)}
     * gerufen, und wenn <code>true</code>, dann {@link #makePath(int)} zurueckgeliefert.
     * @param sourceId Vertex Start Id
     * @param targetId Vertex Ziel Id
     * @param maxCost double 0 bzw {@link Double#MAX_VALUE}, wenn keine Zeit- oder Laengen-Begrenzung
     * @param params {@link RouterParameters} weitere Einstellungen (NULLABLE).
     * @return int[]
     *      Indizes (nicht die IDs!) der Kanten vom Start zum Ziel
     *      oder <code>null</code>, wenn nicht gefunden.
     * @see AbstractRouter
     * @see #findPath(int, int, boolean, boolean)
     */
    public final int[] findPath(int sourceId, int targetId, double maxCost, RouterParameters params) {
        if (null == params) params = new RouterParameters(); // Default
        if (!this.traverse(sourceId, new RouterConstraints(maxCost, targetId), params)) return null;
        else return this.makePath(targetId);
    }

    /**
     * Komfort-Version, die {@link #findPath(int, int, double, RouterParameters)} mit konkreten
     * Parametern ruft und sonst alles auf Standard belaesst.
     * @param sourceId Vertex Start Id
     * @param targetId Vertex Ziel Id
     * @param shortestPath boolean true: Kuerzesten Pfad statt schnellsten finden.
     * @param aStar boolean true: Setzt {@link RouterParameters#setHeuristicFactor(double)} auf 1.
     * @return int[]
     *      Indizes (nicht die IDs!) der Kanten vom Start zum Ziel
     *      oder <code>null</code>, wenn nicht gefunden.
     * @see AbstractRouter
     * @see #findShortestPath(int, int)
     * @see #findFastestPath(int, int)
     */
    public final int[] findPath(int sourceId, int targetId, boolean shortestPath, boolean aStar) {
        RouterParameters params = new RouterParameters()
                .setFindShortestPath(shortestPath)
                .setHeuristicFactor(aStar ? 1d : 0d);
        return this.findPath(sourceId, targetId, 0, params);
    }

    /**
     * Komfort-Version, die {@link #findPath(int, int, boolean, boolean)} aufruft.<br>
     * Setzt auf Standard-Dijkstra statt A* und auf Kuerzester-Pfad-Suche.
     * @param sourceId Vertex Start Id
     * @param targetId Vertex Ziel Id
     * @return int[]
     *      Indizes (nicht die IDs!) der Kanten vom Start zum Ziel
     *      oder <code>null</code>, wenn nicht gefunden.
     * @see AbstractRouter
     * @see AbstractRouter#findFastestPath(int, int)
     */
    public final int[] findShortestPath(int sourceId, int targetId) {
        return this.findPath(sourceId, targetId, true, false);
    }

    /**
     * Komfort-Version, die {@link #findPath(int, int, boolean, boolean)} aufruft.<br>
     * Setzt auf Standard-Dijkstra statt A* und auf Schnellster-Pfad-Suche.
     * @param sourceId Vertex Start Id
     * @param targetId Vertex Ziel Id
     * @return int[]
     *      Indizes (nicht die IDs!) der Kanten vom Start zum Ziel
     *      oder <code>null</code>, wenn nicht gefunden.
     * @see AbstractRouter
     * @see #findShortestPath(int, int)
     */
    public final int[] findFastestPath(int sourceId, int targetId) {
        return this.findPath(sourceId, targetId, false, false);
    }

}