/*
 * $Id: EdgeRouter.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.GeoUtils;

/**
 * Edge-Based Implemetation eines OneToOne-Routings auf einem
 * statischen Graphen mit AStar-Unterstuetzung.<br>
 * Die interne PriorityQueue arbeitet dabei auf Kanten-Basis, so dass
 * jede Kante 1x besucht werden kann. Das heisst, ein Segment wird maximal
 * 2x besucht, 1x vorwaerts und 1x in Rueckrichtung, was genauere Resultate
 * fuer Abbiegevorschriften erzeugt.
 * <br><br>
 * Diese Implementation beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * <li> {@link RouterParameters#getHeuristicFactor()} </li>
 * </ul>
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see VertexRouter
 * @see VisitedVerticesAware
 */
public class EdgeRouter extends AbstractRouter<Graph>
implements VisitedVerticesAware {

    public EdgeRouter(Graph graph, Log log) {
        super(graph, log);
    }

    private final static int PQ_INITIAL_CAPACITY = 100000;
    private int pqSize; // Aktuelle ArrayGroesse der PQ
    private int pqCapacity; // Maximale ArrayGroesse der PQ
    private int[] pqIdxs;
    private double[] pqCostsR; // Reelle Weg- oder Zeit-Kosten.
    private double[] pqCostsH; // Heuristische Weg- oder Zeit-Kosten.

    private int[] ses; // settled Edges, gleichzeitig Hash fuer TargetEdges.
    private int targetEdgeIdx; // letzte Edge zum Ziel, -1 wenn nicht gefunden.

    // Aktuelle Edge (nach pqRemove)
    private int pqrEdgeIdx;
    private double pqrCostsR; // Kosten bis hierher real

    // Kosten werden entweder mit km oder h belegt.
    private float[] edgeCosts;

    // Members fuer Heuristiken
    private float[] lats, lons; // Alle Koordinaten.
    private float targetLat, targetLon; // Zielkoordinate.
    private double heuristicFactor = 0d; // Bewertungsfaktor 0 Dijkstra, 1 AStar

    // Modi
    private boolean findShortestPath;
    private boolean ignoreRestrictions;
    private boolean ignoreOneWays;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        if (constraints.targetIds.length > 1) throw new IllegalArgumentException(
                "OneToMany-Routing with more than One Target not supported");

        this.initParams(params);
        this.initPqAndSes();
        this.initAStar(constraints.targetIds[0]); // Unkritisch, wenn Targets > 1
        this.initSource(sourceId, constraints.maxCost);
        this.targetEdgeIdx = this.traverseInternal(constraints.targetIds[0], constraints.maxCost);
        return this.targetEdgeIdx >= 0;
    }

    private void initParams(RouterParameters params) {

        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();
        this.heuristicFactor = params.getHeuristicFactor();

        // Kosten entweder fuer Shortest oder Fastest
        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
            // Wenn Fastest, dann muss die Toleranz vergroessert werden,
            // damit die AStar-Bedingung eingehalten wird.
            // Bedingung: Len(A) + Heuristic(A) <= Len(A) + Len(B) + Heuristic(B)
            // Hier 110 kmh: D.h. zwei aufeinanderfolgende Wege duerfen
            // in ihren Geschwindigkeiten maximal um 110 kmh differieren.
            this.heuristicFactor /= 110;
        }
    }

    private void initPqAndSes() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Maximale ArrayGroesse der PQ
        this.pqIdxs = new int[pqCapacity];
        this.pqCostsR = new double[pqCapacity]; // Reelle Weg- oder Zeit-Kosten.
        this.pqCostsH = new double[pqCapacity]; // Heuristische Weg- oder Zeit-Kosten.

        int nEdges = this.graph.getEdgeIds().length;
        this.ses = new int[nEdges];
        this.targetEdgeIdx = -1;
    }

    private void initAStar(int toTargetId) {
        if (!this.graph.isSupported(Graph.SUPPORT_LATLON)) {
            this.heuristicFactor = 0; // Kein AStar
            return;
        }
        this.lats = this.graph.getLats();
        this.lons = this.graph.getLons();
        this.targetLat = this.lats[toTargetId];
        this.targetLon = this.lons[toTargetId];
    }

    private void initSource(int fromSourceId, double maxCost) {
        int[] edgeIds = this.graph.getEdgeIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIdxs;

        edgeIdxs = this.graph.findOutgoingEdges(fromSourceId);
        for (int i = 0; i < edgeIdxs.length; i++) {
            int edgeIdx = edgeIdxs[i];
            boolean isWrongWay = Graph.isWrongWay(edgeIds[edgeIdx]);
            if (!isWrongWay || this.ignoreOneWays) {
                double costR = this.calcEdgeCost(edgeIdx);
                double costH = costR + this.calcVertexCost(edgeTargetIds[edgeIdx]);
                if (costR < maxCost) {
                    this.pqInsert(edgeIdx, costR, costH);
                    this.ses[edgeIdx] = Integer.MAX_VALUE;
                }
            }
            edgeIdx++;
        }
    }

    private int traverseInternal(int toTargetId, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        byte[] edgeNoTurnBits = this.graph.getEdgeNoTurnBits();

        while (this.pqSize > 0) {
            this.pqRemove();
            int targetId = edgeTargetIds[this.pqrEdgeIdx];
            int uturnId = edgeSourceIds[this.pqrEdgeIdx];

            // Jepp!!! Hier ist der kuerzeste Weg.
            if (toTargetId == targetId) return this.pqrEdgeIdx;

            // Anschlusssegmente suchen.
            int sourceId = targetId; // Rollenwechsel
            int toEdgeIdx = edgeEntries[sourceId];

            int noTurnBits = edgeNoTurnBits[this.pqrEdgeIdx];
            int noTurnBit = 1;

            do { // Anschlusssegmente durchlaufen.
                int se = this.ses[toEdgeIdx];
                if (se == 0) { //TODO Noetig?: || se == -1 */) { // Edge wurde noch nicht durchfahren
                    targetId = edgeTargetIds[toEdgeIdx];
                    if (uturnId != targetId) { // Kein UTurn!
                        // Abbiegevorschriften und Einbahnstrassen beachten.
                        boolean mayTurn = true;
                        if (noTurnBits != 0 && !this.ignoreRestrictions)
                            mayTurn &= ((noTurnBits & noTurnBit) == 0);
                        if ((edgeIds[toEdgeIdx] & 0x80000001) == 0x80000000)  // WrongWay
                            mayTurn &= this.ignoreOneWays;
                        if (mayTurn) {
                            int fromEdgeIdx = this.pqrEdgeIdx; // Retten

                            /*TODO Noetig?: if (se == -1) { // Anschlusssegment ist moegliche TargetEdge
                                fromEdgeIdx |= 0x80000000;
                            }*/

                            // Reelle Kosten bis hier
                            double costR = this.pqrCostsR + this.calcEdgeCost(toEdgeIdx);
                            double costH = costR + this.calcVertexCost(targetId);

                            if (costR < maxCost) {
                                this.pqInsert(toEdgeIdx, costR, costH);
                                // Jetzt schon in Closed-List aufnehmen. Gaaaanz wichtig. Sonst DeadLoop!
                                this.ses[toEdgeIdx] = fromEdgeIdx;
                            }
                        }
                    }
                }
                toEdgeIdx++;
                noTurnBit <<= 1;
            } while (sourceId == edgeSourceIds[toEdgeIdx]);
        }
        return -1;
    }

    @Override
    public final int[] getVisited() {
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int len = this.ses.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int i = 0; i < len; i++) {
            int edgeIdx = this.ses[i];
            if (edgeIdx > 0) { // Kante wurde besucht
                if (buffer.length < i) // Ueberlauf korrigieren
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = edgeTargetIds[i];
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    protected double calcEdgeCost(int edgeIdx) {
        return this.edgeCosts[edgeIdx];
    }

    /**
     * Liefert die heuristischen Kosten eines Vertex.
     * Diese Implementation liefert entweder 0 fuer Dijkstra
     * oder die heuristische Distanz zum Ziel fuer AStar.
     * Eine abgeleitete Klasse koennte an dieser Stelle z.B.
     * auch noch konstante Kosten einer Vertex-Klasse hinzuaddieren.
     * So z.B. fuer Ampeln, Mautstationen oder Bahnuebergaenge.
     * @param vertexId Id des Vertex
     * @return double Kosten.
     */
    protected double calcVertexCost(int vertexId) {
        double h = 0d;
        if (this.heuristicFactor != 0d) { // nur rechnen wenn !=0 (schneller).
            h = GeoUtils.calcDistanceEuklidCoarse(
                    this.lats[vertexId], this.lons[vertexId],
                    this.targetLat, this.targetLon);
            h *= this.heuristicFactor;
        }
        return h;
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqIdxs = Arrays.copyOf(this.pqIdxs, this.pqCapacity);
        this.pqCostsR = Arrays.copyOf(this.pqCostsR, this.pqCapacity);
        this.pqCostsH = Arrays.copyOf(this.pqCostsH, this.pqCapacity);
        log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsert(int edgeIdx, double costsR, double costsH) {
        int pqPos = ++this.pqSize;
        if (this.pqSize >= this.pqCapacity) this.pqGrow();

        this.pqIdxs[pqPos] = edgeIdx;
        this.pqCostsR[pqPos] = costsR;
        this.pqCostsH[pqPos] = costsH;

        this.pqUpheap(pqPos);
    }

    private void pqRemove() {
        this.pqrEdgeIdx = this.pqIdxs[1];
        this.pqrCostsR = this.pqCostsR[1];
        //this.pqrCostsH = this.pqCostsH[1]; Nur fuer PQ-Sort benoetigt.
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap(1);
    }

    private void pqSwap(int pos1, int pos2) {
        // PQ-Swap
        int idx1 = this.pqIdxs[pos1];
        double costsR1 = this.pqCostsR[pos1];
        double costsH1 = this.pqCostsH[pos1];
        int idx2 = this.pqIdxs[pos2];
        double costsR2 = this.pqCostsR[pos2];
        double costsH2 = this.pqCostsH[pos2];

        this.pqIdxs[pos1] = idx2;
        this.pqCostsR[pos1] = costsR2;
        this.pqCostsH[pos1] = costsH2;
        this.pqIdxs[pos2] = idx1;
        this.pqCostsR[pos2] = costsR1;
        this.pqCostsH[pos2] = costsH1;
    }

    private void pqUpheap(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[parentPos] < this.pqCostsH[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos) {
        int childPos = pos << 1; // * 2 = linkes Kind
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqCostsH[childRightPos] < this.pqCostsH[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[pos] < this.pqCostsH[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

    @Override
    public boolean isVisited(int targetId) {
        if (this.targetEdgeIdx < 0) return false;
        return this.graph.getEdgeTargetIds()[this.targetEdgeIdx] == targetId;
    }

    @Override
    public int[] makePath(int targetId) {
        if (!this.isVisited(targetId)) return null;

        int edgeIdx = this.targetEdgeIdx;
        final int INITIAL_CAPACITY = 1000;
        int[] pathIdxs = new int[INITIAL_CAPACITY];
        int n = 0;
        while (edgeIdx != Integer.MAX_VALUE) {
            if (pathIdxs.length == n) pathIdxs = Arrays.copyOf(
                    pathIdxs, pathIdxs.length + INITIAL_CAPACITY);
            pathIdxs[n++] = edgeIdx;
            edgeIdx = this.ses[edgeIdx]; // Vorgaenger
        }

        pathIdxs = Arrays.copyOf(pathIdxs, n);

        // Reverse
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIdxs[i]; pathIdxs[i] = pathIdxs[j]; pathIdxs[j] = z;
        }

        if (this.log != null && log.isEnabled(Level.DATA)) {
            log.data("Segments: " + pathIdxs.length + "\n" + Arrays.toString(pathIdxs));
        }

        return pathIdxs;
    }

} // EOC
