package de.cm.osm2po.service;

import static java.time.format.DateTimeFormatter.ofPattern;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Base64;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESedeKeySpec;

import com.sun.net.httpserver.Headers;
import com.sun.net.httpserver.HttpExchange;

/**
 * a cookie ;-).
 * @see #builder(String, String)
 */
public final class Cookie {

    private static final ZoneId GMT = ZoneId.of("GMT");
    public static final String SECRET_PREFIX = "###~"; // for starts-with-check if encrypted

    /**
     * Specifies whether/when cookies are sent with cross-site requests.
     * @see SameSite#STRICT
     * @see SameSite#LAX
     * @see SameSite#NONE
     */
    public enum SameSite {
        /** sent from origin site only */
        STRICT("Strict"),
        /** Default. Same as Strict, but works with links from foreign sites */
        LAX("Lax"),
        /** if none, then {@link Cookie#secure()} must be set */
        NONE("None");

        public final String value;

        private SameSite(String value) {
            this.value = value;
        }
    }

    private String name;
    private String value;
    private LocalDateTime expires;
    private Integer maxAge;
    private String domain;
    private String path;
    private Boolean secure;
    private Boolean httpOnly;
    private SameSite sameSite;

    private String secretKey;

    private Cookie(String name, String value) {
        this.name = name;
        this.value = value;
    }

    /**
     * Cookie Builder, initialized with required name and value.
     * As long as {@link #maxAgeSeconds(int)} or {@link #expires(int, Month, int, int, int)}
     * are not set, the lifetime is usually bound to the browser-session.
     * However, many browsers restore session-cookies after computer crash.
     * Even maxAge and expire seem to work (or not work) differently under different
     * browsers. So don't rely on it.
     * @param name NOT NULL. Expects url-friendly attribute name.
     * @param value NOT NULL. Exects already url-Encoded String if necessary.
     * @return new {@link Cookie}
     */
    public static Cookie builder(String name, String value) {
        Objects.requireNonNull(name);
        Objects.requireNonNull(value);
        return new Cookie(name, value);
    }

    /**
     * @param maxAge in seconds or NULL in order to ignore this value (=SessionCookie).
     * @return this
     */
    public final Cookie maxAge(Integer maxAge) {
        this.maxAge = maxAge;
        return this;
    }

    /**
     * @param maxAge in seconds or 0 in order to delete Cookie
     * @return this
     */
    public final Cookie maxAgeSeconds(int maxAge) {
        this.maxAge = maxAge;
        return this;
    }

    /**
     * @param maxAge in minutes or 0 in order to delete Cookie
     * @return this
     */
    public final Cookie maxAgeMinutes(int maxAge) {
        return maxAgeSeconds(maxAge * 60);
    };

    /**
     * @param maxAge in hours or 0 in order to delete Cookie
     * @return this
     */
    public final Cookie maxAgeHours(int maxAge) {
        return maxAgeMinutes(maxAge * 60);
    };

    /**
     * Specific path, e.g. /foo to match /foo/x/y but not /bar or / (root).
     * If not set, the Browser will derive this from origin.
     * @param path String
     * @return this
     */
    public final Cookie path(String path) {
        this.path = path;
        return this;
    }

    /**
     * Specifies which server can receive the cookie.
     * If not set subdomains are excluded. Setting it to
     * e.g. 'foo.org', also includes 'sub.foo.org'
     * @param domain String
     * @return this
     */
    public final Cookie domain(String domain) {
        this.domain = domain;
        return this;
    }

    /**
     * Sets the Secure flag, meaning HTTPS only.
     * @return this
     */
    public final Cookie secure() {
        this.secure = true;
        return this;
    }

    /**
     * Sets httpOnly, so JavaScript in Browser cannot see this Cookie.
     * @return this
     */
    public final Cookie httpOnly() {
        this.httpOnly = true;
        return this;
    }

    /**
     * @param policy {@link SameSite}
     * @return this
     */
    public final Cookie sameSite(SameSite policy) {
        this.sameSite = policy;
        return this;
    }

    /**
     * If set (not null), the value will be encrypted and
     * prefixed with {@value #SECRET_PREFIX}.
     * @param secretKey String NULL=no encrpytion
     * @return this
     * @see #fromHeaders(Headers, String)
     */
    public final Cookie encrypt(String secretKey) {
        this.secretKey = secretKey;
        return this;
    }

    /**
     * Sets the absolute timestamp for cookie's end of live.
     * Important: If {@link #maxAgeSeconds(int)} is set, this attribute
     * will not be sent, because the age has higher precedence.
     * @param year int
     * @param month int
     * @param dayOfMonth {@link Month}
     * @param hour int
     * @param minute int
     * @return this
     */
    public final Cookie expires(int year, Month month, int dayOfMonth, int hour, int minute) {
        this.expires = LocalDateTime.of(year, month, dayOfMonth, hour, minute, 0);
        return this;
    }

    /**
     * Expire Cookie today at given time.
     * @param hour int
     * @param minute int
     * @return this
     */
    public final Cookie expiresToday(int hour, int minute) {
        this.expires = LocalDateTime.now().withHour(hour).withMinute(minute).withSecond(0);
        return this;
    }

    /**
     * Expire Cookie tomorrow at given time.
     * @param hour int
     * @param minute int
     * @return this
     */
    public final Cookie expiresTomorrow(int hour, int minute) {
        this.expires = LocalDateTime.now().plusDays(1).withHour(hour).withMinute(minute).withSecond(0);
        return this;
    }

    /**
     * Expire Cookie at UnixTimeStamp (seconds).
     * @param unixTimestamp Long NULLABLE (if null then omitted)
     * @return this
     */
    public final Cookie expires(Long unixTimestamp) {
        if (null == unixTimestamp) {
            this.expires = null;
        } else {
            Instant instant = Instant.ofEpochMilli(unixTimestamp.longValue() * 1000);
            this.expires = LocalDateTime.ofInstant(instant, GMT);
        }
        return this;
    }

    /**
     * Cakks {@link #build()} and adds the Cookie to the
     * {@link HttpExchange#getResponseHeaders()}.
     * @param htx {@link HttpExchange}
     */
    public void buildAndAddToResponse(HttpExchange htx) {
        htx.getResponseHeaders().add("Set-Cookie", this.build());
    }

    public String build() {
        String value = (null == this.secretKey ? this.value : encrypt(this.secretKey, this.value));
        StringBuilder sb = new StringBuilder(0x100).append(this.name + "=" + value);

        if (this.maxAge != null) sb.append("; Max-Age=" + this.maxAge);
        if (this.domain != null) sb.append("; Domain=" + this.domain);
        if (this.path != null) sb.append("; Path=" + this.path);
        if (this.secure != null) sb.append("; Secure");
        if (this.httpOnly != null)sb.append("; HttpOnly");
        if (this.sameSite != null) sb.append("; SameSite=" + sameSite.value);

        if (this.expires != null && null == this.maxAge) {
            // RFC_1123_DATE_TIME is not 100% compatible with HTTP 1.1 DateTime
            // see https://stackoverflow.com/questions/7707555/getting-date-in-http-format-in-java
            final DateTimeFormatter RFC_2616 = ofPattern("EEE, dd MMM yyyy HH:mm:ss z", Locale.ENGLISH);
            final String rfc = RFC_2616.format(
                    ZonedDateTime.of(this.expires, ZoneId.of("GMT")));
            sb.append("; Expires=" + rfc);
        }

        return sb.toString();
    }

    // some useful funcs

    /**
     * Calls {@link #fromHeaders(Headers, String)} with null
     * secret key for Request.
     * @param htx {@link HttpExchange}
     * @return this
     */
    public static Map<String, String> fromRequest(HttpExchange htx) {
        return fromHeaders(htx.getRequestHeaders(), null);
    }

    /**
     * Calls {@link #fromHeaders(Headers, String)} for Request.
     * @param htx {@link HttpExchange}
     * @param secretKey String NULLABLE
     * @return this
     */
    public static Map<String, String> fromRequest(HttpExchange htx, String secretKey) {
        return fromHeaders(htx.getRequestHeaders(), secretKey);
    }

    /**
     * Returns Cookie's Name-Value-Pairs as Map.
     * @param headers {@link Headers}
     * @param secretKey key if encrypted, NULLABLE (null = no decryption)
     * @return Map of Name-Value Pairs (both Strings)
     * <br><strong>Important:</strong><br>
     * If key is not null and the Cookie-Value starts with
     * {@value #SECRET_PREFIX} then the value will be decrypted.
     */
    public static Map<String, String> fromHeaders(Headers headers, String secretKey) {
        Map<String, String> cookieMap = new HashMap<String, String>();
        List<String> cookieList = headers.get("Cookie");
        if (cookieList != null) for (String cookieString : cookieList) {
            for (String cookie : cookieString.split(";")) {
                String[] cookiePair = cookie.split("=");
                String name = cookiePair[0].trim();
                String value = cookiePair[1].trim();
                if (secretKey != null && value.startsWith(SECRET_PREFIX)) {
                    value = decrypt(secretKey, value.substring(SECRET_PREFIX.length()));
                    if (null == value)
                        continue; // Decryption failed
                }
                cookieMap.put(name, value);
            }
        }
        return Collections.unmodifiableMap(cookieMap);
    }

    public static void toRequest(Map<String, String> nameValueMap, HttpExchange htx) {
        String s = null;
        if (nameValueMap != null) for (Entry<String, String> e : nameValueMap.entrySet()) {
            if (null == s) s = ""; else s += "; ";
            s += e.getKey() + "=" + e.getValue(); // URLEncode or enclose in " ?
        }
        if (s != null) htx.getRequestHeaders().set("Cookie", s);
    }


    private static SecretKey bytes2SecretKeyDESede(byte[] bytes) throws Exception {
        DESedeKeySpec keyspec = new DESedeKeySpec(bytes);
        SecretKeyFactory keyfactory = SecretKeyFactory.getInstance("DESede");
        return keyfactory.generateSecret(keyspec);
    }


    private static String encrypt(final String secret, final String val) {
        try {
            SecretKey secretKey = bytes2SecretKeyDESede(secret.getBytes());
            Cipher cipher = Cipher.getInstance("DESede");
            cipher.init(Cipher.ENCRYPT_MODE, secretKey);
            byte[] encryptedVal = cipher.doFinal(val.getBytes());

            String secretValBase64 = SECRET_PREFIX + new String(
                    Base64.getUrlEncoder().encode(encryptedVal));
            return secretValBase64;
        } catch (Exception e) {
            return null;
        }
    }

    private static String decrypt(final String secret, String val) {
        try {
            SecretKey secretKey = bytes2SecretKeyDESede(secret.getBytes());
            byte[] bytes = Base64.getUrlDecoder().decode(val);
            Cipher cipher = Cipher.getInstance("DESede");
            cipher.init(Cipher.DECRYPT_MODE, secretKey);
            byte[] decryptedBytes = cipher.doFinal(bytes);
            return new String(decryptedBytes);
        } catch (Exception e) {
            return null;
        }
    }

}