/*
 * $Id: RateLimitServer.java 5987 2024-11-01 07:35:37Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;

import java.io.File;
import java.util.Optional;

import de.cm.osm2po.Config;
import de.cm.osm2po.ShutdownHookSequence.ShutdownAware;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.service.Sockets.Server;
import de.cm.osm2po.service.security.RateLimit.Status;
import de.cm.osm2po.service.security.RateLimiter;

/**
 * @author (c) 2024 - Carsten Moeller, info@osm2po.de
 */
public final class RateLimitServer implements ShutdownAware {

    /** StandardPort {@value} */
    public static final int DEFAULT_PORT = 9999;

    private Server socketServer;
    private RateLimiter rateLimiter;

    public RateLimitServer(Config config) throws Exception {
        final String prefix = "service.ratelimiter.";
        final Log log = Optional.ofNullable(config.getLog()).orElse(Log.stderr());

        File keystoreFile = null;
        String keystorePath = config.getProperty(prefix + "keystore");
        if (keystorePath != null) keystoreFile = new File(keystorePath);

        File backupFile = null;
        String backupPath = config.getProperty(prefix + "backup");
        if (backupPath != null) backupFile = new File(backupPath);

        // RefreshInterval in Minuten. Default = 1
        final long refreshInterval = 60000 * Long.parseLong(
                config.getProperty(prefix + "refresh", "1"));

        // MaxIdle von RateLimit-Elementen in Minuten. Default = 5
        final long maxIdle = 60000 * Long.parseLong(
                config.getProperty(prefix + "ttl", "5"));

        final int port = Integer.parseInt(
                config.getProperty(prefix + "port", String.valueOf(DEFAULT_PORT)));

        final String apiKey = config.getProperty(prefix + "apikey");
        final boolean devMode = Boolean.valueOf(config.getProperty(prefix + "devmode"));

        this.rateLimiter = new RateLimiter(log, backupFile,
                refreshInterval, 10, maxIdle, keystoreFile).start();

        this.socketServer = new Server(port, log, apiKey, key->{
            if (devMode && "<EXIT>".equals(key)) {
                log.info(this.getClass().getSimpleName() + " calling System.exit(-1)");
                new Thread(()->System.exit(-1)).start(); // muss im eigenen Thread laufen
                return Status.SERVICEDOWN.name();
            } else {
                return rateLimiter.validate(key);
            }
        });
    }

    @Override
    public final void shutDown() {
        if (this.socketServer != null) this.socketServer.close();
        if (this.rateLimiter != null) this.rateLimiter.close();
    }

}
