/*
 * $Id: ServiceHandlerProxy.java 5870 2024-03-30 18:12:49Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;

import static de.cm.osm2po.service.errors.ErrorEnum.ERR_HTTP_METHOD_NOT_ALLLOWED;
import static java.lang.reflect.Modifier.PUBLIC;
import static java.lang.reflect.Modifier.STATIC;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import com.sun.net.httpserver.HttpExchange;

import de.cm.osm2po.Config;
import de.cm.osm2po.annotations.HandlerMethod;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * Behandelt Requests des {@link Osm2poService} per Class-Reflection.<br>
 * Eine Handler-Methode muss lediglich eine {@link ServiceResponse} zurueckliefern
 * und public sein.<br>
 * Akzeptiert sie exakt einen {@link ServiceRequest} als Parameter, so wird sie
 * direkt erkannt. Alternativ darf sie auch die folgenden optionalen Parameter
 * in beliebiebiger reihenfolge zur Verfuegung stellen, die wie folgt
 * interpretiert werden:<br>
 * {@link Properties} queryParams, {@link String} cmd (normalerweise
 * der Methodenname selbst) und {@link HttpExchange}.<br>
 * Zusaetzliche Informationen koennen mittels {@link HandlerMethod}
 * annotiert werden.<br>
 * Der Konstruktor wird ebenfalls per Reflection gerufen und kann folgende
 * Parameter aufnehmen:<br>
 * {@link Config}, {@link Graph}, {@link Osm2poService} und {@link Log}.
 * Existieren mehrere Konstruktoren, so gewinnt der mit den meisten
 * uebereinstimmenden Parametern.<br>
 * Die Klasse selbst muss per Konfiguration oder Kommandozeile mittels
 * z.B. <code>service.handler.1.class=my.pack.Class</code> bekannt gemacht werden.
 *
 * @author (c) 2017 - Carsten Moeller - info@osm2po
 */
public class ServiceHandlerProxy {

    private final Osm2poService service;
    private final Config config;
    private final Graph graph;
    private final Log log;
    private Map<String, HandlerMethodProxy> handlerMethodMap;
    private Map<String, String> cmdMappingMap;

    public ServiceHandlerProxy(Osm2poService service,
            Config config, Graph graph) {
        this.service = service;
        this.config = config;
        this.graph = graph;
        this.log = config.getLog();
        this.handlerMethodMap = new HashMap<String, HandlerMethodProxy>();
        this.cmdMappingMap = new HashMap<String, String>();
    }

    private static class HandlerMethodProxy {
        private final Object instance;
        private final Method method;
        private final String httpMethod;
        private final boolean heavy;
        private final boolean perfectMatch;
        private HandlerMethodProxy(Object instance,
                Method method, String httpMethod, boolean heavy, boolean perfectMatch) {
            this.instance = instance;
            this.method = method;
            this.httpMethod = httpMethod;
            this.heavy = heavy;
            this.perfectMatch = perfectMatch;
        }
    }

    public final void shutDown() {
        this.cmdMappingMap.clear();
        this.handlerMethodMap.clear();
    }

    public final Map<String, String> getCmdMapping() {
        return this.cmdMappingMap;
    }

    public final boolean registerHandlerClass(Object instanceOrClass) throws Exception {
        boolean registered = false;

        final boolean isClassArg = instanceOrClass instanceof Class;
        Class<?> clazz = isClassArg ? (Class<?>) instanceOrClass : instanceOrClass.getClass();
        Object instance = isClassArg ? null : instanceOrClass;

        if (isClassArg) {
            // Vermeide clazz.getDeclaredConstructor().newInstance(); finde stattdessen passendsten Konstruktor
            Constructor<?> constructor = null;
            Object[] constructorParams = null;
            int nMaxAssignableParams = -1;

            for (Constructor<?> ctor : clazz.getConstructors()) {
                // getConstructors() liefert nur publics!
                Class<?>[] clazzes = ctor.getParameterTypes();
                Object[] params = new Object[clazzes.length];
                int nAssignableParams = 0;
                for (int i = 0; i < clazzes.length; i++) {
                    nAssignableParams++;
                    if (Config.class.isAssignableFrom(clazzes[i])) params[i] = this.config;
                    else if (Graph.class.isAssignableFrom(clazzes[i])) params[i] = this.graph;
                    else if (Log.class.isAssignableFrom(clazzes[i])) params[i] = this.log;
                    else if (Osm2poService.class.isAssignableFrom(clazzes[i])) params[i] = this.service;
                    else {params[i] = null; nAssignableParams--;}
                }
                if (nAssignableParams > nMaxAssignableParams) {
                    nMaxAssignableParams = nAssignableParams;
                    constructor = ctor;
                    constructorParams = params;
                }
            }

            if (null == constructor) {
                log.error("No public constructor found in " + clazz.getName());
                return false;
            }

            instance = constructor.newInstance(constructorParams);
        }


        for (Method method : clazz.getDeclaredMethods()) {
            if (ServiceResponse.class.isAssignableFrom(method.getReturnType())) {
                int modifiers = method.getModifiers();
                if ((STATIC & modifiers) != 0) continue;
                if (0 == (PUBLIC & modifiers)) continue;

                String cmd = method.getName(); // default
                String httpMethod = "GET";
                boolean heavy = false;
                boolean perfectMatch = false;

                Class<?>[] paramTypes = method.getParameterTypes();
                if (paramTypes.length == 1
                        && ServiceRequest.class.isAssignableFrom(paramTypes[0])) {
                    perfectMatch = true;
                }

                HandlerMethod handlerMethod = method.getAnnotation(HandlerMethod.class);
                if (handlerMethod != null) {
                    if (handlerMethod.ignore()) continue; // Kein Handler!
                    if (!"".equals(handlerMethod.cmd())) cmd = handlerMethod.cmd();
                    heavy = handlerMethod.heavy();
                    httpMethod = handlerMethod.method();

                    for (String mapping : handlerMethod.mapping()) {
                        if (!mapping.startsWith("/")) mapping = "/" + mapping;
                        this.cmdMappingMap.put(mapping, cmd);
                    }
                }

                if (this.handlerMethodMap.containsKey(cmd)) log.warn(
                        "Overwriting Handler for '" + cmd + "' with "
                        + clazz.getSimpleName() + "." + method.getName() + "()");

                HandlerMethodProxy proxy = new HandlerMethodProxy(
                        instance, method, httpMethod, heavy, perfectMatch);
                this.handlerMethodMap.put(cmd, proxy);

                registered = true;
            }
        }
        return registered;
    }


    public final ServiceResponse invokeHandler(
            ServiceRequest serviceRequest) throws Exception {
        try {
            HandlerMethodProxy proxy = this.handlerMethodMap.get(serviceRequest.cmd);

            if (proxy != null) {
                String requestMethod = serviceRequest.htx.getRequestMethod();
                if (!requestMethod.equalsIgnoreCase(proxy.httpMethod)) {
                    throw new ServiceException(ERR_HTTP_METHOD_NOT_ALLLOWED, requestMethod);
                }

                if (proxy.heavy) this.service.checkResources();
                if (proxy.perfectMatch) {
                    return (ServiceResponse) proxy.method.invoke(proxy.instance, serviceRequest);
                } else {
                    Class<?>[] clazzes = proxy.method.getParameterTypes();
                    Object[] params = new Object[clazzes.length];

                    for (int i = 0; i < clazzes.length; i++) {
                        if (String.class.isAssignableFrom(clazzes[i]))
                            params[i] = serviceRequest.cmd;
                        else if (Properties.class.isAssignableFrom(clazzes[i]))
                            params[i] = serviceRequest.params;
                        else if (HttpExchange.class.isAssignableFrom(clazzes[i]))
                            params[i] = serviceRequest.htx;
                        else params[i] = null;
                    }

                    return (ServiceResponse) proxy.method.invoke(proxy.instance, params);
                }
            }

        } catch (InvocationTargetException itx) {
            if (itx.getCause() instanceof Exception) {
                throw (Exception) itx.getCause(); // Originalfehler im Service behandeln
            } else {
                log.fatal(itx); // Anderer Grund
                throw itx;
            }
        }
        return null;
    }

    @Override
    public String toString() {
        String s = null;
        for (Entry<String, HandlerMethodProxy> entry : this.handlerMethodMap.entrySet()) {
            if (null == s) s = ""; else s += '\n';
            String className = entry.getValue().instance.getClass().getName();
            String methodName = entry.getValue().method.getName();
            s += entry.getKey() + " : " + className + "." + methodName + "()";
        }

        if (!this.cmdMappingMap.isEmpty()) {
            s += "\nMappings:";
            for (Entry<String, String> entry : this.cmdMappingMap.entrySet()) {
                s += '\n' + entry.getKey() + " : " + entry.getValue();
            }
        }
        return s;
    }

}