/*
 * $Id: Smtp.java 5972 2024-10-26 20:46:41Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;

import static de.cm.osm2po.misc.Utils.UTF8;
import static de.cm.osm2po.service.Smtp.MultiPartType.MIXED;
import static de.cm.osm2po.service.Smtp.PartType.TEXT;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.Base64;

/**
 * Ein einfacher SMTP-Mail-Sender.<br>
 * Unterstuetzt Standard (Port 25) und SSL/TLS (normalerweise Port 465).<br>
 * STARTTLS wird nicht unterstuetzt.
 * @author (c) 2024 - Carsten Moeller, info@osm2po.de
 * @see #Smtp(Log, String, String, int, boolean, String, String)
 */
public class Smtp {

    private final Log log;
    private final String from;
    private final String host;
    private final int port;
    private final boolean ssltls;
    private final String userpass;

    /**
     * CTOR.
     * @param log {@link Log}
     * @param from EMail-Adresse des Absenders
     * @param host String SMTP-Host z.B. mail.gmx.net
     * @param port int SMTP-Port, z.B. 25 (standard) oder 465 (ssltls)
     * @param ssltls true: SSL-Socket mit TLS, false: Unverschluesselt
     * @param user String SMTP-AnmeldeName fuer SMTP-Host
     * @param pass String SMTP-AnmeldePassWort fuer SMTP-Host
     * @see Part
     */
    public Smtp(Log log, String from, String host,
            int port, boolean ssltls, String user, String pass) {
        if (null == log) log = Log.empty();
        this.log = log;
        this.from = new String(from.getBytes(UTF8));
        this.host = host;
        this.port= port;
        this.ssltls = ssltls;
        this.userpass = Base64.encodeMime(
                ('\0' + user + '\0' + pass).getBytes());
    }

    private static final class CmdAck {
        private final String command;
        private final int responseCode; // expected
        private CmdAck(int responseCode) {this(null, responseCode);}
        private CmdAck(String command, int responseCode) {
            this.command = command;
            this.responseCode = responseCode;
        }
    }

    /**
     * Sendet eine simple Text-Mail.
     * @param subject String Betreff
     * @param to String Empfaenger EMail-Adresse
     * @param text String Mail-Text
     * @return true: Mail wurde versendet, false: Fehler
     */
    public boolean send(String subject, String to, String text) {
        return this.send(subject, to, new Part(text));
    }

    /**
     * Sendet eine Mail.
     * @param subject String Betreff
     * @param to String Empfaenger EMail-Adresse
     * @param part Part, NOT NULL
     * @return true: Mail wurde versendet, false: Fehler
     */
    public boolean send(String subject, String to, Part part) {
        return this.send(subject, new Recipients().to(to), part);
    }

    /**
     * Sendet eine Mail.
     * @param subject String Betreff
     * @param recipients Rcpt, not empty.
     * @param part {@link Part}
     * @return true: Mail wurde versendet, false: Fehler
     * <br><br><b>Wichtig:</b><br>
     * Hier werden keine Exceptions geworfen, {@link Throwable}
     * wird aufgefangen.
     */
    public boolean send(String subject, Recipients recipients, Part part) {
        boolean sent = false;
        try {
            ArrayList<CmdAck> cas = new ArrayList<>();
            cas.add(new CmdAck(220)); // EMail-Server is Ready
            cas.add(new CmdAck("EHLO osm2po", 250)); // gmx und web fordern (fake)host hinter EHLO
            cas.add(new CmdAck("AUTH PLAIN " + this.userpass, 235));
            cas.add(new CmdAck("MAIL FROM: " + removeAlias(this.from), 0));
            for (String rcpt : recipients.rcptSet) cas.add(
                    new CmdAck("RCPT TO: " + rcpt, 0));
            cas.add(new CmdAck("DATA", 354));
            cas.add(new CmdAck("X-Mailer:osm2po.de", 0));
            cas.add(new CmdAck("Content-Type: " + part.pti.contentType, 0));
            if (part.pti.base64) cas.add(
                    new CmdAck("Content-Transfer-Encoding: base64", 0));
            cas.add(new CmdAck("From: " + this.from, 0));
            if (recipients.to != null) cas.add(
                    new CmdAck("To: " + recipients.to, 0));
            if (recipients.cc != null) cas.add(
                    new CmdAck("Cc: " + recipients.cc, 0));
            // Adding BCC here is possible, but nonsense, because it is visible in Thunderbird e.g. then.
            cas.add(new CmdAck("Subject: " + new String(subject.getBytes(UTF8)), 0));
            cas.add(new CmdAck("", 0)); // Muss laut Doku, geht aber ohne
            cas.add(new CmdAck(part.body, 0));
            cas.add(new CmdAck("", 0));
            cas.add(new CmdAck(".", 250)); // EndOfMessage
            cas.add(new CmdAck("QUIT", 221));

            Socket smtpSocket = null;
            if (this.ssltls) {
                SSLContext sslContext = SSLContext.getInstance("TLS");
                sslContext.init(null, null, null);
                SSLSocketFactory sf = sslContext.getSocketFactory();
                smtpSocket = sf.createSocket();
            } else {
                smtpSocket = new Socket();
            }
            smtpSocket.setSoTimeout(3000); // InputStream Read-Timeout
            smtpSocket.connect(new InetSocketAddress(this.host, this.port), 1000); // Connect-Timeout

            OutputStream out = smtpSocket.getOutputStream();
            BufferedReader br = new BufferedReader(
                    new InputStreamReader(smtpSocket.getInputStream()));

            final byte[] CRLF = "\r\n".getBytes();
            for (CmdAck ca : cas) {

                if (log.isEnabled(Level.DATA)) log.data("-> "
                        + (0 == ca.responseCode ? "" : "[" + ca.responseCode + "] ")
                        + (null == ca.command ? "" : ca.command));

                if (ca.command != null) {
                    out.write(ca.command.getBytes());
                    out.write(CRLF);
                }

                if (ca.responseCode != 0) {
                    for (;;) {
                        String response = br.readLine();
                        if (log.isEnabled(Level.DATA)) log.data("<- " + response);
                        int code = Integer.valueOf(response.substring(0, 3));

                        if (code >= 400) {
                            smtpSocket.close();
                            throw new IllegalArgumentException(
                                    this.host + ": " + response);
                        }

                        if (code == 221) sent = true; // closing channel
                        if (code == ca.responseCode) break;
                    }
                }
            }

            smtpSocket.close();

        } catch(Throwable t) {
            log.error(t.getClass().getSimpleName() + ": " + t.getMessage());
        }
        return sent;
    }

    public static final class PartTypeInfo {
        public final String contentType;
        public final boolean base64;
        public PartTypeInfo(String contentType, boolean base64) {
            this.contentType = contentType;
            this.base64 = base64;
        }
        public String dataUrl(final byte[] rawBytes) {
            final String data64 = "data:" + this.contentType + ";base64," + Base64.encodeMime(rawBytes);
            return data64; // no direct return, better for Breakpoint.
        }
    }

    public enum PartType {
        TEXT(new PartTypeInfo("text/plain; charset=\"utf-8\"", false)),
        HTML(new PartTypeInfo("text/html; charset=\"utf-8\"", false)),
        JPG(new PartTypeInfo("image/jpeg", true)),
        PNG(new PartTypeInfo("image/png", true)),
        PDF(new PartTypeInfo("application/pdf", true));

        public final PartTypeInfo pti;

        PartType(PartTypeInfo pti) {
            this.pti = pti;
        }

        public String dataUrl(final byte[] rawBytes) {
            return this.pti.dataUrl(rawBytes);
        }
    }

    public enum MultiPartType {
        MIXED("multipart/mixed"),
        ALTERNATIVE("multipart/alternative"),
        RELATED("multipart/related");

        public final String contentType;

        private MultiPartType(String contentType) {
            this.contentType = contentType;
        }
    }

    /**
     * Message-Part.<br><br>
     * Darf einfach oder Multi oder eine Kombination aus beiden sein.<br>
     * Fuer Multi existieren zwei Konstruktoren, die wiederum mehrere Parts
     * als VarArgsArray entgegennehmen. Bei einem laesst sich zudem der
     * {@link MultiPartType} bestimmen. Der andere setzt
     * auf den Standard {@link MultiPartType#MIXED}.<br>
     * Alle Konstruktoren erwarten entweder <code>String</code>
     * oder <code>byte[]</code>-Parameter als Body-Payload. Dieser darf vorher
     * nicht in irgend einer Weise formatiert oder kodiert sein.<br>
     * {@link PartType} ist ein Enum, welches die wichtigsten Typen mittels
     * {@link PartTypeInfo} vorkonfiguriert, also Content-Type und ob
     * <code>Content-Transfer-Encoding</code> mit base64 gesetzt werden muss.
     * Das Info-Object kann aber auch direkt an einen weiteren Konstruktor
     * uebergeben werden und laesst so auch frei definierbare Typen zu.<br>
     * Einige Konstruktoren setzen zudem das Member {@link #file}. Ist dieses
     * nicht <code>null</code>, dann wird zusaetzlich ein
     * <code>Content-Disposition: attachment;</code> mit diesem Dateinamen gesendet.
     * Wenn der Name jedoch ein @ enthaelt, z.B. "foo@bar", dann wird
     * stattdessen <code>Content-Id</code> mit diesem Wert besetzt, was nur
     * in einem {@link MultiPartType#RELATED}-Block sinnvoll ist. Beispiel:<br>
     * Wichtig: Der erste Part muss das Hauptdokument sein
     * <pre>
     * Part part = new Part(MultiPartType.RELATED,
     *   new Part("&lt;img src=\"" + "cid:file@jpg" + "\"&gt;&lt;/img&gt;", PartType.HTML),
     *   new Part(imgBytes, PartType.JPG, "file@jpg")
     * );
     * </pre>
     * Hinweis:<br>
     * Viele Mail-Apps, insbesondere die auf Mobilgeraeten, koennen das nicht mehr
     * richtig anzeigen. Ausserdem gilt die related-Mechanik als veraltet.
     * Daher sollte man in HTML-Mails besser Data-URLs einsetzen, z.B. mittels
     * {@link PartTypeInfo#dataUrl(byte[])}, oder gleich auf externe URLs verlinken.
     * Fuer bereits vordefinierte Typen laesst sich dieses Info-Objekt auch direkt
     * via {@link PartType#pti} ermitteln.
     * @see Smtp
     */
    public static final class Part {
        public final String body;
        public final PartTypeInfo pti;
        public final String file;

        public Part(String text) {this(TEXT, text);}
        public Part(PartType type, String text) {this(type, text.getBytes(UTF8));}
        public Part(PartType type, byte[] bytes) {this(type, bytes, null);}
        public Part(PartType type, byte[] bytes, String file) {this(type.pti, bytes, file);}
        public Part(PartTypeInfo pti, byte[] bytes, String file) {
            this.body = pti.base64 ? Base64.encodeMime(bytes) : new String(bytes);
            this.pti = pti;
            this.file = (null == file ? null : new String(file.getBytes(UTF8)));
        }
        public Part(Part... parts) {this(MIXED, parts);}
        public Part(MultiPartType type, Part... parts) {
            final String boundary =
                    "X" + UUID.randomUUID().toString().replace("-", "").toUpperCase() + "X";
            final String contentType = type.contentType + "; boundary=" + boundary;
            this.pti = new PartTypeInfo(contentType, false);
            this.body = createMultiPartMessage(boundary, parts);
            this.file = null;
        }
    }

    private static String createMultiPartMessage(String boundary, Part[] parts) {
        final String CRLF = "\r\n";
        StringBuilder sb = new StringBuilder(0x4000 /*16k*/);

        if (parts != null) for (Part part : parts) {
            sb.append("--").append(boundary).append(CRLF);
            sb.append("Content-Type: ").append(part.pti.contentType).append(CRLF);
            if (part.pti.base64) sb.append("Content-Transfer-Encoding: base64").append(CRLF);
            if (part.file != null) {
                if (part.file.contains("@")) sb
                    .append("Content-ID: <" + part.file + ">").append(CRLF);
                else sb
                    .append("Content-Disposition: attachment; filename=")
                    .append("\"" + part.file + "\"").append(CRLF);
            }
            sb.append(CRLF);
            sb.append(part.body).append(CRLF);
        }

        sb.append("--").append(boundary).append("--").append(CRLF);
        return sb.toString();
    }

    public static final class Recipients {
        private static enum Type {TO, CC, BCC}
        public Recipients() {};

        private Set<String> rcptSet = new HashSet<>();
        private String to = null;
        private String cc = null;
        private String bcc = null;

        public Recipients to(String address) {return rcpt(address, Type.TO, true);};
        public Recipients cc(String address) {return rcpt(address, Type.CC, true);};
        public Recipients bcc(String address) {return rcpt(address, Type.BCC, true);};

        /**
         * Fuegt einen Empfaenger hinzu.<br>
         * Wird von {@link #to(String)}, {@link #cc(String)}
         * und {@link #bcc(String)} gerufen.
         * @param address String: Email-Adresse. Alias-Form moeglich.
         * @param type {@link Type}: z.B. CC.
         * @param doSend boolean: Mail an Empfaenger senden ja/nein.<br>
         *      Wenn nein, dann wird er im Kopf nur gelistet,
         *      erhaelt aber keine Mail.<br>Ob sinnvoll, sei dahingestellt.
         * @return <code>this</code>
         */
        private Recipients rcpt(final String address, Type type, boolean doSend) {
            final String adr = new String(address.trim().getBytes(UTF8));
            final String rcpt = removeAlias(adr);
            if (!rcptSet.contains(rcpt)) {
                if (doSend) rcptSet.add(rcpt);
                switch (type) {
                case TO: this.to = (null == this.to ? adr : this.to + ", " + adr); break;
                case CC: this.cc = (null == this.cc ? adr : this.cc + ", " + adr); break;
                case BCC: this.bcc = (null == this.bcc ? adr : this.bcc + ", " + adr); break;
                }
            }
            return this;
        }
    }

    private static String removeAlias(final String addi) {
        int aliasIndex = addi.indexOf('<');
        if (-1 == aliasIndex) return addi;
        return addi.substring(aliasIndex +1, addi.length() -1).trim();
    }

}

