/*
 * $Id: Sockets.java 5995 2024-11-07 19:23:29Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Function;

import de.cm.osm2po.logging.Log;

/**
 * Zwei einfache Sockets fuer Client und Server.<br>
 * Messages sind hier Strings. Entgegen vieler Beispiele im Netz,
 * wird hier jedoch als String-Ende-Sequenz kein CRLF, sondern
 * Byte 0, also analog C erwartet. Das heisst, Strings duerfen
 * auch Zeilenumbrueche enthalten.
 */
public class Sockets {

    private static final String CLIENT = "SOC Client ";
    private static final String SERVER = "SOC Server ";

    private static final String BYE = "<BYE>";
    private static final String OK = "<OK>";
    private static final String SPLITTER = ","; // Caution with regexp e.g. |

    private Sockets() {/* STATIC */}

    private static String readInputStream(InputStream is) throws Exception {
        ByteArrayOutputStream baos = new ByteArrayOutputStream(0x400/*1k*/);

        for (;;) {
            int bint = is.read();
            if (-1 == bint) return null; // Socket died
            if ('\r' == bint) continue;
            if ('\n' == bint) break;
            baos.write(bint);
        }

        // Above possibly faster on small payloads than this here:
        /*
        byte[] buffer = new byte[0x100]
        for (;;) {
            int nRead = is.read(buffer);
            if (-1 == nRead) return null; // Socket died
            boolean lastChunk = buffer[nRead-1] == 0;
            if (lastChunk) nRead -= 1;
            baos.write(buffer, 0, nRead);
            if (lastChunk) break;
        }
        */

        return new String(baos.toByteArray());
    }

    private static void writeOutputStream(OutputStream os, String s) throws Exception {
        os.write(s.getBytes());
        os.write('\r');
        os.write('\n');
    }

    // ------------------------------------------------------------

    public static final class Server extends Thread {
        private static int SERVERID = 0;

        private final ServerSocket serverSocket;
        private final Log log;
        private final Set<ClientHandler> clientHandlerSet = new HashSet<>();
        private final Function<String, String> ioFunction;
        private final int serverId;
        private final String apiKey;

        public Server(int port, Log log, final String apiKey,
                Function<String, String> ioFunction) throws Exception {
            this.log = (null == log ? Log.stderr() : log);
            try {
                this.ioFunction = ioFunction;
                this.log.info(SERVER + "starting on Port " + port);
                this.serverSocket = new ServerSocket(port); // NULL if e.g. Port already in use
                this.serverId = ++SERVERID;
                this.apiKey = apiKey;
                this.setName("SocketServer" + this.serverId); // ThreadName
                this.start();

            } catch (Exception e) {
                this.log.error(SERVER + "not started. " + e);
                throw e;
            }
        }

        @Override
        public void run() {
            try {
                int nextClientId = 1;
                while (true) {
                    Socket clientSocket = Server.this.serverSocket.accept();
                    new ClientHandler(Server.this, clientSocket, nextClientId++,
                            this.log, this.apiKey, this.ioFunction).start();
                }
            } catch (SocketException se) {
                if (!this.serverSocket.isClosed()) log.error(se);
            } catch (Exception e) {
                log.fatal(e); // TODO handle this worst case
            }

        }

        public void close() {
            try {
                log.info(SERVER + "closing with "
                        + this.clientHandlerSet.size() + " active Clients");
                this.serverSocket.close(); // Dont accept new clients
                for (ClientHandler ch : new ArrayList<>(this.clientHandlerSet)) {
                    ch.clientSocket.close(); // forces Exception in ClientHandler
                    ch.join(); // Wait for each to shutdown
                }

            } catch (Exception e) {
                log.warn(SERVER + "close() failed with: " +  e);
            }
        }

        private static class ClientHandler extends Thread {
            private final Server server;
            private final Socket clientSocket;
            private final int clientId;
            private final Log log;
            private final String apiKey;
            private final Function<String, String> ioFunction;

            public ClientHandler(Server server, Socket socket,
                    int clientId, Log log, String apiKey, Function<String, String> io) {
                this.server = server;
                this.clientSocket = socket;
                this.clientId = clientId;
                this.log = log;
                this.apiKey = apiKey;
                this.ioFunction = io;
                this.setName(server.getName() + "Client" + clientId); // TheadName
            }

            @Override public void run() {
                log.info(SERVER + "accepted Client ID " + clientId);
                this.server.clientHandlerSet.add(this);

                try {
                    OutputStream os = this.clientSocket.getOutputStream();
                    final String clientInfo = this.clientId + SPLITTER + BYE;
                    writeOutputStream(os, clientInfo); // let client know it's ID and BYE-String
                    final String clientApiKey = readInputStream(this.clientSocket.getInputStream()).trim();
                    log.info(SERVER + "received AUTH from Client ID " + this.clientId);
                    final boolean authOk = null == this.apiKey || this.apiKey.equals(clientApiKey);

                    if (authOk) {
                        writeOutputStream(os, OK);
                    } else {
                        log.error(SERVER + "AUTH failed Client ID " + this.clientId);
                        writeOutputStream(os, "AUTH failed");
                    }

                    if (authOk) for (;;) {
                        String request = readInputStream(this.clientSocket.getInputStream());
                        if (null == request) {
                            log.info(SERVER + "Client died ID " + this.clientId);
                            break;
                        }
                        if (BYE.equals(request)) {
                            log.info(SERVER + "received " + request + " from Client ID " + this.clientId);
                            writeOutputStream(os, BYE + " received. Closing Connection ID " + this.clientId);
                            break;
                        }

                        final String responseLine = ioFunction.apply(request);
                        writeOutputStream(os, responseLine);
                    }

                    log.info(SERVER + "closing connection to Client ID " + this.clientId);
                    this.clientSocket.close();

                } catch (Exception e) {
                    this.log.info(SERVER + "Client closed by " + e + " ID " + this.clientId);
                }

                this.server.clientHandlerSet.remove(this);
            }
        }
    }

    // -------------------------------------------------------------


    public static final class Client {
        private Log log; // final will be not buildable, because ot not initialized? Java-Gaga

        private final String host;
        private final int port;
        private final String apiKey;

        private int clientId;
        private boolean offline;
        private boolean forbidden;
        private String bye;
        private Socket socket;

        public Client(String host, int port, String apiKey, Log log) {
            this.log = (null == log ? Log.stderr() : log); // must be first line
            this.host = host;
            this.port = port;
            this.apiKey = "" + apiKey; // not null
            this.connect(0); // Wichtig 1x synchron, setzt closed=true bei Fehler
            if (!this.forbidden) this.reconnector.start();
        }

        private Thread reconnector = new Thread(()->{
            long attempt = 1;
            final long maxSleep = 300000; // 5 Min.
            final long minSleep = 5000; // 5 Sek.
            while (true) {
                try {
                    if (this.offline) {
                        connect(attempt);
                        if (this.forbidden) break;
                        if (this.offline) attempt++; else attempt = 1;
                    }
                    long sleepFor = minSleep + (attempt * minSleep);
                    if (sleepFor > maxSleep) sleepFor = maxSleep;
                    Thread.sleep(sleepFor);
                } catch (InterruptedException ie) {
                    if (this.log != null) this.log.info(CLIENT + "Reconnector interrupted.");
                    break;
                } catch (Exception e) {
                    this.log.error(e);
                }
            }
        });

        private void closeSocket() {
            try {
                this.offline = true; // important for retry-loop
                if (this.socket != null) this.socket.close();
            } catch (Exception e){
                this.log.warn(CLIENT + "socket closed with " + e + " ID " + this.clientId);
            }
        }

        private void connect(long attempt) {
            try {
                this.offline = true;
                this.forbidden = false;

                this.log.info(CLIENT + "#" + this.hashCode() +  " connecting to "
                        + this.host + ":" + port + " [" + attempt + ']');
                this.socket = new Socket(host, port);
                final String serverHello = readInputStream(this.socket.getInputStream());
                final String[] clientInfos = serverHello.split(SPLITTER);
                for (int i = 0; i < clientInfos.length; i++) clientInfos[i] = clientInfos[i].trim();
                this.clientId = Integer.valueOf(clientInfos[0]);
                this.bye = clientInfos[1];
                this.log.info(CLIENT + "#" + this.hashCode() + " received Info " + serverHello);
                this.offline = false; // activate message sender
                this.forbidden = !OK.equals(this.sendMessage(this.apiKey));
                if (this.forbidden) this.closeSocket(); // sets offline true again

            } catch (Exception e) {
                this.log.error(CLIENT + "#" + this.hashCode() + " not connected - " + e);
                this.closeSocket();
            }
        }

        public final boolean isOffline() {
            return this.offline;
        }

        public final boolean isForbidden() {
            return this.forbidden;
        }

        public synchronized String sendMessage(String msg) {
            if (this.offline) return null;
            try {
                writeOutputStream(this.socket.getOutputStream(), msg);
                return readInputStream(this.socket.getInputStream());

            } catch (Exception e) { // ignore, log.error and return null instead
                log.error(CLIENT + "message not sent. " + e + " ID " + this.clientId);
                this.closeSocket();
                return null;
            }
        }

        public synchronized void close() {
            try {
                if (this.offline) {
                    this.log.warn(CLIENT + "already closed ID " + this.clientId);
                } else {
                    this.log.info(CLIENT + "closing connection ID " + this.clientId);
                    if (null == this.sendMessage(this.bye)) this.log.info(
                            CLIENT + "Server already closed connection ID " + this.clientId);
                    this.reconnector.interrupt();
                    this.reconnector.join();
                    this.closeSocket();
                }
            } catch (Exception e) {
                this.log.info(CLIENT + "closed by " + e + " ID " + this.clientId);
            }
        }
    }

}

