/*
 * $Id: RateLimitFilter.java 5991 2024-11-01 10:20:54Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.filters;

import static de.cm.osm2po.service.Osm2poService.getHeaderValue;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_ACCOUNT_EXPIRED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_DOWN_MAINTENANCE;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_RATELIMIT_EXCEEDED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNAUTHORIZED;
import static de.cm.osm2po.service.security.RateLimit.Status.SERVICEDOWN;
import static de.cm.osm2po.service.security.RateLimit.Status.UNKNOWN;
import static de.cm.osm2po.service.security.RateLimit.Status.VALID;
import static de.cm.osm2po.service.security.RateLimiter.validationParamsToString;

import java.util.Date;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import com.sun.net.httpserver.Authenticator;
import com.sun.net.httpserver.HttpPrincipal;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.RateLimitServer;
import de.cm.osm2po.service.ServiceFilter;
import de.cm.osm2po.service.ServiceRequest;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.Sockets.Client;
import de.cm.osm2po.service.errors.ServiceException;
import de.cm.osm2po.service.security.RateLimit;
import de.cm.osm2po.service.security.RateLimit.Status;
import de.cm.osm2po.service.security.RateLimiter;

/**
 * Ein {@link ServiceFilter}, der Aufrufe mittels {@link RateLimiter}
 * ueberwacht. Sonderrechte koennen per ApiKey (x-api-key) aber auch
 * {@link HttpPrincipal} (via {@link Authenticator} im {@link Osm2poService})
 * getriggert werden.<br>
 * Anders als der {@link RateLimitFilter}, baut diese Variante ueber einen
 * {@link Client} eine Verbindung zu einem {@link RateLimitServer} auf.
 * <strong>Config Beispiel:</strong>
 * <pre>
 * service.filter.1.class=de.cm.osm2po.service.filters.RateLimitFilterClient
 * service.filter.1.host=localhost (default)
 * service.filter.1.port={@link RateLimitServer#DEFAULT_PORT}
 * </pre>
 * @see RateLimit
 * @see RateLimiter
 * @see RateLimitServer
 * @author (c) 2024 - Carsten Moeller, info@osm2po.de
 */
public class RateLimitFilter implements ServiceFilter {
    private final static String HTX_RLC_RESUME = "HTX_RLC_RESUME_KEY";

    private Set<String> ignoreCmdSet;
    private Set<String> resumeCmdSet;
    private Log log;
    private Client clientSocket;

    @Override
    public void init(Config config, int index) {
        this.log = config.getLog();
        if (null == this.log) this.log = Log.stderr();

        final String prefix = "service.filter." + index + ".";

        this.ignoreCmdSet = new HashSet<String>(0x20);
        String ignoreCmdCsv = config.getProperty(prefix + "ignoreCmds");
        if (ignoreCmdCsv != null) for (String cmd : ignoreCmdCsv.split(",")) {
            this.ignoreCmdSet.add(cmd.trim());
        }
        this.resumeCmdSet = new HashSet<String>(0x20);
        String resumeCmdCsv = config.getProperty(prefix + "resumeCmds");
        if (resumeCmdCsv != null) for (String cmd : resumeCmdCsv.split(",")) {
            this.resumeCmdSet.add(cmd.trim());
        }

        final String host = config.getProperty(prefix + "host", "localhost");
        final int port = Integer.valueOf(config.getProperty(prefix + "port",
            String.valueOf(RateLimitServer.DEFAULT_PORT)));
        final String apiKey = config.getProperty(prefix + "apikey");

        log.info("Starting " + this.getClass().getSimpleName() + " at " + new Date().toString());
        this.clientSocket = new Client(host, port, apiKey, this.log);
    }

    @Override
    public boolean preHandle(ServiceRequest sr) throws Exception {
        sr.htx.setAttribute(HTX_RLC_RESUME, null);

        if (this.clientSocket.isOffline())
            throw new ServiceException(ERR_DOWN_MAINTENANCE);

        String cmd = sr.cmd;
        if (null == cmd) {
            cmd = "NULL";
        }
        if (this.ignoreCmdSet.contains(cmd)) {
            return false;
        }

        String userKey = Optional.ofNullable(sr.htx.getPrincipal())
                .map(p->p.getName() + "@" + p.getRealm())
                .orElse(null);
        if (null == userKey) { // Wenn kein BasicAuth
            userKey = getHeaderValue(sr.htx.getRequestHeaders(), "x-api-key");
            if (null == userKey) userKey = sr.params.getProperty("xak");
        }

        final String request = validationParamsToString(sr.ip, userKey, sr.cmd);
        final String response = this.clientSocket.sendMessage(request);

        Status status = UNKNOWN;
        if (null == response) {
            status = SERVICEDOWN;
        } else {
            status = Status.valueOf(response);
        }

        if (status != VALID) {
            // if (KEY_DDOS == status) Thread.sleep(1000);
            switch (status) {
            case RATELIMIT: throw new ServiceException(ERR_RATELIMIT_EXCEEDED);
            case EXPIRED: throw new ServiceException(ERR_ACCOUNT_EXPIRED);
            case SERVICEDOWN: throw new ServiceException(ERR_DOWN_MAINTENANCE);
            default: throw new ServiceException(ERR_UNAUTHORIZED); // auch bei DDOS
            }
        } else if (userKey != null && this.resumeCmdSet.contains(sr.cmd)) {
            // Fuer Fehler von ausserhalb, Fuer PostHandle merken.
            final String resumeKey = userKey + "," + sr.cmd;
            sr.htx.setAttribute(HTX_RLC_RESUME, resumeKey);
        }

        return false;
    }

    @Override
    public void shutDown() {
        this.clientSocket.close();
    }

    @Override public void postHandle(
            ServiceRequest req, ServiceResponse resp, boolean sent) {

        if (this.clientSocket.isOffline()) return;

        if (!sent || null == resp || resp.httpCode != 200) { // Wenn externer Fehler
            // Ein Benutzer bezahlt nur fuer erfolgreiche Berechnungen, daher
            // muss das RateLimit bei externem Fehler rehabilitiert werden.
            // Sicherheitsluecke: Wir wissen hier nicht, ob es ein valider User ist,
            // sondern nur, dass ein AuthKey uebergeben wurde, wenn RateLimit != null.
            final String resumeKey = (String) req.htx.getAttribute(HTX_RLC_RESUME);
            if (resumeKey != null) {
                log.error(getClass().getSimpleName() + resumeKey + " counter must be decremented");
            }
        }
    }

}
