/*
 * $Id: TspHandler.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.handlers;

import static de.cm.osm2po.misc.ArrayUtils.intArrayConcat;
import static de.cm.osm2po.misc.Utils.DF4;
import static de.cm.osm2po.misc.Utils.TS;
import static de.cm.osm2po.misc.Utils.msToDhms;
import static de.cm.osm2po.routing.RoutingResultWriter.writePathAsCsv;
import static de.cm.osm2po.routing.RoutingResultWriter.writePathAsFeatureCollection;
import static de.cm.osm2po.service.ServiceHelpers.parseDoubleProperty;
import static de.cm.osm2po.service.ServiceHelpers.parseNode;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TSP_NOT_FOUND;
import static de.cm.osm2po.tsp.Tsp2po.createMatrix;
import static de.cm.osm2po.tsp.Tsp2po.solve;
import static de.cm.osm2po.tsp.TspUtils.isSymmetryGiven;

import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.routing.AbstractGraph;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayEdge;
import de.cm.osm2po.routing.OverlayGraph;
import de.cm.osm2po.routing.OverlayNode;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.ServiceHandler;
import de.cm.osm2po.service.ServiceResourcesChecker;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ServiceException;
import de.cm.osm2po.tsp.Tsp2po.Matrix;
import de.cm.osm2po.tsp.TspUtils;

/**
 * Behandelt Requests des {@link Osm2poService}.
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class TspHandler implements ServiceHandler {
    @Override
    public ServiceResponse handle(final String cmd, Config config, Graph graph,
            Properties request, ServiceResourcesChecker lttc) throws Exception {

        boolean isGeoJson = "geojson".equals((request.getProperty("format")));
        Log log = config.getLog();

        if ("ft".equals(cmd)) { // Find TSP
            lttc.checkResources();

            final int baosSize = isGeoJson ? 0x400000/*4M*/ : 0x40000/*256k*/;
            ByteArrayOutputStream baos = new ByteArrayOutputStream(baosSize);

            // Router-Parameter abgreifen und validieren
            float maxCost = (float) parseDoubleProperty(request, "maxCost", 0);
            if (maxCost == 0) maxCost = Float.MAX_VALUE;

            final double tolerance = parseDoubleProperty(request, "heuristicFactor", 0d);
            final String tspChain = request.getProperty("tspChain");

            OverlayGraph overlayGraph = new OverlayGraph(graph);
            String[] strNodes = request.getProperty("tsp", "").split(",");
            int nVertices = strNodes.length;
            int[] vertexIds = new int[strNodes.length];
            boolean ovlMode = false;
            for (int i = 0; i < strNodes.length; i++) {
                OverlayNode overlayNode = overlayGraph.insertNode(
                        parseNode(graph, strNodes[i].trim(), true, false));
                ovlMode |= !overlayNode.isOnVertex();
                vertexIds[i] = overlayNode.getId();
            }
            if (ovlMode) log.debug("OverlayNodes present");

            final boolean isRing = vertexIds[0] == vertexIds[nVertices-1];
            if (isRing) vertexIds = Arrays.copyOf(vertexIds, nVertices -1);
            final int tspN = nVertices -2;

            // TSP-Matrix bauen
            log.info("TSP [" + tspN + "] Creating Matrix " + TS());
            AbstractGraph ag = ovlMode ? overlayGraph : graph;
            final int dbl2LongFactor = 100000; // ~ cm oder 10tel Sekunde
            Matrix mx = createMatrix(ag, vertexIds, maxCost, log, request, dbl2LongFactor);

            if (log.isEnabled(Level.NOTE)) {
                log.note("TSP [" + tspN + "] Symmetry given: " + isSymmetryGiven(mx.costs));
            }
            if (log.isEnabled(Level.DEBUG)) {
                log.debug("TSP [" + tspN + "]\n" + TspUtils.prettyPrintMatrix(mx.costs));
            }

            // TSP berechnen
            log.info("TSP [" + tspN + "] Solving Tour - " + TS());
            long tspTime = System.currentTimeMillis();
            final int[] solvedTour = solve(mx, log, isRing, tolerance, tspChain);
            if (null == solvedTour)throw new ServiceException(ERR_TSP_NOT_FOUND);

            tspTime = System.currentTimeMillis() - tspTime;
            log.info("TSP [" + tspN + "] Tour solved in " + msToDhms(tspTime) + " - " + TS());

            int[][] paths = mx.getTourPaths(solvedTour);

            double sumLength = 0, sumTime = 0;
            for (int i = 0; i < paths.length; i++) {
                int sourceId = vertexIds[solvedTour[i]];
                int targetId = vertexIds[solvedTour[i+1]];

                double length = overlayGraph.calcPathLength(paths[i]);
                double time = overlayGraph.calcPathTime(paths[i]);
                sumLength += length;
                sumTime += time;

                log.debug("P[" + i + "] from " +  sourceId
                        + " to " + targetId + " : " + DF4(length) + "km " + DF4(time) + "h");
            }

            int[] path = intArrayConcat(paths);
            OverlayEdge[] route = overlayGraph.translatePath(path);
            if (isGeoJson) writePathAsFeatureCollection(baos, graph, route);
            else writePathAsCsv(baos, graph, route);
            // Example: http://localhost:8888/Osm2poService
            // ?cmd=ft&tsp=20583,23537,5303,3142,1834,20583&format=geojson

            log.note("TSP [" + tspN + "] Order : "
                    + Arrays.toString(solvedTour));
            log.note("TSP [" + tspN + "] Total : "
                    + DF4(sumLength) + "km " + DF4(sumTime) + "h");

            byte[] response = baos.toByteArray();
            return new ServiceResponse(HTTP_200, isGeoJson ? MIME_JSON : MIME_TEXT, response);
        }

        return null;
    }

}