/*
 * $Id: BasicAuth.java 5925 2024-08-17 09:46:58Z carsten $
 *
 * Copyright 2023 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.security;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import com.sun.net.httpserver.Authenticator;
import com.sun.net.httpserver.HttpExchange;

import de.cm.osm2po.Config;
import de.cm.osm2po.misc.Base64;
import de.cm.osm2po.service.errors.ErrorEnum;

/**
 * Simple BasicAuth.
 * @see #checkCredentials(String, String)
 * @author (c) 2023 - Carsten Moeller - info@osm2po
 */
public class BasicAuth extends Auth {

    protected String realm;

    private boolean handleFailure;

    @Override
    public void init(Config config) {
        this.realm = config.getProperty("service.auth.basic.realm", "osm2po");
        this.handleFailure = Boolean.valueOf(config.getProperty("service.auth.handlefailure", "true"));
    }

    @Override
    public final Result authenticate(HttpExchange htx) {

        String auth = htx.getRequestHeaders().getFirst("Authorization");
        if (auth == null) {
            htx.getResponseHeaders().set("WWW-Authenticate",
                    "Basic realm=" + "\"" + this.realm + "\"");
            return new Authenticator.Retry(401);
        }
        if (!auth.startsWith("Basic ")) {
            return basicFailure(401);
        } else {
            auth = auth.substring(6);
        }

        String[] userpass = new String(Base64.decode(auth)).split(":");

        // Pitfall: userpass < 2 if user clicks ok without having entered either or both
        if (userpass.length == 2) {
            final String user = userpass[0];
            final String pass = userpass[1];

            Set<String> roles = checkCredentials(user, pass);
            if (null == roles) {
                htx.getResponseHeaders().set("WWW-Authenticate",
                    "Basic realm=" + "\"" + this.realm + "\"");
                return basicFailure(401);
            } else {
                return new Authenticator.Success(new Principal2po(
                    user + '-' + Math.abs(pass.hashCode()), this.realm, roles));
            }
        } else {
            htx.getResponseHeaders().set("WWW-Authenticate",
                    "Basic realm=" + "\"" + this.realm + "\"");
            return basicFailure(401);
        }
    }

    /**
     * Checks Username and Password and returns a Set of Roles.
     * This implementation just asserts that both have
     * at least 5 characters. Must be overwritten in
     * real life, of course.
     * @param username String NOT NULL
     * @param password String NOT NULL
     * @return Set of Roles (may be empty) or NULL for unauthorized
     */
    public Set<String> checkCredentials(String username, String password) {
        // Just some Dummy-Checks
        if (username.contains(" ")) return null;
        if (password.contains(" ")) return null;
        if (username.length() < 5) return null;
        if (password.length() < 5) return null;
        return new HashSet<>(Arrays.asList("ROLE_GUEST"));
    }

    private Result basicFailure(int httpCode) {
        if (this.handleFailure) {
            return new Authenticator.Failure(httpCode);
        } else {
            // TODO check httpCode for correct Error
            return new Authenticator.Success(
                new Principal2po(ErrorEnum.ERR_UNAUTHORIZED, this.realm));
        }
    }

}
