/*
 * $Id: OAuth.java 5956 2024-10-22 16:04:45Z carsten $
 *
 * Copyright 2024 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.security;

import static de.cm.osm2po.misc.Utils.UTF8;
import static de.cm.osm2po.misc.Utils.loadStreamIntoByteArray;
import static de.cm.osm2po.misc.Utils.readArgsIntoProps;
import static de.cm.osm2po.reflection.JsonBuilder.prettyPrint;
import static de.cm.osm2po.service.Cookie.SameSite.LAX;
import static de.cm.osm2po.service.Cookie.SameSite.STRICT;

import java.security.interfaces.RSAPublicKey;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Properties;
import java.util.Set;
import java.util.UUID;

import com.sun.net.httpserver.Authenticator;
import com.sun.net.httpserver.HttpExchange;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.Base64;
import de.cm.osm2po.misc.HttpClient;
import de.cm.osm2po.misc.HttpClient.HttpClientResponse;
import de.cm.osm2po.reflection.JsonDeserializer;
import de.cm.osm2po.reflection.JsonParser;
import de.cm.osm2po.reflection.Reflector.Reflectable;
import de.cm.osm2po.service.Cookie;
import de.cm.osm2po.service.errors.ErrorEnum;

/**
 * @author (c) 2024 - Carsten Moeller - info@osm2po
 */
public class OAuth extends Auth {

    private Log log;

    private String providerName;
    private String realm;
    private String realmUrl;
    private String clientId;
    private String clientSecret;
    private String scope;
    private String redirectUrl;
    private String backChannelLogout;
    private RealmInfo realmInfo;

    private String rolesPath;

    private boolean handleFailure;
    private String cookieSecretKey;

    private Set<String> logouts = new HashSet<>();

    private final static String COOKIE_ACCESS_TOKEN = "OAUTH2PO_ACCESS";
    private final static String COOKIE_REFRESH_TOKEN = "OAUTH2PO_REFRESH";
    private final static String COOKIE_ID_TOKEN = "OAUTH2PO_ID";
    private final static String COOKIE_XSRF = "OAUTH2PO_XSRF";

    @Override
    public void init(Config config) {
        this.log = config.getLog();

        final String prefix = config.getProperty("service.auth.provider.prefix", "kc");
        final String defaultUrl = config.getProperty("service.url.https");

        this.providerName = config.getProperty("service.auth.provider.name", "KeyCloak");
        this.handleFailure = Boolean.valueOf(config.getProperty("service.auth.handlefailure", "true"));
        this.cookieSecretKey = config.getProperty("service.auth.cookie.secret");

        this.realm = config.getProperty(prefix + ".realm.name");
        this.realmUrl = config.getProperty(prefix + ".realm.url");
        this.clientId = config.getProperty(prefix + ".realm.client.id");
        this.clientSecret = config.getProperty(prefix + ".realm.client.secret");
        this.scope = config.getProperty(prefix + ".realm.client.scope", "").trim().replace('+', ' ');
        this.redirectUrl = config.getProperty(prefix + ".realm.client.redirect.url", defaultUrl);
        this.backChannelLogout = config.getProperty(prefix + ".realm.client.backchannel.logout", "/kc/logout");

        this.rolesPath = config.getProperty(prefix + ".realm.client.roles.path", "resource_access." + this.clientId + ".roles");

        if (this.cookieSecretKey != null && this.cookieSecretKey.length() < 24) {
            this.cookieSecretKey = null;
            log.fatal("Cookie encryption disabled");
            log.error("SECRET KEY for Cookie must have at least 24 characters!");
        }

        this.initRealmInfo();
    }

    private void initRealmInfo() {
        final HttpClientResponse resp = new HttpClient(this.realmUrl).doGetRequest();
        if (resp.isJson200()) {
            final String jsonRealmInfo = new String(resp.responseBody, UTF8);
            if (log.isEnabled(Level.DEBUG)) {
                log.debug("OAUTH IdP is " + this.providerName);
                log.debug("REALM INFO\n" + prettyPrint(jsonRealmInfo));
            }
            this.realmInfo = JsonDeserializer.deserialize(jsonRealmInfo, RealmInfo.class);

        } else {
            log.fatal("RealmInfos not yet fetched - IdP online?");
        }
    }

//    private String getUsersOfRealmAsJson() { // TODO JUST A TEST
//        // ClientCredentialsFlow not possible in this version. Known Bug.
//        logOAuthTask("LIST OF USERS");
//        TokenTriplet tt = this.oauthPasswordGrant("admin-cli", "admin", "admin");
////        TokenTriplet tt = this.oauthClientCredentialsGrant(this.clientId, this.clientSecret);
//        log.debug(tt.toString());
//
//        HttpClientResponse resp = new HttpClient(
//                "http://foo.bar.baz:8080/admin/realms/master/users?enabled=true") // must match token issuer "iss"
//                .putBearer(tt.accessToken)
//                .doGetRequest();
//
//        if (!resp.isJson200()) return null;
//
//        final String json = new String(resp.responseBody, UTF8);
//        return prettyPrint(json);
//    }

    private static enum TokenState {
        VERIFIED, EMAILNOTVERIFIED, EXPIRED,
        WRONGAUTHPARTY, WRONGSIGNATURE, WRONGFORMAT,
        NULLTOKEN, BACKCHANNELLOGOUT
    }

    public static class TokenHeader {
        public String alg;
        public String kid;
    }

    public static class Jwt {
        public final String[] rawParts;

        public final String header;
        public final String payload;
        public final byte[] signature;

        public final Object parsedPayload;
        public final Long exp; // Expires in Unix-Time
        public final String sub; // Subject (User)
        public final String azp; // Authorized Party
        public final String aud; // Audience
        public final String iss; // Issuer
        public final Boolean emailVerified;


        /**
         * @param urlEncodedToken -Parts as String of pattern ey123.ey345.xyz
         */
        public Jwt(String urlEncodedToken) {
            this.rawParts = urlEncodedToken.split("\\.");
            this.header = new String(Base64.decodeUrl(rawParts[0]), UTF8);
            this.payload = new String(Base64.decodeUrl(rawParts[1]), UTF8);
            this.signature = Base64.decodeUrl(rawParts[2]);

            this.parsedPayload = JsonParser.parse(this.payload);

            this.exp = (Long) JsonParser.getValue(this.parsedPayload, "exp");
            this.sub = (String) JsonParser.getValue(this.parsedPayload, "sub");
            this.aud = (String) JsonParser.getValue(this.parsedPayload, "aud");
            this.azp = (String) JsonParser.getValue(this.parsedPayload, "azp");
            this.iss = (String) JsonParser.getValue(this.parsedPayload, "iss");
            this.emailVerified = Boolean.TRUE.equals(
                    JsonParser.getValue(this.parsedPayload, "email_verified"));
        }
    }

    public static class RealmInfo {
        @Reflectable(name="realm") public String realm;
        @Reflectable(name="public_key")public String publicKey;
        @Reflectable(name="token-service")public String tokenServiceUrl;
        @Reflectable(name="account-service")public String accountServiceUrl;
    }

    private static class TokenTriplet {
        public final String accessToken;
        public final String refreshToken;
        public final String idToken;

        private TokenTriplet(String accessToken, String refreshToken, String idToken) {
            this.accessToken = accessToken;
            this.refreshToken = refreshToken;
            this.idToken = idToken;
        }

        @Override
        public String toString() {
            String s = "";
            if (accessToken != null) s += "\n" + tokenToString("ACCESSTOKEN", accessToken);
            if (refreshToken != null) s += "\n\n" + tokenToString("REFRESHTOKEN", refreshToken);
            if (idToken != null) s += "\n\n" + tokenToString("IDTOKEN", idToken);
            return s;
        }

        /**
         * Sets all NON-NULL Tokens (Access, Refresh and ID) as Cookies in the Response-Headers.
         * Except for the AccessToken, which will be returned with SameSite=LAX,
         * the other ones are kept as STRICT. This is because of the Redirect behavior and the
         * interpretation of most Browsers, that regard this to be a second party call,
         * like a link.
         * @param htx {@link HttpExchange}
         * @param secretKey NULLABLE key for Cookie-Encryption
         */
        public final void addToResponseCookies(HttpExchange htx, final String secretKey) {
            if (this.accessToken != null) Cookie.builder(
                    COOKIE_ACCESS_TOKEN, this.accessToken)
                .secure().httpOnly().path("/").sameSite(LAX) // Important! Must not be strict
                .expires(new Jwt(this.accessToken).exp)
                .encrypt(secretKey)
                .buildAndAddToResponse(htx);
            if (this.refreshToken != null) Cookie.builder(
                    COOKIE_REFRESH_TOKEN, this.refreshToken)
            .secure().httpOnly().path("/").sameSite(STRICT)
                .expires(new Jwt(this.refreshToken).exp)
                .encrypt(secretKey)
                .buildAndAddToResponse(htx);
            if (this.idToken != null) Cookie.builder(
                    COOKIE_ID_TOKEN, this.idToken)
            .secure().httpOnly().path("/").sameSite(STRICT)
                .expires(new Jwt(this.idToken).exp)
                .encrypt(secretKey)
                .buildAndAddToResponse(htx);
        }
    }

    private Properties readQueryOrPostParamsIntoProps(HttpExchange htx) {
        final String ctype = htx.getRequestHeaders().getFirst("Content-Type");
        final String method = htx.getRequestMethod();
        String params = Optional.ofNullable(htx.getRequestURI().getQuery()).orElse("");
        if ("POST".equals(method) && "application/x-www-form-urlencoded".equalsIgnoreCase(ctype)) {
            try {
                byte[] data = loadStreamIntoByteArray(htx.getRequestBody());
                params = new String(data, UTF8);
            } catch (Exception e) {
                log.error(e);
                params = "";
            }
        }
        return readArgsIntoProps(params.split("&"));
    }

    private void logOAuthTask(final String info) {
        log.info(this.providerName + "." + this.realm + "." + this.clientId + ": " + info);
    }

    @Override
    public Result authenticate(HttpExchange htx) {
        if (null == this.realmInfo) {
            if (isPageCall(htx)) this.initRealmInfo(); // Try again
            if (null == this.realmInfo) { // Failed again
                return oauthFailure(500);
            }
        }

        final String requestPath = htx.getRequestURI().getPath();
        // TODO handle paths with trailing / in subsequent comparisons

        // BACKCHANNEL-Handling (Logout)--------------------------
        if ("POST".equals(htx.getRequestMethod())
                && requestPath.equals(this.backChannelLogout)) {
            return oauthBackChanelLogout(htx);
        }

        // CLIENT-Handling ----------------------------------------------------

        final Properties queryParams = readQueryOrPostParamsIntoProps(htx);

        // PASSWORD-GRANT (Login) ----------------------------
        // TEST WITH https://localhost/login?user=carsten&pass=coffee
        if (requestPath.endsWith("/login")) { // TODO Pitfall may be /login/
            TokenTriplet tokenTriplet = this.oauthPasswordGrantFromRemote(htx);
            if (null == tokenTriplet) {
                return oauthFailure(401);
            }
            if (log.isEnabled(Level.DEBUG)) log.debug("Received Tokens:" + tokenTriplet.toString());
            tokenTriplet.addToResponseCookies(htx, this.cookieSecretKey);
            htx.getResponseHeaders().set("Location", this.redirectUrl);
            return new Authenticator.Retry(302);
        }

        // CODE-GRANT-EXCHANGE ----------------------------
        final String code = queryParams.getProperty("code");
        if (code != null) { // AuthorizationCodeGrant - Get Token for Code
            // TODO just checking for "code" is too naive. Should be more specific, e.g. /code-Endpoint
            // TODO Handle "NO CONSENT"-Click from KC -> EndlessLoop cause code=null

            final String state = queryParams.getProperty("state", "");
            String csrf = Cookie.fromRequest(htx, this.cookieSecretKey).get(COOKIE_XSRF);
            if (!state.equals(csrf)) {
                return oauthFailure(403);
            }

            final TokenTriplet tokenTriplet = this.oauthCodeGrantExchangeCode(code);
            if (null == tokenTriplet) {
                return oauthFailure(401);
            }

            if (log.isEnabled(Level.DEBUG)) log.debug("Received Tokens:" + tokenTriplet.toString());
            tokenTriplet.addToResponseCookies(htx, this.cookieSecretKey);
            Cookie.builder(COOKIE_XSRF, "").maxAge(0).buildAndAddToResponse(htx); // delete it
            htx.getResponseHeaders().set("Location", this.redirectUrl);
            // This Redirect forces the Browser to elimitate the the response from the
            // keycloak with all its parameters and switch back to root URL
            return new Authenticator.Retry(302); // Redirect Dance
        }

        String accessToken = Optional.ofNullable(Optional.ofNullable(
            htx.getRequestHeaders().getFirst("Authorization"))
                .filter(a->a.startsWith("Bearer "))
                .map(b->b.substring(7))
                .orElse(Cookie.fromRequest(htx, this.cookieSecretKey).get(COOKIE_ACCESS_TOKEN)))
                .orElse(null);

        final String refreshToken = Cookie
                .fromRequest(htx, this.cookieSecretKey)
                .get(COOKIE_REFRESH_TOKEN);

        final TokenState tokenState = verifyToken(accessToken, this.realmInfo.publicKey);
        final boolean pageCall = isPageCall(htx);

        // CAUTION: Many FallThru's here in case block
        switch (tokenState) {
        case BACKCHANNELLOGOUT:
            // Delete cookie via max age 0
            Cookie.builder(COOKIE_ACCESS_TOKEN, "")
                .maxAge(0).path("/").buildAndAddToResponse(htx);
            Cookie.builder(COOKIE_REFRESH_TOKEN, "")
                .maxAge(0).path("/").buildAndAddToResponse(htx);
            Cookie.builder(COOKIE_ID_TOKEN, "")
                .maxAge(0).path("/").buildAndAddToResponse(htx);
            htx.getResponseHeaders().set("Location", this.redirectUrl);
            return new Authenticator.Retry(302);

        case EXPIRED: // If Cookie expired, then never evaluated
        case NULLTOKEN: // May also accur when AccessToken-Cookie expired
            if (null == refreshToken) { // also after logout
                return pageCall ? this.oauthCodeGrantStart(htx) : oauthFailure(401);
            } else {
                TokenTriplet tokenTriplet = this.oauthRefreshTokens(htx, refreshToken);
                if (null == tokenTriplet) { // refresh failed
                    return pageCall ? this.oauthCodeGrantStart(htx) : oauthFailure(401);
                }

                if (log.isEnabled(Level.DEBUG)) log.debug("Reveived Tokens:" + tokenTriplet.toString());
                tokenTriplet.addToResponseCookies(htx, this.cookieSecretKey);
                accessToken = tokenTriplet.accessToken;

                // FALL THRU TO VERIFIED with new valid access token
                // ... but do more checks here in future
            }

        case VERIFIED:
            Jwt jwt = new Jwt(accessToken);
            String userId = jwt.sub;

            // USER-Logout-Handling --------------------------
            if (requestPath.equals("/user/logout")) {
                return oauthUserLogout(userId, refreshToken, htx);
            }

            Set<String> roleSet = null;
            Object roles = JsonParser.getValue(jwt.parsedPayload, this.rolesPath);

            if (roles != null) {
                @SuppressWarnings("unchecked")
                List<String> roleList = (List<String>) roles;
                roleSet = new HashSet<>(roleList);
            }

            return new Authenticator.Success(
                    new Principal2po(userId, this.realm, roleSet));

        case WRONGSIGNATURE: // FALL THRU TO Default (Error)
        case WRONGFORMAT:
        case EMAILNOTVERIFIED:
        default:
            return oauthFailure(401); // UnAuthorized : even works with osm2po-UI-REST-Calls
        }
    }

    private static boolean isPageCall(HttpExchange htx) {
        // Try to find out if this is a normal call from an HtmlPage
        // and no JSON-XMLHttpRequest, so we can send redirects to the
        // Browser and not into the dark tunnel.
        // TODO check X-requested-with=[XMLHttpRequest] - Header as well.
        final String accept = htx.getRequestHeaders().getFirst("Accept");
        final boolean pageCall = accept != null && accept.contains("html");
        return pageCall;
    }

    private Result oauthFailure(int httpCode) {
        if (this.handleFailure) {
            return new Authenticator.Failure(httpCode);
        } else {
            // TODO check httpCode for correct Error
            return new Authenticator.Success(
                new Principal2po(ErrorEnum.ERR_UNAUTHORIZED, this.realm));
        }
    }

    private Result oauthBackChanelLogout(HttpExchange htx) {
        logOAuthTask("BACKCHANNEL LOGOUT"); // Called by e.g. KeyCloak even after User-Logout

        final Properties params = readQueryOrPostParamsIntoProps(htx);
        final String logoutToken = params.getProperty("logout_token");
        if (logoutToken != null) {
            log.warn(tokenToString("LOGOUTTOKEN", logoutToken));

            if (this.verifyToken(logoutToken, this.realmInfo.publicKey) != TokenState.VERIFIED) {
                return new Authenticator.Failure(403 /*Forbidden*/); // Response for IdP
            }

            Jwt parts = new Jwt(logoutToken);

            Object payload = JsonParser.parse(parts.payload);
            final String sub = (String) JsonParser.getValue(payload, "sub");
            if (sub != null) this.logouts.add(sub);
        }

        return new Authenticator.Success( // Response to IdP
                new Principal2po(this.providerName, this.realm, null));
    }

    private Result oauthUserLogout(String userId, String refreshToken, HttpExchange htx) {
        logOAuthTask("USER LOGOUT " + userId); // Called by e.g. Logout-Click in Frontend

        final String redirectUrl = ""
                + this.realmInfo.tokenServiceUrl + "/logout"
                + "?client_id=" + this.clientId
                + "&redirect_url=" + this.redirectUrl;

        htx.getResponseHeaders().set("Authorization", "Bearer " + refreshToken);
        htx.getResponseHeaders().set("Location", redirectUrl);
        return new Authenticator.Retry(302);
    }

    private TokenTriplet oauthPasswordGrantFromRemote(HttpExchange htx) {
        // PitFall: CONSENT to scopes must be disabled for this!
        // see: https://keycloak.discourse.group/t/grant-type-password-fails-if-client-consent-is-required/24805

        logOAuthTask("PASSWORD GRANT LOGIN");

        final Properties params = readQueryOrPostParamsIntoProps(htx);
        final String user = params.getProperty("user","");
        final String pass = params.getProperty("pass","");

        HttpClient client = new HttpClient(this.realmInfo.tokenServiceUrl)
                .setContentTypeForm()
                .putFormParam("grant_type", "password")
                .putFormParam("username", user)
                .putFormParam("password", pass)
                .putFormParam("client_id", this.clientId)
                .putFormParam("client_secret", this.clientSecret);

        final boolean hasScope = this.scope != null && this.scope.length() > 0;
        if (hasScope) {
            client.putFormParam("scope", this.scope);
        }

        HttpClientResponse response =  client.doPostRequest("/token");
        if (!response.isJson200()) return null;

        final String json = new String(response.responseBody, UTF8);
        Object jwt = JsonParser.parse(json);

        return new TokenTriplet(
                (String) JsonParser.getValue(jwt, "access_token"),
                (String) JsonParser.getValue(jwt, "refresh_token"),
                (String) JsonParser.getValue(jwt, "id_token"));
    }

//    private TokenTriplet oauthPasswordGrant(
//            final String clientId, final String user, final String pass) {
//
//        logOAuthTask("PASSWORD GRANT LOGIN");
//
//        HttpClient client = new HttpClient(this.realmInfo.tokenServiceUrl)
//                .setContentTypeForm()
//                .putFormParam("grant_type", "password")
//                .putFormParam("username", user)
//                .putFormParam("password", pass)
//                .putFormParam("client_id", clientId);
//
//        HttpClientResponse response =  client.doPostRequest("/token");
//        if (!response.isJson200()) return null;
//
//        final String json = new String(response.responseBody, UTF8);
//        Object jwt = JsonParser.parse(json);
//
//        return new TokenTriplet(
//                (String) JsonParser.getValue(jwt, "access_token"),
//                (String) JsonParser.getValue(jwt, "refresh_token"),
//                (String) JsonParser.getValue(jwt, "id_token"));
//    }

//    private TokenTriplet oauthClientCredentialsGrant(
//            final String clientId, final String clientSecret) {
//
//        logOAuthTask("CLIENT CREDENTIALS GRANT");
//
//        HttpClient client = new HttpClient(this.realmInfo.tokenServiceUrl)
//                .setContentTypeForm()
//                .putFormParam("grant_type", "client_credentials")
//                .putFormParam("client_id", clientId)
//                .putFormParam("client_secret", clientSecret);
//
//        HttpClientResponse response =  client.doPostRequest("/token");
//        if (!response.isJson200()) return null;
//
//        final String json = new String(response.responseBody, UTF8);
//        Object jwt = JsonParser.parse(json);
//
//        return new TokenTriplet(
//                (String) JsonParser.getValue(jwt, "access_token"),
//                (String) JsonParser.getValue(jwt, "refresh_token"),
//                (String) JsonParser.getValue(jwt, "id_token"));
//    }

    private Result oauthCodeGrantStart(HttpExchange htx) {
        logOAuthTask("CODE GRANT START");

        // e.g. forces Keycloak-Confirm-Dialog, if consent configured,
        // otherwise no userinfo in token
        final boolean hasScope = this.scope != null && this.scope.length() > 0;

        final String csrf = UUID.randomUUID().toString();

        final String redirectUrl = ""
                + this.realmInfo.tokenServiceUrl + "/auth?response_type=code"
                + "&client_id=" + this.clientId
                + "&redirect_uri=" + this.redirectUrl
                + (hasScope ? "&scope=" + this.scope : "")
                + "&state=" + csrf;


        Cookie.builder(COOKIE_XSRF, csrf) // session cookie, but will be deleted asap
            .secure().httpOnly().path("/").sameSite(LAX) // Important! Must not be strict
            .encrypt(this.cookieSecretKey)
            .buildAndAddToResponse(htx);

        htx.getResponseHeaders().set("Location", redirectUrl);
        return new Authenticator.Retry(302);
    }


    private TokenTriplet oauthCodeGrantExchangeCode(final String code) {
        logOAuthTask("CODE GRANT EXCHANGE");

        HttpClient client = new HttpClient(realmInfo.tokenServiceUrl)
                .setContentTypeForm()
                .putFormParam("grant_type", "authorization_code")
                .putFormParam("code", code)
                .putFormParam("client_id", this.clientId)
                .putFormParam("redirect_uri", this.redirectUrl)
                .putFormParam("client_secret", this.clientSecret);

        HttpClientResponse response = client.doPostRequest("/token");
        if (!response.isJson200()) return null;

        final String json = new String(response.responseBody, UTF8);
        Object jwts = JsonParser.parse(json);

        return new TokenTriplet(
                (String) JsonParser.getValue(jwts, "access_token"),
                (String) JsonParser.getValue(jwts, "refresh_token"),
                (String) JsonParser.getValue(jwts, "id_token"));
    }

    @SuppressWarnings("unchecked")
    private TokenTriplet oauthRefreshTokens(final HttpExchange htx, final String refreshToken) {
        logOAuthTask("REFRESH TOKEN");

        HttpClient client = new HttpClient(realmInfo.tokenServiceUrl)
            .setContentTypeForm()
            .putFormParam("grant_type", "refresh_token")
            .putFormParam("refresh_token", refreshToken)
            .putFormParam("client_id", this.clientId)
            .putFormParam("client_secret", this.clientSecret);

        HttpClientResponse response = client.doPostRequest("/token");

        if (!response.isJson200()) {
            return null;
        }

        final String json = new String(response.responseBody, UTF8);
        Map<String, Object> jsonObj = (Map<String, Object>) JsonParser.parse(json);

        TokenTriplet tokenTriplet = new TokenTriplet(
            (String)jsonObj.get("access_token"),
            (String)jsonObj.get("refresh_token"),
            (String)jsonObj.get("id_token"));

        return tokenTriplet;
    }

    private TokenState verifyToken(final String token, final String pubKeyPem64) {
        if (null == token) return TokenState.NULLTOKEN;

        Jwt jwt = null;
        try {
            jwt = new Jwt(token);
        } catch (Exception e) {
            log.error("Unparsable Token (" +  e + ") : " + token);
            return TokenState.WRONGFORMAT;
        }

        byte[] bytesToCheck = (jwt.rawParts[0] + "." + jwt.rawParts[1]).getBytes();
        RSAPublicKey pubKeyRSA = SecurityHelpers.pem64ToPublicKeyRSA(pubKeyPem64);

        if (!SecurityHelpers.verifySignatureSHA256RSA(
                bytesToCheck, jwt.signature, pubKeyRSA))
            return TokenState.WRONGSIGNATURE;


        // LOGOUT ! set by backchannel call (sub=UserId)
        if (jwt.sub != null && this.logouts.contains(jwt.sub)) {
            this.logouts.remove(jwt.sub); // Clear Info for next dance
            return TokenState.BACKCHANNELLOGOUT;
        }

        if (null == jwt.emailVerified) {
            log.warn("Claim email_verified not found. Assuming it's ok then");
        } else if (Boolean.FALSE.equals(jwt.emailVerified)) {
            return TokenState.EMAILNOTVERIFIED;
        }

        // if set as cookie with same expires value, then never triggered, cause Token is null then
        if (jwt.exp != null) { // BackChannel-Logout-Token has no "exp"
            if (jwt.exp.longValue() * 1000 < System.currentTimeMillis()) {
                return TokenState.EXPIRED;
            }
        }

        if (jwt.azp != null && !this.clientId.equals(jwt.azp)) {
            return TokenState.WRONGAUTHPARTY;
        }

        return TokenState.VERIFIED;
    }


    private static String tokenToString(final String headline, final String token) {
        if (null == token) return "<NULL>";

        Jwt parts = new Jwt(token);

        @SuppressWarnings("unchecked")
        Map<String, Object> tokenPayload = (Map<String, Object>) JsonParser.parse(parts.payload);
        Long expires = (Long) Optional.ofNullable(tokenPayload.get("exp")).orElse(0L);

        return headline + " ---> " + new Date(expires.longValue() * 1000)
            + "\nSignature: " + Arrays.toString(parts.signature)
            + "\nHeader: " + prettyPrint(parts.header)
            + "\nPayload: " + prettyPrint(parts.payload);
    }

}
