/*
 * $Id: RateLimiter.java 5984 2024-10-31 10:19:42Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.security;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.TS;
import static de.cm.osm2po.service.security.RateLimit.Status.ERROR;
import static de.cm.osm2po.service.security.RateLimit.Status.EXPIRED;
import static de.cm.osm2po.service.security.RateLimit.Status.FORBIDDEN;
import static de.cm.osm2po.service.security.RateLimit.Status.INITIAL;
import static de.cm.osm2po.service.security.RateLimit.Status.RATELIMIT;
import static de.cm.osm2po.service.security.RateLimit.Status.VALID;
import static java.lang.System.currentTimeMillis;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.service.filters.RateLimitFilter;
import de.cm.osm2po.service.security.RateLimit.Status;
/**
 * Ueberwacht Aufrufe aufgrund der Herkunft z.B. IPs, und/oder
 * Anmeldungen, z.B. ApiKeys, und authorisiert letztere.<br>
 * Merkt sich hierfuer die Zeitstempel aller Schluessel und Adressen, wobei
 * letztere ueber ein Intervall auch wieder aufgeraeumt werden.
 * Zudem koennen optional eine Keystore- und eine Backup-Datei beruecksichtigt
 * werden. Beim Keystore handelt es sich um eine Properties-Datei mit Zeilen der
 * Form <code>[Schluessel = {@link RateLimit} als CSV]</code>
 * Aenderungen an der Keystore-Datei loesen zur Laufzeit im Refresh-Intervall
 * einen Abgleich gegen die Laufzeit-Werte aus.<br>
 * Adressen, also z.B. IPs, als Schluessel lassen sowohl eine adress-basierte
 * Ueberwachung, als auch ein einfaches Blacklisting zu.
 * Die zwar optionale, aber empfohlene Backup-Datei speichert den aktuellen
 * Live-Zustand, so dass ein Neustart hier wieder aufsetzen kann.
 * <br><br><b>Wichtig:</b><br>
 * Hier wird nicht authentifiziert, sondern nur gegen Schluessel authorisiert.
 * <br><br>
 * <strong>KeyStore (Datei) Beispiel:</strong>
 * <pre>
 * admin$=1
 * 66.6.6.6$=0
 * $myfunc=1,10
 * premiumUser$myfunc=10,1
 * basicAuthUser@1355030580=3,10
 * </pre>
 * <code>admin</code> hat vollen Zugriff.<br>
 * Die IP <code>66.6.6.6</code> ist komplett geblockt.
 * <code>myfunc</code> kann hier von <code>premiumUser</code> 10 mal in der
 * Sekunde ausgefuehrt werden, von anderen Benutzern hingegen nur 1 Mal in 10 Sekunden.
 * Ein Schluessel kann sowohl eine IP, ein API-KEY, als auch ein BasicAuth
 * sein. Durch eine geschickte Vergabe, kann also auch eine IP (hier 66.6.6.6)
 * gesperrt werden.<br>
 * Das Schluessel-Format fuer BasicAuth ist <code>user@password</code>,
 * wobei letzteres als {@link String#hashCode()} im KeyStore vorliegen muss.<br>
 * Eine Sperrung bezieht sich jedoch immer nur auf einen Schluessel.
 * Existieren mehrere Schluessel aufgrund unterschiedlicher Endpunkte,
 * so muessen diese ebenfalls explizit gesperrt, bzw. auskommentiert werden.
 *
 * @author (c) 2017/2020 - Carsten Moeller, info@osm2po.de
 * @see #validate(String, String, String)
 * @see RateLimit
 * @see RateLimitFilter
 */
public final class RateLimiter {

    private final static int DDOS_THRESHOLD = 10; // max Anzahl Fehler bis DDOS
    private final static int BACKUP_INTERVAL = 10; // Backup bei N-tem Refresh

    private final static String LOGPREFIX = "RLC "; // LogPrefix

    private Map<String, RateLimit> liveMap;

    private Log log;
    private File keystoreFile;
    private File backupFile;
    private Props internalKeystore;
    private Thread refreshRunnerThread;
    private long lastKeystoreModified;

    private long refreshInterval;
    private long maxIdle;
    private int backupInterval;

    public RateLimiter(Log log) {
        if (null == log) log = Log.empty();
        this.log = log;
        this.refreshInterval = 60000; //1 Min
        this.backupInterval = BACKUP_INTERVAL;
        this.maxIdle = 300000; //5 Min;
        this.internalKeystore = null;
        this.keystoreFile = null;
        this.backupFile = null;
    }

    public RateLimiter(Log log, File backupFile,
            long refreshInterval, int backupTrigger, long maxIdle) {
        this(log);
        this.backupFile = backupFile;
        if (this.backupFile != null) this.log.info(
                LOGPREFIX + "BackupFile: " + this.backupFile.getAbsolutePath());
        this.refreshInterval = refreshInterval;
        this.backupInterval = backupTrigger;
        this.maxIdle = maxIdle;
    }

    public RateLimiter(Log log, File backupFile,
            long refreshInterval, int backupTrigger, long maxIdle, Props keystore) {
        this(log, backupFile, refreshInterval, backupTrigger, maxIdle);
        this.internalKeystore = keystore;
    }

    public RateLimiter(Log log, File backupFile,
            long refreshInterval, int backupTrigger, long maxIdle, File keystoreFile) {
        this(log, backupFile, refreshInterval, backupTrigger, maxIdle);
        if (null == keystoreFile) {
            keystoreFile = new File("apikeys.txt");
            log.error(LOGPREFIX + "KeyStoreFile not set. Using Default");
        }
        this.keystoreFile = keystoreFile;

        this.log.info(LOGPREFIX + "KeyStoreFile: " + this.keystoreFile.getAbsolutePath());
        if (!this.keystoreFile.isFile()) log.error(LOGPREFIX + "KeyStoreFile not created yet");
    }

    public synchronized final RateLimiter start() {
        if (this.refreshRunnerThread != null) {
            log.fatal(LOGPREFIX + getClass().getSimpleName()
                    + " cannot be started twice. Call shutdown() first");
        } else {
            this.liveMap = new HashMap<String, RateLimit>();
            this.liveMap.put("$", new RateLimit(true, "1"));

            this.restoreLiveMap();
            this.refreshRunnerThread = new Thread(new RefreshRunner());
            this.refreshRunnerThread.start();
        }
        return this; // 4Chaining
    }

    private class RefreshRunner implements Runnable {
        @Override public void run() {
            try {
                int n = 0;
                boolean needsBackup = false;
                // Gleich loslegen! Wegen UnitTest
                while (true) {
                    needsBackup |= refreshLiveMap();
                    if (++n % backupInterval == 0) {
                        n = 0;
                        if (needsBackup) {
                            needsBackup = false;
                            backupLiveMap();
                        }
                    }
                    Thread.sleep(refreshInterval);
                }
            }
            catch (InterruptedException ie) {
                log.info(LOGPREFIX + "INTERRUPT "
                        + getClass().getSimpleName() + " " + TS());}
            catch (Throwable t) {
                log.error(t);}
        }
    }

    public final void close() {
        try {
            log.info(LOGPREFIX + "SHUTDOWN "
                    + getClass().getSimpleName() + " " + TS());
            this.refreshRunnerThread.interrupt();
            this.refreshRunnerThread.join();
            this.backupLiveMap();
            this.liveMap.clear();
            this.refreshRunnerThread = null; // ggf. neu-start() moeglich

        } catch (Throwable t) {log.error(t);}
    }

    public static final String validationParamsToString(
            String anonKey, String userKey, String endpoint) {
        if (null == anonKey) anonKey = "";
        if (null == userKey) userKey = "";
        if (null == endpoint) endpoint = "";
        return anonKey + ';' + userKey + ';' + endpoint;
    }

    /**
     * Validiert gegen Anmelde-Schluessel.<br>
     * Diese Variante arbeitet komplett auf String-Basis und
     * delegiert In wie Out an {@link #validate(String, String, String)}.
     * @param key String der Form 'anonKey;userKey;endpoint'
     * @return {@link Status} als String
     */
    public final synchronized String validate(final String key) {
        if (null == key) return FORBIDDEN.name();

        String[] ks = key.split(";"); // Achtung bei z.B. | da regexp
        String[] ps = new String[]{null, null, null};

        for (int i = 0; i < ks.length; i++) {
            final String k = ks[i].trim();
            if (k.length() > 0) ps[i] = k;
        }

        return this.validate(ps[0], ps[1], ps[2]).name();
    }

    /**
     * Validiert gegen Anmelde-Schluessel.
     * Zum Beispiel API-Key, BasicAuth aber auch anonyme Adresse (IP).
     * @param anonKey String Fallback: anonymer Schluessel, z.B IP. NOT NULL
     * @param userKey String optionaler Anmelde-Schluessel, z.B. API-Key, NULLABLE
     * @param endpoint String optional zusaetzlicher Endpunkt (Funktion)
     * @return Status status
     * @see RateLimit#validate()
     */
    public final synchronized Status validate(
            final String anonKey, final String userKey, final String endpoint) {
        if (null == anonKey) return ERROR;

        String foundKey = null;
        for (int priority = 0; priority < 8; priority++) {
            String key = null;
            switch (priority) {
            case 0: key = userKey + '$'; break; // Alles-Faenger. Prio 1. z.B. fuer Blacklisting
            case 1: key = anonKey + '$'; break;
            case 2: key = userKey + '$' + endpoint; break;
            case 3: key = anonKey + '$' + endpoint; break;
            case 4: key = '$' + endpoint; break; // Fallback: Template of Global Default with Endpoint
            case 5: key = userKey; break;
            case 6: key = anonKey; break;
            case 7: key = "$"; break; // Fallback: Template of Global Default without Endpoint
            }
            if (this.liveMap.containsKey(key)) {
                foundKey = key;
                break;
            }
        }

        if (null == foundKey) {
            // nix in live-map, keine defaults, dann auch nix erlaubt
            foundKey = anonKey;
            this.liveMap.put(foundKey, new RateLimit());
        }

        else if (foundKey.startsWith("$")) { // Fallback-Templates
            RateLimit tempLimit = this.liveMap.get('$' + endpoint);
            if (tempLimit != null) foundKey = anonKey + '$' + endpoint;
            else {
                tempLimit = this.liveMap.get("$");
                if (tempLimit != null) foundKey = anonKey;
            }

            if (!this.liveMap.containsKey(foundKey)) {
                RateLimit anonRateLimit = new RateLimit();
                anonRateLimit.updateConstraints(false, tempLimit);
                anonRateLimit.resetMonitoring();
                this.liveMap.put(foundKey, anonRateLimit);
                log.info(LOGPREFIX + "KEY ADDED [" + foundKey + "] " + anonRateLimit + " " + TS());
            }
        }

        RateLimit rateLimit = this.liveMap.get(foundKey); // Muss hier ungleich NULL sein!

        if (rateLimit.isDeclared() && INITIAL == rateLimit.getStatus()) {
            log.message(LOGPREFIX + "FIRST CALL [" + foundKey + "] " + rateLimit + " " + TS());
        }

        rateLimit.validate();
        Status status = rateLimit.getStatus();

        if (status != VALID) { // Status fuer Aussensicht ggf. umbiegen.
            long malCounter = rateLimit.getMalCounter();
            if (malCounter >= DDOS_THRESHOLD) {

                // DDOS setzen und Wartezeit um 10 Sekungen pro Call verlaengern
                rateLimit.penalizeDdos(10000);

                Level logLevel = DDOS_THRESHOLD == malCounter ? Level.ALERT : Level.WARN;
                log.log(LOGPREFIX + "KEY DDOS (" + DF(malCounter) + ") [" + foundKey
                        + "] " + rateLimit + " " + TS(), logLevel, null);
            } else {
                if (EXPIRED == status) log.warn(LOGPREFIX
                        + "KEY EXPIRED [" + foundKey + "] " + rateLimit + " " + TS());
                else if (RATELIMIT == status) log.warn(LOGPREFIX
                        + "KEY RATELIMIT REACHED [" + foundKey + "] " + rateLimit + " " + TS());
                else if (FORBIDDEN == status) log.warn(LOGPREFIX
                        + "KEY UNAUTHORIZED [" + foundKey + "] " + rateLimit + " " + TS());
            }
        }

        return rateLimit.getStatus(); // kann z.B. zu DDOS veraender worden sein.
    }


    private synchronized boolean refreshLiveMap() {
        boolean modified = false;
        List<String> expiredKeyList = new ArrayList<String>();

        log.progress(LOGPREFIX + "REFRESH ["
                + this.liveMap.size() + "] " + TS());

        // EXPIRED (transient keys)
        for (Entry<String, RateLimit> entry : this.liveMap.entrySet()) {
            String key = entry.getKey();
            RateLimit rl = entry.getValue();
            if (!rl.isDeclared()) {
                if (rl.getIdle() > this.maxIdle
                        && currentTimeMillis() > rl.getIntervalEnd()) {
                    expiredKeyList.add(key);
                }
            }
        }

        Props keystore = this.internalKeystore;
        if (null == keystore && this.keystoreFile != null) {
            try {
                // Tricky: lastModified liefert 0, wenn Datei nicht existiert!
                if (this.keystoreFile.lastModified() > this.lastKeystoreModified) {
                    InputStream is = new FileInputStream(this.keystoreFile);
                    keystore = new Props(true);
                    keystore.load(is);
                    is.close();
                }
            } catch (Exception e) { // implizit Exists- und Permission-Check
                keystore = null;
                log.error(LOGPREFIX + "KEYSTORE SYNC FAILED '" + this.keystoreFile + "'");
            }
        }

        // KEYSTORE SYNC (wenn vorhanden nach LiveMap)
        if (keystore != null) {

            long lastUpdate = keystore.getLastModified();
            if (this.lastKeystoreModified < lastUpdate) {
                this.lastKeystoreModified = lastUpdate;

                // KEYSTORE SYNC - CREATE / UPDATE
                String val = null;
                for (String key : keystore) {
                    if ('.' == key.charAt(0)) continue; // Hidden Template-Key

                    try {
                        val = keystore.getRawProperty(key); // Rohwert falls Exception
                        val = keystore.getProperty(key); // getProperty kann Exception werfen!
                        RateLimit rateLimit = this.liveMap.get(key);
                        if (null == rateLimit) { // CREATE NEW
                            rateLimit = new RateLimit(true, val);
                            this.liveMap.put(key, rateLimit);
                            log.info(LOGPREFIX + "KEYSTORE CREATED [" + key
                                    + "] " + rateLimit + " " + TS());
                            modified = true;
                        } else if (rateLimit.updateConstraintsCsv(true, val)) { // UPDATE
                            rateLimit.resetMonitoring();
                            log.info(LOGPREFIX + "KEYSTORE UPDATED [" + key + "] "
                                    + rateLimit + " " + TS());
                            modified = true;
                        }
                    } catch (Exception e) {log.error(
                            LOGPREFIX + "KEYSTORE KEY '" + key + "=" + val + "'");
                    }
                }

                // KEYSTORE SYNC - DELETE
                for (Entry<String, RateLimit> entry : this.liveMap.entrySet()) {
                    String key = entry.getKey();
                    RateLimit rateLimit = entry.getValue();
                    if (rateLimit.isDeclared() && !keystore.containsKey(key)) {
                        log.info(LOGPREFIX + "KEYSTORE REMOVED [" + key
                                + "] " + rateLimit + " " + TS());
                        expiredKeyList.add(key);
                    }
                }
            }
        }

        // DELETE (aus LiveMap)
        int nRemove = expiredKeyList.size();
        if (nRemove > 0) {
            int i = 0;
            for (String key : expiredKeyList) {
                RateLimit rateLimit = this.liveMap.remove(key);
                log.info(LOGPREFIX + "KEY REMOVED(" + (++i) + "/" + nRemove
                        + ") [" + key + "] " + rateLimit + " " + TS());
                modified = true;
            }
        }
        return modified;
    }

    /**
     * Schreibt die aktuellen {@link RateLimit}s in ein Backup.<br>
     * Jedoch nur, wenn Keys hinzugekommen oder geloescht,
     * nicht aber, wenn lediglich Zeitstempel neu gesetzt wurden.
     */
    private synchronized void backupLiveMap() {
        if (null == this.backupFile) {
            log.warn(LOGPREFIX + "BACKUP disabled! " + TS());
            return;
        }
        log.info(LOGPREFIX + "BACKUP [" + liveMap.size() + "] " + TS());
        boolean logData = log.isEnabled(Level.DATA);
        try {
            OutStream outStream = new OutStreamDisk(this.backupFile);
            for (Entry<String, RateLimit> entry : this.liveMap.entrySet()) {
                String key = entry.getKey();
                RateLimit rateLimit = entry.getValue();
                if (logData) log.data(LOGPREFIX + "BACKUP " + key + ":" + rateLimit
                        + "/" + rateLimit.getIdle() + "/" + rateLimit.getAge());

                new VarString(key).writeToStream(outStream);
                rateLimit.writeToStream(outStream);
            }
            outStream.close();
        } catch (Exception e) {log.error(e);}
    }

    private synchronized void restoreLiveMap() {
        if (null == this.backupFile) return;
        log.info(LOGPREFIX + "RESTORING " + TS());
        try {
            InStream inStream = new InStreamDisk(this.backupFile);
            this.liveMap.clear();
            while (!inStream.isEof()) {
                String key = new VarString().readFromStream(inStream).toString();
                RateLimit val = new RateLimit().readFromStream(inStream);
                log.debug(LOGPREFIX + "RESTORED " + val + " [" + key + "] ");
                this.liveMap.put(key, val);
            }
            inStream.close();
            log.info(LOGPREFIX + "RESTORED [" + this.liveMap.size() + "] " + TS());
        } catch (Exception e) {
            log.warn(LOGPREFIX + "RESTORE FAILED '" + this.backupFile + "'");
        }
    }

}
