/*
 * $Id: Tsp2po.java 5952 2024-10-21 11:41:05Z carsten $
 *
 * Copyright 2012/2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.tsp;

import static de.cm.osm2po.misc.Utils.DF4;
import static de.cm.osm2po.routing.AbstractRouter.newRouter;
import static de.cm.osm2po.tsp.TspUtils.INFINITY;
import static de.cm.osm2po.tsp.TspUtils.calcTourCosts;
import static de.cm.osm2po.tsp.TspUtils.createGreedyTour;
import static de.cm.osm2po.tsp.TspUtils.createInitialTour;
import static de.cm.osm2po.tsp.TspUtils.createStarTour;
import static de.cm.osm2po.tsp.TspUtils.improveInsertion;
import static de.cm.osm2po.tsp.TspUtils.improveSimulatedAnnealing;
import static de.cm.osm2po.tsp.TspUtils.improveTwoOpt;

import java.util.Arrays;
import java.util.Properties;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.Log.Level;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.routing.AbstractGraph;
import de.cm.osm2po.routing.AbstractRouter;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayGraph;
import de.cm.osm2po.routing.OverlayVertexRouter;
import de.cm.osm2po.routing.RouterConstraints;
import de.cm.osm2po.routing.RouterParameters;
import de.cm.osm2po.routing.TargetCostsAware;
import de.cm.osm2po.routing.VertexRouter;
import de.cm.osm2po.tsp.TspResultLogger.Event;
import de.cm.osm2po.tsp.TspResultLogger.EventListener;

/**
 * Osm2po spezifische Funktionen zum Loesen eines TSP.
 * @author (c) 2022 - Carsten Moeller - info@osm2po
 */
public class Tsp2po {

    private Tsp2po() {/*static*/}

    public final static class Matrix { // osm2po-Special
        public final long[][] costs;
        public final int[][][] paths;
        public Matrix(long[][] costs, int[][][] paths) {
            this.costs = costs;
            this.paths = paths;
        }

        /**
         * Generiert aus einer Besuchsreihenfolge
         * die entsprechenden Pfade.
         * @param tour int[] Besuchsreihenfolge.
         * @return int[][] Pfade.
         */
        public final int[][] getTourPaths(final int[] tour) {
            int n = tour.length -1;
            int[][] paths = new int[n][];
            for (int i = 0; i < n; i++) {
                int sourceIdx = tour[i];
                int targetIdx = tour[i + 1];
                paths[i] = this.paths[sourceIdx][targetIdx];
            }
            return paths;
        }
    }

    public static <G extends AbstractGraph> Matrix createMatrix(
            G graph, int[] vertexIds, double maxCost,
            Log log, Properties params, final double dbl2LongFactor) throws Exception {

        if (null == log) log = new Log();
        int n = vertexIds.length;

        for (int i = 0; i < n -1; i++) {
            for (int j = i + 1; j < n; j++) {
                if (vertexIds[i] == vertexIds[j]) {
                    throw new IllegalArgumentException(
                            "Intermediate Points must be differently");
                }
            }
        }

        boolean ovlMode = graph instanceof OverlayGraph;
        AbstractRouter<?> router = ovlMode
                ? newRouter(OverlayVertexRouter.class, (OverlayGraph)graph, log)
                : newRouter(VertexRouter.class, (Graph)graph, log);

        int[] cpVertexIds = Arrays.copyOf(vertexIds, n);
        long[][] matrixCosts = new long[n][n];
        int[][][] matrixPaths = new int[n][n][];


        // Untersuche alle Erreichbarkeiten und erstelle die Matrix.
        for (int y = 0; y < n; y++) {
            // Routen. Jeder ist einmal Source.
            int swap = cpVertexIds[0]; cpVertexIds[0] = cpVertexIds[y];  cpVertexIds[y] = swap;

            int sourceId = cpVertexIds[0];
            int[] targetIds = Arrays.copyOfRange(cpVertexIds, 1, n);

            log.progress("Populating Matrix Row [" + y + "] from SourceId " + sourceId);
            router.traverse(sourceId, new RouterConstraints(maxCost, targetIds), new RouterParameters(params));
            TargetCostsAware tca = (TargetCostsAware) router;

            // Matrix aufbauen
            for (int z = 0; z < n; z++) { // Zeilen
                int x = z + y; // Rotation beachten x0 ist die Diagonale.
                if (x >= n) x -= n;

                // Erreichbarkeit pruefen.
                if (x == y) {
                    matrixCosts[y][x] = 0;
                } else if (router.isVisited(vertexIds[x])) {
                    final long costs = StrictMath.round(
                            dbl2LongFactor * tca.getTargetCost(vertexIds[x]) + 0.5);
                    matrixCosts[y][x] = costs;
                    matrixPaths[y][x] = router.makePath(vertexIds[x]);
                } else {
                    matrixCosts[y][x] = INFINITY;
                    matrixPaths[y][x] = new int[0]; // Leerer Pfad
                    log.warn("TSP  Matrix [" + y + "][" + x + "] unreachable!");
                }
            }
        }
        return new Matrix(matrixCosts, matrixPaths);
    }

    public static int[] solve(Matrix mx, Log log, boolean asRing, double tolerance, String tspChain) {
        int[] tour = createInitialTour(mx.costs, asRing);
        final long initialCost = calcTourCosts(tour, mx.costs);

        final int tspN = tour.length -2;

        final EventListener tspLogger = (e,t,c,p) -> {
            final String msg = Utils.stringCropMid(Arrays.toString(t), 40)
                    + " " + e.symbol + " " + (c != 0 ? "[" + c + "] " : "") + p;
            log.log(msg, Event.IMPROVED==e ? Level.INFO : Level.PROGRESS, null);
        };

        if (tspChain != null) { // Debug Kram gesetzt
            log.info("TSP [" + tspN + "] Creating debug tour '" + tspChain + "'");
            for (char c : tspChain.toCharArray()) {
                switch (c) { // c || 0x20 = would ignore UpperCase
                case 'G': tour = createGreedyTour(mx.costs, tour); break;
                case 'S': tour = createStarTour(mx.costs, tour); break;
                case 'u': tour = improveSimulatedAnnealing(mx.costs, tour); break;
                case 'i': tour = improveInsertion(mx.costs, tour); break;
                case 't': tour = improveTwoOpt(mx.costs, tour); break;
                case 'x': tour = TspX.improvePermuteAndPrune(mx.costs, tour, tolerance, tspLogger).getAnyTour(); break;
                case 's': tour = TspS.improvePermuteAndPrune(mx.costs, tour, tolerance, tspLogger).getAnyTour(); break;
                case 'a': tour = TspA.improvePermuteAndPrune(mx.costs, tour, tolerance, tspLogger).getAnyTour(); break;
                }
            }
            if (calcTourCosts(tour, mx.costs) > initialCost) return null;
            else return tour;
        }

        // Ab hier osm2po's einigermassen austariertes Standardverhalten

        log.note("TSP [" + tspN + "] Creating initial tour");
        long improvedCost = initialCost;
        while (true) {
            // Sichern fuer alternative zweite Eroeffnung
            int[] tour2 = Arrays.copyOf(tour, tour.length);

            // Erste Eroeffnung
            tour = improveTwoOpt(mx.costs, // Third
                    improveSimulatedAnnealing(mx.costs, // Second
                            improveInsertion(mx.costs, tour))); // First
            long newCost = calcTourCosts(tour, mx.costs);

            // Zweite Eroeffnung
            tour2 = improveInsertion(mx.costs, // Third
                    improveSimulatedAnnealing(mx.costs, // Second
                            improveTwoOpt(mx.costs, tour2))); // First
            long newCost2 = calcTourCosts(tour2, mx.costs);

            if (newCost2 < newCost) { // Zweite Eroeffnung besser als erste.
                newCost = newCost2;
                tour = tour2;
            }

            if (newCost >= improvedCost) break; // keine Verbesserung mehr
            improvedCost = newCost;
        }


        if (tspN <= 40) { // Besser rechnen, wenn nicht zu viele Stationen (bei 40 sinds 5% ... bei 45 zu wenig!)
            log.note("TSP [" + tspN + "] Tour: " + Arrays.toString(tour));
            log.note("TSP [" + tspN + "] Cost: " + improvedCost);
            log.note("TSP [" + tspN + "] Trying to improve tour (may take a while)");

            if (tolerance <= 0) { // wenn Optimum gesucht
                final int tol = tspN - 25;// Optimal noch fix rechenbar
                if (tol > 0) { // Zu viele Stationen fuer optimale Berechnung
                    tolerance = tol * 0.0025; // Pro Station 0,25% draufrechnen
                    log.note("TSP [" + tspN + "] Tolerance set to " + DF4(tolerance));
                }
            }

            tour = TspX.improvePermuteAndPrune(mx.costs, tour, tolerance, tspLogger).getAnyTour();

            final long newCost = calcTourCosts(tour, mx.costs);
            if (newCost < improvedCost) {
                log.note("TSP [" + tspN + "] Tour could be improved from " + improvedCost + " to " + newCost);
                improvedCost = newCost;
            } else {
                log.note("TSP [" + tspN + "] Tour could not be improved");
            }

            log.note("TSP [" + tspN + "] " + (tolerance <= 0 ? "Tour may be optimal"
                    : "Tour is the best one found within a tolerance of " + DF4(tolerance)));
        }

        if (improvedCost > initialCost) return null;
        else return tour;
    }

} // EOC
