/*
 * $Id: TspUtils.java 6004 2024-12-17 15:04:16Z carsten $
 *
 * Copyright 2012/2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.tsp;

import static de.cm.osm2po.misc.ArrayUtils.intArrayMove;
import static de.cm.osm2po.misc.ArrayUtils.intArrayReverse;
import static de.cm.osm2po.misc.ArrayUtils.intArraySwap;
import static java.lang.System.arraycopy;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Standard- und Hilfsfunktionen fuer die Berechnung eines TSP.<br>
 * @author (c) 2022 - Carsten Moeller - info@osm2po
 * @see TspA
 * @see TspS
 */
public class TspUtils {
    private TspUtils() {/*static*/}

    public final static DateFormat DF = new SimpleDateFormat("yyMMdd-HH:mm:ss");

    public final static int FINITY_BITS = 54;
    /**
     * Konvention fuer den Wert INFINITY.<br>
     * Ein Integer-Typ kennt kein Infinity oder NaN und wechselt bei
     * Ueberschreitung das Vorzeichen.<br>
     * Daher benoetigen wir eine zentrale Konvention fuer einen hohen Long-Wert,
     * der bei aktuell 511(2^9-1) Aufsummierungen noch im posititiven Bereich verbleibt.
     * Im Umkehrschluss heisst das aber auch, dass der Max-Wert
     * bei 18.014.398.509.481.983 (53 Bits)liegt.
     */
    public final static long INFINITY = (1L << FINITY_BITS);
    public final static long FINITY = INFINITY -1;
    public final static boolean isInfinite(final long v) {return v < 0 || v >= INFINITY;}

    // ---------------------- NICE CLASSES -------------------------------

    /**
     * Partielle Kopie von {@link Random} ohne {@link AtomicLong}.
     * Ziel: Stabil bleiben, auch wenn sich Java mal aendert.
     */
    public final static class Random {
        private static final long multiplier = 0x5DEECE66DL;
        private static final long addend = 0xBL;
        private static final long mask = (1L << 48) - 1;
        private long rnd;
        public Random(long seed) {this.rnd = (seed ^ multiplier) & mask;}
        public final int nextInt() {
            this.rnd = (this.rnd * multiplier + addend) & mask;
            return (int)(this.rnd >>> (48 - 32/*int-bits*/));
        }
    }

    // ------------------------ BASICS ----------------------------------

    public static long[][] copyMatrix(final long[][] mx) {
        final int n = mx.length;
        long[][] mxc = new long[n][n];
        for (int i = 0; i < n; i++) mxc[i] = Arrays.copyOf(mx[i], n);
        return mxc;
    }

    public static int[] createInitialTour(final long[][] mx, final boolean asRing) {
        int[] tour = createInitialRingTour(mx);
        if (!asRing) tour = Arrays.copyOf(tour, tour.length -1);
        return tour;
    }

    public static int[] createInitialRingTour(final long[][] mx) {
        int n = mx.length +1;
        int[] tour = new int[n];
        for (int i = 0; i < n; i++) tour[i] = i;
        tour[n-1] = tour[0];
        return tour;
    }

    public static long calcTourCosts(final int[] ts, final long[][] mx) {
        return calcTourCosts(ts, mx, FINITY);
    }

    public static long calcTourCosts(final int[] ts, final long[][] mx, long limit) {
        if (limit > FINITY || limit < 0) limit = FINITY;
        long costs = 0;
        for (int i = 1; i < ts.length; i++) {
            long mxc = mx[ts[i-1]][ts[i]];
            if (mxc > FINITY || mxc < 0) return INFINITY;
            costs += mxc;
            if (costs > limit) return INFINITY;
        }
        return costs;
    }

    public static long calcSubTourCosts(final int[] ts, final long[][] mx, final int x1, final int x2) {
        long costs = 0;
        for (int x = x1; x < x2; x++) {
            long mxc = mx[ts[x]][ts[x+1]];
            costs += mxc;
        }
        return costs;
    }

    /**
     * Berechnet ob die Dreiecksungleichung dieser Matrix erfuellt ist.<br>
     * Triangle Inequality bedeutet im Prinzip, dass der
     * direkte Weg immer der kuerzeste ist oder zumindest gleich lang,
     * wenn ein weiterer Punkt genau auf der Strecke liegt. Aber Achtung:<br>
     * Die Berechnung hat eine Komplexitaet von O(n^3).<br>
     * Ferner berechnet diese Funktion nur 1 Dreieck. Ist ABCD gegeben und
     * jeweils ABC und BCD korrekt, so muss dies nicht heissen, dass auch
     * ABCD gegenueber AD korrekt ist, also die Ungleichung ueber 2 Stationen
     * gegeben ist.
     * @param mx long[][] Matrix-Kosten. PRE: Anzahl Zeilen gleich Spalten!
     * @return Dreiecksungleichung gegeben ja/nein.
     */
    public static boolean isTriangleInequalityGiven(final long[][] mx) {
        for (int i = 0; i < mx.length; i++) {
            for (int j = 0; j < mx.length; j++) {
                if (i == j) continue;
                for (int k = 0; k < mx.length; k++) {
                    if (k == i || k == j) continue;
                    long m = mx[i][j], v = mx[i][k] + mx[k][j];
                    if (m > v) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * Berechnet, ob die Matrix symmetrisch ist.<br>
     * Symmetrisch bedeutet, dass die Werte ueber die Hauptdiagonale
     * gespiegelt identisch sind - also Kosten Hinweg=Rueckweg.
     * @param mx long[][] Matrix-Kosten. PRE: Anzahl Zeilen gleich Spalten!
     * @return Matrix ist symmetrisch ja/nein.
     */
    public static boolean isSymmetryGiven(final long[][] mx) {
        for (int i = 0; i < mx.length; i++) for (int j = 0; j < mx.length; j++) if (mx[i][j] != mx[j][i]) return false;
        return true;
    }

    /**
     * Erzeugt aus einer Tour eine Sub-Tour.<br>
     * @param tour int[] Tour
     * @param x1 int Start-Index der Tour
     * @param x2 int End-Index der Tour (inklusive)
     * @param makeRing boolean true: Ziel gleich Start setzen
     * @param startAt0 boolean true: Start bleibt Index[0] der Originaltour
     * @return int[] neue SubTour
     * <p>
     * Beispielsweise liefert die Tour <code>0,1,2,3,4,5,0</code>
     * bei Aufruf mit x1=2 und x2=4 die folgenden Resultate fuer
     * <code>makeRing</code> und <code>startAt0</code>
     * </p>
     * <ul>
     * <li>true, true : [0, 2, 3, 4, 0]</li>
     * <li>true, false : [2, 3, 4, 2]</li>
     * <li>false, true : [0, 2, 3, 4]</li>
     * <li>false, false : [2, 3, 4]</li>
     * </ul>
     */
    public static int[] createSubTour(final int[] tour, int x1,
            int x2, boolean makeRing, boolean startAt0) {
        int nStations = x2 - x1 + 1; // Anzahl zu kopierender Stationen aus Originaltour
        int subTourLenth = nStations;
        if (startAt0) subTourLenth += 1; // +1 fuer Start [0]
        if (makeRing) subTourLenth += 1; // +1 fuer Zielstation
        int[] subTour = new int[subTourLenth];
        if (startAt0) {
            arraycopy(tour, x1, subTour, 1, nStations);
            subTour[0] = tour[0]; // nicht =0, Originaltour kann an Pos 1 auch abweichen!
        } else {
            arraycopy(tour, x1, subTour, 0, nStations);
        }
        if (makeRing) subTour[subTourLenth-1] = subTour[0];
        return subTour;
    }

    /**
     * Schnellpruefung, ob die letzten 0-1234-5 Stationen ueber 1234 permutiert (4!) optimal sind.<br>
     * Sind sie es nicht, so kann auch die Start-Tour nicht Teil einer optimalen Loesung sein
     * und ein Pruning darf hier ebenfalls ausgeloest werden.<br>
     * Ein {@link TspS} mit Greedy als Start-Tour fuer Fri26 wird hiemit sogar um Faktor 5 beschleunigt.
     * <br><br><strong>Hinweis:</strong><br>
     * Tatsaechlich werden aktuell nicht alle Permutationen hier geprueft. Ganz selten vorkommende
     * Kombinationen - aus diversen Tests ermittelt - sind derzeit nur als Kommentar im Code.
     * @param x1 int aktuelle Permutation-Tiefe. Pruefung erfolgt bis tour[x1] als gesetztes Ziel.
     * @param mx long[][] Adjazenz-Matrix
     * @param tour int[] aktuelle Tour-Reihenfolge
     * @return boolean true: Die letzten 5 Stationen sind optimal angeordnet.
     */
    public static boolean isLast5TourOptimal(final long[][] mx, final int[] tour, final int x1) {
        if (x1 > 2) {
            final int v5 = tour[x1-0]; // letzter gesetzter vertex der Start-Tour (Target)
            final int v4 = tour[x1-1]; // der Vertex vorm Target
            final int v3 = tour[x1-2]; // und davor
            final int v2 = tour[x1-3]; // und nochmal

            // Erste Array-Dimension nur 1x ermitteln. Java-Speed-Trick.
            final long[] mx2 = mx[v2];
            final long[] mx3 = mx[v3];
            final long[] mx4 = mx[v4];

            // 2! (2-34-5, 2-43-5) 01- noch nicht im Spiel
            final long oldCost2 = mx2[v3] + mx3[v4] + mx4[v5];
            final long newCost2 = mx2[v4] + mx4[v3] + mx3[v5];
            if (newCost2 < oldCost2) return false;

            if (x1 > 3) { // 3! (1-423-5, 1-432-5, 1-342-5, 1-324-5) 0- noch nicht im Spiel
                final int v1 = tour[x1-4]; // und wieder davor
                final long[] mx1 = mx[v1];

                final long oldCost3 = mx1[v2] + oldCost2;
                if (oldCost3 > mx1[v4] + mx4[v2] + mx2[v3] + mx3[v5]) return false;
                if (oldCost3 > mx1[v4] + mx4[v3] + mx3[v2] + mx2[v5]) return false;
                if (oldCost3 > mx1[v3] + mx3[v4] + mx4[v2] + mx2[v5]) return false;
//                if (oldCost3 > mx1[v3] + mx3[v2] + mx2[v4] + mx4[v5]) return false; // Selten false!

                if (x1 > 4) {  // 4! (0-1423-5, 0-1432-5, 0-1342-5, 0-1324-5) usw.
                    final int v0 = tour[x1-5]; // und wieder davor
                    final long[] mx0 = mx[v0];
                    final long oldCost4 = oldCost3 + mx0[v1];

                    final long oldCost4a = oldCost4 - mx0[v4]; // -mx04 nur 1x rechnen, sonst + in jeder Zeile
                    if (oldCost4a > mx4[v3] + mx3[v1] + mx1[v2] + mx2[v5]) return false;
                    if (oldCost4a > mx4[v3] + mx3[v2] + mx2[v1] + mx1[v5]) return false;
                    if (oldCost4a > mx4[v1] + mx1[v2] + mx2[v3] + mx3[v5]) return false;
                    if (oldCost4a > mx4[v1] + mx1[v3] + mx3[v2] + mx2[v5]) return false;
                    if (oldCost4a > mx4[v2] + mx2[v1] + mx1[v3] + mx3[v5]) return false;
                    if (oldCost4a > mx4[v2] + mx2[v3] + mx3[v1] + mx1[v5]) return false; // Selten false!

                    final long oldCost4b = oldCost4 - mx0[v3]; // -mx03 nur 1x rechnen, sonst + in jeder Zeile
                    if (oldCost4b > mx3[v4] + mx4[v2] + mx2[v1] + mx1[v5]) return false;
                    if (oldCost4b > mx3[v4] + mx4[v1] + mx1[v2] + mx2[v5]) return false;
                    if (oldCost4b > mx3[v2] + mx2[v4] + mx4[v1] + mx1[v5]) return false;
//                    if (oldCost4b > mx3[v1] + mx1[v2] + mx2[v4] + mx4[v5]) return false; // Selten false!
//                    if (oldCost4b > mx3[v1] + mx1[v4] + mx4[v2] + mx2[v5]) return false; // Selten false!
//                    if (oldCost4b > mx3[v2] + mx2[v1] + mx1[v4] + mx4[v5]) return false; // Selten false!

                    final long oldCost4c = oldCost4 - mx0[v2]; // -mx02 nur 1x rechnen, sonst + in jeder Zeile
                    if (oldCost4c > mx2[v1] + mx1[v4] + mx4[v3] + mx3[v5]) return false;
                    if (oldCost4c > mx2[v3] + mx3[v4] + mx4[v1] + mx1[v5]) return false;
                    if (oldCost4c > mx2[v4] + mx4[v3] + mx3[v1] + mx1[v5]) return false;
//                    if (oldCost4c > mx2[v4] + mx4[v1] + mx1[v3] + mx3[v5]) return false; // Selten false!
//                    if (oldCost4c > mx2[v3] + mx3[v1] + mx1[v4] + mx4[v5]) return false; // Selten false!
//                    if (oldCost4c > mx2[v1] + mx1[v3] + mx3[v4] + mx4[v5]) return false; // Selten false!
                }
            }
        }
        return true;
    }

    // ------------------------ NAIVE SOLVERS --------------------------------

    public static int[] createGreedyRingTour(final long[][] mx) {
        int[] tour = createInitialRingTour(mx);
        return createGreedyTour(mx, tour);
    }

    public static int[] createGreedyTour(final long[][] mx, final int[] initialTour) {
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        for (int y = 1; y < tour.length -2; y++) {
            long minCosts = INFINITY;
            int a = tour[y -1];
            int minX = 0;
            for (int x = y; x < tour.length -1; x++) {
                int b = tour[x];
                // TODO plus Rueckweg? (btw. durch 2 = Mittelwert noetig)
                long costs = (mx[a][b]/* + mx[b][a] */);
                if (costs < minCosts) {
                    minX = x;
                    minCosts = costs;
                }
            }
            int swap = tour[y]; tour[y] = tour[minX]; tour[minX] = swap;
        }
        return tour;
    }

    public static int[] createBestGreedyRingTour(final long[][] mx) {
        return createBestGreedyTour(mx, true);
    }

    public static int[] createBestGreedyTour(final long[][] mx, final boolean forRing) {
        if (!forRing) return createGreedyRingTour(mx);
        final int n = mx.length;
        long minCost = INFINITY;
        int[] minTour = new int[n+1];
        for (int x = 0; x < n; x++) {
            int[] tour = createInitialRingTour(mx); // 01230
            tour[0] = tour[n] = tour[x]; // bei x==1 11231
            tour[x] = 0; // bei x==1 10231
            tour = createGreedyTour(mx, tour);
            long cost = calcTourCosts(tour, mx);
            if (cost < minCost) {
                cost = minCost;
                int x0 = 0; while (tour[x0] != 0) x0++; // Wo ist die 0?
                arraycopy(tour, x0, minTour, 0, n-x0+1);
                arraycopy(tour, 0, minTour, n-x0, x0);
            }
        }
        return minTour;
    }

    /**
     * Erzeugt eine Sternen-Ring-Tour auf Grundlage einer Adjazenz-Matrix.<br>
     * Das bedeutet, wir wandern von Station 0 zu 1 und zurueck, dann zur naechst
     * gelegenen und zurueck und so weiter.
     * Die Kosten der Tour sind gewoehnlich sehr hoch, dennoch liefert diese Tour haeufig
     * eine allgemein besserere Startkonfiguration als z.B. Greedy. Greedy hat den Nachteil,
     * dass wir zwar haeufig sehr gute Ergebnisse erzielen, aber eben auch auf eine voellig
     * falsche Faehrte geschickt werden koennen (falsche Kreisrichtung),
     * was haeufig in einem Worst-Case-Szenario muenden kann.
     * @param mx long[][] Adjazenz-Matrix (Diagonale: 0, Nicht erreichbar: {@link #INFINITY})
     * @return int[] Tour
     * @see #createStarTour(long[][], int[])
     */
    public static int[] createStarRingTour(final long[][] mx) {
        int[] initialTour = createInitialRingTour(mx);
        return createStarTour(mx, initialTour);
    }

    /**
     * Erzeugt eine Tour mit Zwischenstationen sortiert nach
     * dem Abstand von Start und Ziel.<br>
     * Die Tour begint bei Station [0] und endet bei [length-1].
     * Die Stationen selbst werden nach dem Mittelwert aus Ein-
     * und Ausgangskosten vom Start und zum Ziel sortiert.
     * @param mx long[][] Adjazenz-Matrix (Diagonale: 0, Nicht erreichbar: {@link #INFINITY})
     * @param initialTour int[] aktuelle Tour-Reihenfolge
     * @return int[] Tour
     * @see #createStarRingTour(long[][])
     */
    public static int[] createStarTour(final long[][] mx, int[] initialTour) {
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        for (int i = 1; i < tour.length-1; i++) { // Zwischenstationen
            long min = INFINITY;
            for (int j = i; j < tour.length-1; j++) {
                long a = mx[tour[0]][tour[j]]; // Von Start zur Station
                // Und hin (wenn nicht Ring) bzw. zurueck zum Ziel(Start)
                long b = mx[tour[j]][tour[tour.length-1]];
                long c = (a + b) >>> 1; // Durch 2, also Mittelwert bilden
                if (c < min) {
                    min = c;
                    int swap = tour[i]; tour[i] = tour[j]; tour[j] = swap;
                }
            }
        }
        return tour;
    }

    public static int[] sortStationsByDistanceFrom(final long[][] mx, int source) {
        int[] targets = createInitialTour(mx, false);
        targets[source] = targets[0];
        targets[0] = source;
        for (int i = 1; i < targets.length; i++) {
            long min = INFINITY;
            int minJ = -1;
            for (int j = i; j < targets.length; j++) {
                long cost = mx[source][targets[j]];

                if (cost == min) {
                    if (targets[j] < targets[minJ]) {
                        minJ = j; // Reihenfolge bei Gleichheit sicherstellen
                    }

                } else if (cost < min) {
                    min = cost;
                    minJ = j;
                }
            }
            int swap = targets[i]; targets[i] = targets[minJ]; targets[minJ] = swap;
        }
        return targets;
    }


    public static int[] createTwoOptRingTour(final long[][] mx) {
        int[] tour = createInitialRingTour(mx);
        return improveTwoOpt(mx, tour);
    }

    public static int[] improveTwoOpt(final long[][] mx, final int[] initialTour) {
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        long minCosts = calcTourCosts(tour, mx);
        boolean improved = true;
        while (improved) {
            improved = false;
            for (int dx = tour.length -2; dx > 1; dx--) {
                for (int x = 1, y = x + dx -1; y < tour.length -1; x++, y++) {
                    intArrayReverse(tour, x, y); // flip (umklappen)
                    long newCosts = calcTourCosts(tour, mx, minCosts);
                    if (newCosts < minCosts) {
                        minCosts = newCosts;
                        improved = true;
                    } else {
                        intArrayReverse(tour, x, y); // Zurueckdrehen
                    }
                }
            }
        }
        return tour;
    }

    public static int[] createInsertionRingTour(final long[][] mx) {
        int[] tour = createInitialRingTour(mx);
        return improveInsertion(mx, tour);
    }

    public static int[] improveInsertion(final long[][] mx, final int[] initialTour) {
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        long minCosts = calcTourCosts(tour, mx);
        boolean improved = true;
        while (improved) {
            improved = false;
            for (int y = 1; y < tour.length -2; y++) {
                for (int x = 1; x < tour.length -1; x++) {
                    if (x == y) continue;
                    intArrayMove(tour, x, y, 1);
                    long newCosts = calcTourCosts(tour, mx, minCosts);
                    if (newCosts < minCosts) {
                        minCosts = newCosts;
                        improved = true;
                    } else {
                        intArrayMove(tour, y, x, 1);
                    }
                }
            }
        }
        return tour;
    }

    public static int[] createSimulatedAnnealingRingTour(final long[][] mx) {
        int[] tour = createInitialRingTour(mx);
        return improveSimulatedAnnealing(mx, tour);
    }

    public static int[] improveSimulatedAnnealing(final long[][] mx, final int[] initialTour) {
        int[] minTour = Arrays.copyOf(initialTour, initialTour.length);
        long minCost = calcTourCosts(minTour, mx);

        // reproduzierbares Random erzeugen
        final Random rnd = new Random(Arrays.hashCode(minTour));
        int n = minTour.length;

        int[] accTour = new int[n]; // Akzeptierte mutierte Tour (T)
        int[] curTour = new int[n]; // Aktuelle mutierte Tour (T')

        double heat = 2.0; // Hitze pro Durchlauf => Threshold
        double annealing = 0.9; // Reduktionsfaktor der Hitze
        int nMutations = n * 2000; // Probierschritte pro Hitzegrad

        while (heat > 0.01) {
            long oldCost = minCost;

            arraycopy(minTour, 0, accTour, 0, n);
            long threshold = (long) (minCost * (1.0 + heat));

            for (int i = 0; i < nMutations; i++) {
                arraycopy(accTour, 0, curTour, 0, n);

                // Single-Swap als Mutation (ohne Start und Ziel)
                int x = (rnd.nextInt() & 0x7FFFFFFF) % (n-2) +1;
                int y = (rnd.nextInt() & 0x7FFFFFFF) % (n-2) +1;
                intArraySwap(curTour, x, y);

                long curCost = calcTourCosts(curTour, mx, threshold);
                if (INFINITY == curCost) continue;

                if (curCost < threshold) { // accepted
                    arraycopy(curTour, 0, accTour, 0, n);
                    if (curCost < minCost) { // better
                        minCost = curCost;
                        arraycopy(curTour, 0, minTour, 0, n);
                    }
                }
            }

            double improvement = (double) oldCost / minCost;
            if (improvement > 1.0) {
                heat *= improvement;
            } else {
                heat *= annealing;
            }
        }

        return minTour;
    }

    public static int[] create2poRingTour(final long[][] mx) {
        int[] tour = createInitialRingTour(mx);
        return improve2po(mx, tour);
    }

    public static int[] improve2po(final long[][] mx, final int[] initialTour) {
        final long initialCost = calcTourCosts(initialTour, mx);
        long improvedCost = initialCost;
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);

        while (true) {
            // Sichern fuer alternative zweite Eroeffnung
            int[] tour2 = Arrays.copyOf(tour, tour.length);

            // Erste Eroeffnung
            tour = improveTwoOpt(mx, // Third
                   improveSimulatedAnnealing(mx, // Second
                   improveInsertion(mx, tour))); // First
            long newCost = calcTourCosts(tour, mx);

            // Zweite Eroeffnung
            tour2 = improveInsertion(mx, // Third
                    improveSimulatedAnnealing(mx, // Second
                    improveTwoOpt(mx, tour2))); // First
            long newCost2 = calcTourCosts(tour2, mx);

            if (newCost2 < newCost) { // Zweite Eroeffnung besser als erste.
                newCost = newCost2;
                tour = tour2;
            }

            if (newCost >= improvedCost) break; // keine Verbesserung mehr
            improvedCost = newCost;
        }

        return tour;
    }

    public static int[] createPermuteRingTour(final long[][] mx) {
        int[] initialTour = createInitialRingTour(mx);
        return optimizePermuteTour(mx, initialTour);
    }

    /**
     * Optimiert eine Tour mittels naiver Permuation.<br>
     * Das heisst, diese Funktion sollte besser nicht fuer Tourlaengen groesser 10
     * aufgerufen werden.
     * @param mx long[][] Adjazenz-Matrix (Diagonale: 0, Nicht erreichbar: {@link #INFINITY})
     * @param tour int[] aktuelle Tour-Reihenfolge
     * @return int[] verbesserte Tour oder <code>null</code> wenn bereits optimal.
     */
    public static int[] optimizePermuteTour(final long[][] mx, int[] tour) {
        long mincost = calcTourCosts(tour, mx);
        int[] mintour = new int[tour.length];
        final long newCost = improvePermute(1, tour.length-2, mx, tour, 0, mintour, mincost);
        return newCost < mincost ? mintour : null;
    }

    public static boolean isSubTourOptimal(final long[][] mx, int[] tour, final int x1, final int x2) {
        final long oldCost = calcSubTourCosts(tour, mx, x1, x2);
        final long newCost = improvePermute(x1+1, x2-1, mx, tour, 0, null, oldCost);

        if (oldCost > newCost) {
            return false;
        } else return true;
    }

    private static long improvePermute(int x1, int x2, final long[][] mx,
            final int[] tour, long cost, int[] mintour, long mincost) {
        if (x1 == x2) {
            cost += mx[tour[x1-1]][tour[x1]];
            cost += mx[tour[x1]][tour[x1+1]];
            if (cost <= mincost) {
                mincost = cost;
                if (mintour != null) arraycopy(tour, 0, mintour, 0, tour.length);
            }
        } else {
            for (int k = x1; k <= x2; k++) {
                final long newCost = cost + mx[tour[x1-1]][tour[k]];
                if (newCost <= mincost) {
                    int c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                    mincost = improvePermute(x1 +1, x2, mx, tour, newCost, mintour, mincost);
                    c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                }
            }
        }
        return mincost;
    }

    public static int[] improvePermute10(final long[][] mx, final int[] initialTour) {
        final int tourLen = initialTour.length;
        int[] tour = Arrays.copyOf(initialTour, tourLen);
        for (int x1 = 1, x2 = 10; x2 < tourLen-1; x1++, x2++) {
            int[] mintour = Arrays.copyOf(tour, tourLen);
            improvePermute(x1, x2, mx, tour, 0, mintour, Long.MAX_VALUE);
            tour = mintour;
        }
        return tour;
    }

    // ----------------- PRETTY PRINTER ---------------------

    public static String prettyPrintArray(long[] arr, int colwidth) {
        StringBuilder sb = new StringBuilder(arr.length * (colwidth + 3));
        final String f = "%" + colwidth + "s";
        sb.append('{');
        for (int i = 0; i < arr.length; i++) {
            if (i != 0) sb.append(", ");
            sb.append(String.format(f, arr[i]));
        }
        sb.append('}');
        return sb.toString();
    }

    public static String prettyPrintArray(int[] arr, int colwidth) {
        StringBuilder sb = new StringBuilder(arr.length * (colwidth + 3));
        final String f = "%" + colwidth + "s";
        sb.append('{');
        for (int i = 0; i < arr.length; i++) {
            if (i != 0) sb.append(", ");
            sb.append(String.format(f, arr[i]));
        }
        sb.append('}');
        return sb.toString();
    }

    public static String prettyPrintMatrix(long mx[][]) {
        if (null == mx) return null;
        final long maxValue = findMaxValue(mx);
        int n = mx.length;

        final int fl = maxValue <= 0 ? 1 : (int) StrictMath.log10(maxValue) + 1; // Anzahl max Stellen
        final String fl2 = "%" + (fl+2) + "s";
        final String fl1 = "%" + (fl+1) + "s";

        StringBuilder sb = new StringBuilder(n*n * fl);

        sb.append("{//       ");
        for (int i = 0; i < n; i++) sb.append(String.format(fl2, i));
        sb.append('\n');

        for (int y = 0; y < n; y++) {
            if (y > 0) sb.append('\n');
            sb.append("/*" + String.format("%4s", y) + " */ ");
            sb.append('{');
            for (int x = 0; x < n; x++) {
                if (x > 0) sb.append(",");
                long cost = mx[y][x];
                String scost = isInfinite(cost) ? "oo" : String.valueOf(cost);
                scost = String.format(fl1, scost);
                sb.append(scost);
            }
            sb.append("},");
        }
        sb.append("\n};");
        return sb.toString();
    }

    public static long findMaxValue(final long[][] arr) {
        int n = arr.length;
        long max = 0;
        for (int i = 0; i < n; i++) for (int j = 0; j < n; j++) {
            if (i == j) continue;
            long mxv = arr[i][j];
            if (mxv > max && mxv < INFINITY) max = mxv; // TODO Ausgabe bei Infinite sieht doof aus
        }
        return max;
    }

    public static long findMaxValue(final long[] arr) {
        int n = arr.length;
        long max = 0;
        for (int i = 0; i < n; i++) {
            long val = arr[i];
            if (val > max && val < INFINITY) max = val;
        }
        return max;
    }

    // ----------------------- ANALYSIS --------------------------------------

    public static int[][] createGreedyRingTours(final long[][] mx, boolean shift0ToLeft) {
        final int n = mx.length;
        final int[][] tours = new int[n][];
        for (int x = 0; x < n; x++) {
            int[] tour = createInitialRingTour(mx); // 01230
            tour[0] = tour[n] = tour[x]; // bei x==1 11231
            tour[x] = 0; // bei x==1 10231
            tour = createGreedyTour(mx, tour);
            if (shift0ToLeft) {
                int x0 = 0; while (tour[x0] != 0) x0++; // Wo ist die 0?
                int[] tourShift = new int[n+1];
                arraycopy(tour, x0, tourShift, 0, n-x0+1);
                arraycopy(tour, 0, tourShift, n-x0, x0);
                tour = tourShift;
            }

            tours[x] = tour;
        }
        return tours;
    }


    public static long[][] multiplyMx(final long[][] mx, final long factor) {
        long[][] mx2 = mx.clone();
        for (int y = 0; y < mx.length; y++) {
            for (int x = 0; x < mx.length; x++) {
                mx2[y][x] *= factor;
            }
        }
        return mx2;
    }


} // EOC
